Set Implicit Arguments.


Infix "&":= and (right associativity, at level 100).

Ltac nin w := induction w.
Ltac am := assumption.
Ltac rw u := (rewrite u; try trivial; try am).
Ltac wr u := (rewrite <- u; try trivial; try am).
Ltac rwi u h := rewrite u in h.
Ltac wri u h := rewrite <- u in h.
Ltac ee :=
  match goal with
  | id1:(?X1 /\ ?X2) |- _ => nin id1; ee 
  |  |- _ => idtac
  end.


Ltac cp u :=
  set (recalx := u);
   match goal with
   | recalx:?X1 |- _ => assert X1; [ exact recalx | clear recalx ]
   end.


Inductive Position : Set :=  p1  | p2 | p3 | p4 | p5.

Definition pos_int := Position_rec (fun _ : Position => nat) 1 2 3 4 5.

Lemma pos_inj: forall x y, pos_int x  = pos_int y -> x = y.
Proof. intros. induction x; induction y; simpl in H; 
  try discriminate H;trivial.
Qed.


Definition pnear x y :=
  pos_int x = S (pos_int y) \/  S (pos_int x) = pos_int y.
Definition pbefore x y :=  pos_int x < pos_int y.
Definition pleft x y :=  S (pos_int x) = pos_int y.

Inductive Color: Set :=
  Red | Green | White | Yellow| Blue | Black.

Inductive Nationality : Set :=
  English | Swedish | Danish | Norvegian | German.

Inductive Drink : Set :=
  Tea | Coffee | Milk | Beer | Water.

Inductive Cigarette :Set :=
  Pallmall | Dunhill | Blend | Bluemaster | Prince.

Inductive Animal : Set :=
  Dog | Bird | Cat | Horse | Fish.


Record House: Set :=  { 
  color : Color;
  drink : Drink; 
  cigarette: Cigarette;
  nationality : Nationality;
  animal : Animal; 
  position: Position 
}.

Axiom Pinj: forall (x y:House), 
  (color x = color y -> x = y) &
  (nationality x = nationality y -> x = y) & 
  (drink x = drink y -> x = y) & 
  (cigarette x = cigarette y -> x = y) & 
  (animal x = animal y -> x = y) &
  (position x = position y -> x = y).

Ltac choice :=
  match goal with
  | id1:(?X1 <> ?X1) |- _ => elim id1; trivial
  | |- (?X1 = ?X1) \/ ?X2 =>  left; trivial
  | |- ?X3 \/ (?X1 = ?X1) \/ ?X2 =>  right; left; trivial
  | |- ?X4 \/ ?X3 \/ (?X1 = ?X1) \/ ?X2 =>  right;right; left; trivial
  | |- ?X5 \/ ?X4 \/ ?X3 \/ (?X1 = ?X1)  \/ ?X2 
    => right;right;right; left; trivial
  | |- ?X5 \/ ?X2 \/ ?X4 \/ ?X3 \/ (?X1 = ?X1) 
    =>  right;right;right;right; trivial
  end.

Ltac choice1 :=
  match goal with
    | |- ?X1 \/ ?X2 =>  (left ; am) || (right ; choice1)
    | _ => am
end.
Lemma position_injectivity: forall p x1 x2 x3 x4 x5,
   position x1 <> position x2 ->
   position x1 <> position x3 ->
   position x1 <> position x4 ->
   position x1 <> position x5 ->
   position x2 <> position x3 ->
   position x2 <> position x4 ->
   position x2 <> position x5 ->
   position x3 <> position x4 ->
   position x3 <> position x5 ->
   position x4 <> position x5 ->
   (position x1  = p \/   position x2 = p \/  position x3  = p \/  
     position x4  = p \/  position x5 = p). 
Proof. intros.
  set (q1 := position x1) in *.
  set (q2 := position x2) in *.
  set (q3 := position x3) in *.
  set (q4 := position x4) in *.
  set (q5 := position x5) in *.
  nin p; nin q1; choice ||  
  nin q2;  choice || nin q3; choice || nin q4;  choice || nin q5;  choice.
Qed.

Lemma five_houses: forall (x x1 x2 x3 x4 x5: House),
  x1 <> x2 ->  x1 <> x3 -> x1 <> x4 ->  x1 <> x5 -> x2 <> x3 ->  
  x2 <> x4 ->  x2 <> x5 -> x3 <> x4 ->  x3 <> x5 -> x4 <> x5 ->  
  x1 = x \/ x2 =x \/ x3 =x \/ x4 =x \/ x5 = x.
Proof. intros. 
  assert (forall x y, position x = position y -> x = y). intros. cp (Pinj x0 y).
  ee. apply H15. am.  assert (forall x y, x <> y -> position x <> position y).
  intros. red. intros. elim H10. apply H9. am.
  cp (position_injectivity (position x) x1 x2 x3 x4 x5 (H10 _ _ H) (H10 _ _ H0) 
    (H10 _ _ H1) (H10 _ _ H2) (H10 _ _ H3)  (H10 _ _ H4) (H10 _ _ H5)  
    (H10 _ _ H6) (H10 _ _ H7) (H10 _ _ H8)).
  repeat (nin H11; try rw (H9 _ _ H11); try choice).
Qed.

Lemma color_extensive :
  ((exists x, color x = Red) & 
  (exists x, color x = Green) & 
  (exists x, color x = White) & 
  (exists x, color x = Yellow) & 
  (exists x, color x = Blue) ) ->
  forall x, color x = Red \/ color x = Green \/
    color x = White  \/ color x = Yellow  \/ color x = Blue.
Proof. intros. ee. nin H; nin H0; nin H1; nin H2; nin H3.
  assert (forall x y, color x <> color y -> x <> y).
  red. intros. elim H4.  rw H5.
  assert (x0 = x \/ x1 =x \/ x2 =x \/ x3 =x \/ x4 = x).
  apply five_houses; apply H4; try rw H; try rw H0; try rw H1; try rw H2;  
    try rw H3; discriminate.
  repeat nin H5; try wr H5; try choice1.
Qed.

Definition near x y := pnear (position x) (position y).
Definition variant12 :=
  (exists x, nationality x = English &  color x = Red)&
  (exists x, nationality x = Swedish & animal x = Dog)&
  (exists x, nationality x = Danish & drink x = Tea)&
  (exists x, color x = Green & drink x = Coffee)&
  (exists x, cigarette x = Pallmall & animal x = Bird)&
  (exists x, color x = Yellow & cigarette x = Dunhill)&
  (exists x, pos_int (position x) = 3 & drink x = Milk)&
  (exists x, pos_int (position x) = 1 & nationality x = Norvegian)&
  (exists x, exists y, cigarette x = Blend & animal y = Cat & near x y)&
  (exists x, exists y, animal x = Horse& 
    cigarette y = Dunhill & near x y)&
  (exists x, cigarette x = Bluemaster & drink x = Beer)&
  (exists x, nationality x = German & cigarette x = Prince)&
  (exists x, exists y, nationality x = Norvegian & color y = Blue & near x y)&
  (exists x, exists y, cigarette x = Blend & drink y = Water &near x y).



Definition before x y :=  pbefore (position x) (position y).
Definition is_left x y :=  pleft(position x) (position y).

Definition variant1 :=
  variant12 & 
  (exists x, exists y, color x =Green & color y = White & near x y).
Definition variant2 :=
  variant12 & 
  (exists x, exists y, color x =Green & color y = White & before x y).
Definition variant2 :=
  variant12 & 
  (exists x, exists y, color x =Green & color y = White & is_left x y).

Lemma five_nationalities:  variant12 ->
  exists x1, exists x2, exists x3, exists x4, exists x5,
    (nationality x1 = English) & 
    (nationality x2 = Swedish) & 
    (nationality x3 = Danish) & 
    (nationality x4 = Norvegian) & 
    (nationality x5 = German).
Proof. ir. red in H. ee. nin H. nin H0. nin H1. nin H11. nin H11. nin H10. ee.
  exists x. exists x0. exists x1. exists x2.  exists x4. intuition.
Qed.

Lemma five_houses: variant12 ->
  exists x1, exists x2, exists x3, exists x4, exists x5,
    (position x1) = p0 &
    (position x2) = p1&
    (position x3) = p2&
    (position x4) = p3&
    (position x5) = p4.
Proof. ir. red in H. ee. nin H. nin H. nin H0. nin H0. nin H1. nin H1.
  nin H11. nin H11. nin H11. nin H10. nin H10. nin H10.
Qed.


Lemma solution1 : variant1 -> 
  exists x,  nationality (f x) = German & animal (f x) = Fish.
Proof. ir. red in H. ee.
  assert (R1: nationality (f (Red)) = English). nin H. nin H. wr H14. clear H.
  assert (R2: drink (f (Green)) = Coffee). nin H3. nin H. wr H. clear H3.
  assert (R3: cigarette (f (Yellow)) = Dunhill). nin H5. nin H. wr H. clear H5.

Qed.


  assert 
Qed.

Axiom R1 :  (exists x, nationality (f x) = English &  x = Red).
Axiom R2 :  (exists x, nationality (f x) = Swedish & animal (f x) = Dog).
Axiom R3 :  (exists x, nationality (f x) = Danish & drink (f x) = Tea).
Axiom R4 :  (before Green White).
Axiom R5 :  (exists x, x = Green & drink (f x) = Coffee).
Axiom R6 :  (exists x, cigarette (f x) = Pallmall & animal (f x) = Bird).
Axiom R7 :  (exists x, x = Yellow & cigarette (f x) = Dunhill).
Axiom R8 :  (exists x, pos_int x = 3 & drink (f x) = Milk).
Axiom R9 :  (exists x, pos_int x = 1 & nationality (f x) = Norvegian).
Axiom R10 :  (exists x, exists y, cigarette (f x) = Blend & animal (f y) = Cat & near x y).
Axiom R11 : (exists x, exists y, animal (f x) = Horse & cigarette (f y) = Dunhill & near x y).
Axiom R12 :  (exists x, cigarette (f x) = Bluemaster & drink (f x) = Beer).
Axiom R13 :  (exists x, nationality (f x) = German & cigarette (f x) = Prince).
Axiom R14 :  (exists x, exists y, nationality (f x) = Norvegian & y = Blue & near x y).
Axiom R15 :  (exists x, exists y, cigarette (f x) = Blend & drink (f y) = Water & near x y).



Lemma pos_blue: pos_int (Blue) = 2.
Proof. nin R14. nin H. nin R9. ee.  
  assert (x = x1). apply f_inj. pose (Pinj (f x) (f x1)). ee. pose Pinj. ee.
  apply H5. rw H1. rwi H2 H3. wri H4 H0. red in H3. rwi H0 H3. symmetry.
  induction H3. assert (0= pos_int Blue). auto with arith.
  nin (pos2 Blue). wri H5 H6. elim (le_Sn_O 0). am. am.
Qed.

Lemma pos_green1 : pos_int (Green) = 1 \/ pos_int (Green) = 4.
Proof. assert (before Green White). apply R4. red in H. nin (pos2 White).
  nin (pos2 Green). set (x:= pos_int Green) in *. 
  cp (lt_le_trans _ _ _ H H1). cp (lt_n_Sm_le _ _ H4).
  nin (le_lt_or_eq _ _ H5).  cp (lt_n_Sm_le _ _ H6).
  nin (le_lt_or_eq _ _ H7). cp (lt_n_Sm_le _ _ H8).
  nin (le_lt_or_eq _ _ H9).  cp (lt_n_Sm_le _ _ H10).
  left. apply le_antisym. am. am.
  assert (pos_int Green = pos_int Blue). rw pos_blue.  cp (pos1 H11).
  discriminate H12.
  assert (drink (f (Green)) = Coffee). nin R5.  nin H9. wr H9.
  nin R8. nin H10. wri H8 H10. unfold x in H10. rwi (pos1 H10) H11.
  rwi H9 H11. discriminate H11. right. am.
Qed.

Lemma nat_red: nationality (f (Red)) = English.
Proof. nin R1. nin H. wr H0.

Lemma drink_green: drink (f (Green)) = Coffee.
Proof. nin R5.  nin H. wr H. Qed.

Lemma cig_yellow: cigarette (f (Yellow)) = Dunhill.
Proof. nin R7.  nin H. wr H. Qed.


Lemma E1: exists x, nationality x = English.
Proof. induction R1. induction H. exists x. assumption. Qed.
Lemma E2: exists x, nationality x = Swedish.
Proof. induction R2. induction H. exists x. assumption. Qed.
Lemma E3: exists x, nationality x = Danish.
Proof. induction R3. induction H. exists x. assumption. Qed.
Lemma E4: exists x, nationality x = Norvegian.
Proof. induction R9. induction H. exists x. assumption. Qed.
Lemma E5: exists x, nationality x = German.
Proof. induction R13. induction H. exists x. assumption. Qed.





EOF.
