(** * Elements of mathematics:  Algebra : Z
  Copyright INRIA (2010) Apics Team (Jose Grimm).
*)

Require Export ZArith.
Require Export set11.


Set Implicit Arguments.
(* $Id: salg1.v,v 1.1 2010/08/09 06:30:35 grimm Exp $ *)



Ltac qprops := auto with qprops.
Ltac qw := autorewrite with qw;  tv.
Ltac qwi u:= autorewrite with qw in u.
Ltac qwii u:= autorewrite with qw in u; tv.


Import InfiniteSets.




Module RationalIntegers.
(** ** The set of rational integers *)

(** Z is the disjoint union of N* and N; [J x TPb] is positive, [J y TPa] is negative *)

(** We define 0 and the injections [N -> Zp] and [N -> Zm] *)
Definition BZ_of_nat x := J x TPb.
Definition BZ_zero := BZ_of_nat card_zero.
Definition BZm_of_nat x := Yo (x= card_zero) BZ_zero (J x TPa).

(** We define 0 1 2 3 4 and -1 as elements of Z *)
Definition BZ_one := BZ_of_nat card_one.
Definition BZ_two := BZ_of_nat card_two.
Definition BZ_three := BZ_of_nat card_three.
Definition BZ_four := BZ_of_nat card_four.
Definition BZ_mone := BZm_of_nat card_one.

(** We define Z Zp Zm Zps Zms and Ns (p=positive, m=negative s = nonzero) *)
Definition Bnats:= Zo Bnat (fun x=> x <> card_zero). 
Definition BZ := canonical_du2 Bnats Bnat.
Definition BZms:= product Bnats (singleton TPa).
Definition BZp:= product Bnat (singleton TPb).
Definition BZps:= product Bnats (singleton TPb).
Definition BZm:= tack_on BZms BZ_zero.

(** We show that some sets are subsets of others *) 

Lemma sub_BZp_BZ : sub BZp BZ.
Proof. red. ir. ufi BZp H. uf BZ. rw canonical_du2_rw. app union2_second.
Qed.
Lemma sub_BZps_BZp : sub BZps BZp.
Proof. red. ir. ufi BZps H. awi H. uf BZp. aw. ee. am.  ufi Bnats H0. Ztac. 
  am. am.
Qed.

Lemma sub_BZps_BZ : sub BZps BZ.
Proof. app (sub_trans sub_BZps_BZp sub_BZp_BZ).
Qed.

Lemma sub_BZms_BZm : sub BZms BZm.
Proof. red. ir. uf BZm. fprops.
Qed.

Lemma sub_BZms_BZ : sub BZms BZ.
Proof. red. ir. ufi BZms H. uf BZ.  rw canonical_du2_pr. awi H. ee. am.
  left. ee. am. am.
Qed.

Lemma sub_BZm_BZ : sub BZm BZ.
Proof. uf BZm. red. ir. rwi tack_on_inc H. nin H. app  sub_BZms_BZ.
  rw H. uf BZ_zero. uf BZ_of_nat. app sub_BZp_BZ. uf BZp. aw. eee.
Qed.

Hint Resolve sub_BZm_BZ sub_BZms_BZ sub_BZp_BZ sub_BZps_BZ : qprops. 
Hint Resolve sub_BZms_BZm sub_BZps_BZp : qprops.

(** We show that the injections have values in Zp and Zm *)
Lemma inc_BZ_of_natp : forall x, inc x Bnat -> inc (BZ_of_nat x) BZp.
Proof. ir. uf BZp. uf BZ_of_nat. aw. eee.
Qed.

Lemma inc_BZ_of_nat : forall x, inc x Bnat -> inc (BZ_of_nat x) BZ.
Proof. ir. app sub_BZp_BZ. app inc_BZ_of_natp.
Qed.

Lemma inc_BZm_of_natms : forall x, inc x Bnat -> x <> card_zero ->
  inc (BZm_of_nat x) BZms.
Proof. ir. uf BZms. uf BZm_of_nat. rw Y_if_not_rw. aw. eee. uf Bnats. Ztac. am.
Qed.

Lemma inc_BZm_of_natm : forall x, inc x Bnat -> inc (BZm_of_nat x) BZm.
Proof. ir. uf BZms. uf BZm. nin (equal_or_not x card_zero).
  uf BZm_of_nat. rww Y_if_rw. fprops. cp (inc_BZm_of_natms H H0). fprops.
Qed.

Lemma inc_BZm_of_nat : forall x, inc x Bnat -> inc (BZm_of_nat x) BZ.
Proof. ir. cp (inc_BZm_of_natm H). qprops.
Qed.

Lemma inc_BZps_nonzero: forall x, inc x BZps -> P x <> card_zero.
Proof. ir.  ufi BZps H. awi H. ee. ufi Bnats H0. Ztac. am. 
Qed.

Lemma inc_BZps: forall x, inc x BZps = (inc x BZp & x <> BZ_zero).
Proof. ir. uf BZps; uf BZp; uf BZ_zero; uf BZ_of_nat. app iff_eq.
  ir. awi H. nin H. nin H0.  ufi Bnats H0. Ztac.
  aw. eee. dneg. rw H4. aw. 
  ir. nin H.  awi H. ee. aw. eee. uf Bnats. Ztac. dneg. wr H3. wr H2. aw.
Qed.


Ltac TP_tac:= 
  match goal with
    | H: TPa = TPb |- _ => elim (two_points_distinct H)
    | H: TPb = TPa |- _ => elim (two_points_distinctb H)
    | H1: ?x = TPa, H2: ?x = TPb  |- _ =>
      rwi H2 H1;  elim (two_points_distinctb H1)
    | H: ?x <> ?x  |- _ => elim H; tv
    | H: ?x = TPa |- ?x <> TPb  =>  rw H; fprops
    | H: ?x = TPb |- ?x <> TPa  =>  rw H; fprops
end.

(** We show that [N -> Zp] and [N -> Zm] are injective *)
Lemma BZ_of_nat_inj: forall x y,
  BZ_of_nat x = BZ_of_nat y -> x = y.
Proof. uf BZ_of_nat.  ir. cp (f_equal P H). awi H0. am. Qed.

Lemma BZm_of_nat_inj: forall x y,
  BZm_of_nat x = BZm_of_nat y -> x = y.
Proof. uf BZm_of_nat.  ir. nin (equal_or_not x card_zero).
  rwi Y_if_rw H.  nin (equal_or_not y card_zero). ue. rwi Y_if_not_rw H. 
  ufi BZ_zero H. ufi BZ_of_nat H. cp (f_equal Q H). awi H2. TP_tac. am. am.
  rwii Y_if_not_rw H.  nin (equal_or_not y card_zero).  rwii Y_if_rw H. 
  ufi BZ_zero H. ufi BZ_of_nat H. cp (f_equal Q H). awi H2. TP_tac. 
  rwii Y_if_not_rw H. cp (f_equal P H). awi H2. am.
Qed.

Lemma BZm_of_nat_inj_bis: forall x y,
  BZm_of_nat x = BZ_of_nat y -> (x = y & x = card_zero).
Proof. uf BZm_of_nat.  ir. nin (equal_or_not x card_zero).
  rwi Y_if_rw H. ufi BZ_zero H. cp (BZ_of_nat_inj H). rw H0. rw H1. au. am.
  rwi Y_if_not_rw H.  ufi BZ_of_nat H. cp (f_equal Q H). awi H1. TP_tac. am.
Qed.

Lemma BZ_zero_if_abs0: forall x, inc x BZ -> P x = card_zero -> x = BZ_zero.
Proof. ir. ufi BZ  H. rwi canonical_du2_pr H. ee. nin H1. nin H1.
  ufi Bnats H1. Ztac. contradiction. uf BZ_zero. uf BZ_of_nat. nin H1. 
  wr H2. wr H0. aw.
Qed.


(** We show that 0, 1, etc are rational integers *)
Lemma inc0_BZ : inc BZ_zero BZ.
Proof. uf BZ_zero. app inc_BZ_of_nat. fprops.
Qed.

Lemma inc0_BZp : inc BZ_zero BZp.
Proof. uf BZ_zero. uf BZp.  uf BZ_of_nat. aw. eee. 
Qed.

Lemma inc1_BZ : inc BZ_one BZ.
Proof. uf BZ_one. app inc_BZ_of_nat. fprops.
Qed.

Lemma inc1_BZps : inc BZ_one BZps.
Proof. uf BZ_one. uf BZ_of_nat. uf BZps. aw. eee. uf Bnats. Ztac.
  fprops. app card_one_not_zero.
Qed.

Lemma inc2_BZ : inc BZ_two BZ.
Proof. uf BZ_two. app inc_BZ_of_nat. fprops.
Qed.

Lemma inc3_BZ : inc BZ_three BZ.
Proof. uf BZ_three. app inc_BZ_of_nat. uf card_three. fprops.
Qed.

Lemma inc4_BZ : inc BZ_four BZ.
Proof. uf BZ_four. app inc_BZ_of_nat. uf card_four. app  inc_succ_Bnat.
  uf card_three. fprops.
Qed.

Lemma inc_m1_BZ : inc BZ_mone BZ.
Proof. uf BZ_mone. app inc_BZm_of_nat. fprops.
Qed.


Hint Resolve inc0_BZ inc1_BZ inc_m1_BZ: qprops.


(**  More properties of integers *)
Lemma inc_BZ_P: forall a, inc a BZ -> inc (P a) Bnat.
Proof. ir. ufi BZ H. rwi canonical_du2_rw H. nin (union2_or H). 
  awi H0.  ee. ufi Bnats H1.  Ztac. am. awi H0. ee. am.
Qed.


Lemma BZ_of_nat_P: forall x, P (BZ_of_nat x) = x.
Proof. ir. uf BZ_of_nat. aw. Qed.

Lemma BZm_of_nat_P: forall x, P (BZm_of_nat x) = x.
Proof. ir. uf BZm_of_nat. nin (equal_or_not x card_zero).
  rww Y_if_rw. uf BZ_zero. sy. rww  BZ_of_nat_P.
  rww Y_if_not_rw. aw. 
Qed.

Lemma BZp_Q: forall x, inc x BZp -> Q x = TPb.
Proof. ir. ufi BZp H. awi H. nin H. ee. am. Qed.

Lemma BZps_Q: forall x, inc x BZps -> Q x = TPb.
Proof. ir. ufi BZps H. awi H. nin H. ee. am. Qed.
Lemma BZms_Q: forall x, inc x BZms -> Q x = TPa.
Proof. ir. ufi BZms H. awi H. nin H. ee. am. Qed.


Lemma BZ_zero_P: P BZ_zero = card_zero.
Proof. uf BZ_zero. uf BZ_of_nat. aw. Qed.

Lemma BZ_zero_Q: Q BZ_zero = TPb.
Proof. uf BZ_zero. uf BZ_of_nat. aw. Qed.


Hint Rewrite  BZ_zero_P BZ_zero_Q: qw.

Lemma inc_BZms_pr: forall x, inc x BZms -> 
  (P x <> card_zero & BZm_of_nat (P x) = x).
Proof. uf BZms; uf BZm_of_nat.  uf Bnats. ir. awi H. nin H. nin H0. Ztac.
  ee. am. rww Y_if_not_rw. wr H1.  aw. 
Qed.

Lemma inc_BZp_pr: forall x, inc x BZp -> (BZ_of_nat (P x) = x).
Proof. uf BZp; uf BZ_of_nat. ir. awi H. nin H. nin H0. wr H1. aw. 
Qed.

Lemma inc_BZm_pr: forall x, inc x BZm -> 
  BZm_of_nat (P x) = x.
Proof. uf BZm.  ir. rwi tack_on_inc H. nin H. nin (inc_BZms_pr H); am.
  rw H.  qw. uf BZm_of_nat. rww Y_if_rw.
Qed.

Lemma inc_BZ_pr: forall a, inc a BZ ->
  a =  BZ_of_nat (P a) \/ a =  BZm_of_nat (P a).
Proof. ir. ufi BZ H. rwi canonical_du2_rw H. nin (union2_or H). right.
  uf BZm_of_nat.  awi H0. ee. rww Y_if_not_rw.  wr H2. aw.
  ufi Bnats H1. Ztac. am. left. awi H0. ee. uf BZ_of_nat. wr H2. aw.
Qed.


Lemma BZ_one_not_zero: BZ_one <> BZ_zero.
Proof. uf BZ_one.  uf BZ_zero. red. ir. cp (BZ_of_nat_inj H).
  elim card_one_not_zero. am.
Qed.

Lemma BZ_mone_not_zero: BZ_mone <> BZ_zero.
Proof. uf BZ_mone.  uf BZ_zero. red. ir. cp (f_equal P H).
  rwi BZm_of_nat_P H0. rwi BZ_of_nat_P H0. elim card_one_not_zero. am.
Qed.

(** We have a partition of Z as Zp and Zms, but there are more such partitions *)

Lemma inc_BZ: forall x, inc x BZ = (inc x  BZms \/ inc x BZp).
Proof.  ir. ap iff_eq. uf BZ; uf BZms; uf BZp. rw canonical_du2_pr. 
  ir. nin H. nin H0. ee. left. aw. eee. ee. right. aw. eee. ir. 
  nin H; qprops. 
Qed.

Lemma inc_BZ1: forall x, inc x BZ =
  (x = BZ_zero \/ inc x BZps \/ inc x BZms).
Proof. ir. app iff_eq. rw inc_BZ. ir. nin (equal_or_not x BZ_zero). au.
  nin H. au. right. left. rw inc_BZps. au.
  ir. nin H. rw H. qprops. nin H; qprops.
Qed.

Lemma inc_BZ2: forall x, inc x BZ = (inc x  BZps \/ inc x BZm).
Proof.  ir. ap iff_eq. ir. nin (equal_or_not x BZ_zero). right.
  uf BZm. rw H0. fprops. rwi inc_BZ H. nin H. right. uf BZm. fprops.
  left. rw inc_BZps.  au. ir. nin H; qprops.
Qed.

Lemma BZ_disjoint_neg_0 : forall x, inc x  BZms -> x = BZ_zero -> False.
Proof. ir. assert (Q x = TPa). rww BZms_Q. assert (Q x = TPb). rw H0. 
  qw. TP_tac.
Qed.

Lemma BZ_disjoint_pos_0 : forall x, inc x BZps -> x = BZ_zero -> False.
Proof. ir. rwi inc_BZps H. nin H. contradiction.
Qed.

Lemma BZ_disjoint_neg_pos : forall x, inc x BZms -> inc x BZp -> False.
Proof. ir. assert (Q x = TPa). rww BZms_Q. assert (Q x = TPb). rww BZp_Q.
  TP_tac.
Qed.

Lemma BZ_disjoint_pos_neg : forall x, inc x  BZps -> inc x BZm -> False.
Proof. uf BZps ; uf BZm; uf BZms. uf BZ_zero.  uf BZ_of_nat. ir.
  awi H. ee. rwi tack_on_inc H0. nin H0. awi H0. ee.  rwi H4 H2.  
  TP_tac. rwi H0 H1. awi H1. ufi Bnats H1. Ztac. TP_tac.
Qed.

Lemma BZ_disjoint_neg_spos : forall x, inc x  BZms -> inc x BZps -> False.
Proof. ir. cp (sub_BZms_BZm H). ap  (BZ_disjoint_pos_neg H0 H1).
Qed.

(** We show that Ns and Z are infinite countable *)

Lemma equipotent_Bnats_Bnat: equipotent Bnats Bnat.
Proof. 
  assert (is_countable_set Bnats). rw countable_prop. assert (sub Bnats Bnat). 
  uf Bnats. Ztac. ap Z_sub. ap (sub_smaller H). 
  nin (countable_finite_or_N_b H).
  assert (Bnat = tack_on Bnats card_zero). uf Bnats. set_extens.  
  nin (equal_or_not x card_zero). rw H2. fprops. 
  app inc_tack_on_y. Ztac. rwi tack_on_inc H1.  nin H1. Ztac. am. rw H1. fprops.
  assert (~ (inc card_zero  Bnats)).  red. ir. ufi Bnats H2. Ztac. elim H4. tv.
  cp (cardinal_succ_pr H2). wri H1 H3. red in H0. cp (is_finite_succ1 H0).
  assert (succ Bnats = succ (cardinal Bnats)). uf succ. app cardinal_sum_pr2. 
  fprops.  fprops. wri H5 H4. wri H3 H4.  nin infinite_Bnat. elim H7; am. am.
Qed.

Lemma infinite_countable_Z: cardinal BZ = cardinal Bnat.
Proof. ir.
  set (x := card_plus Bnats Bnat). assert (x = cardinal BZ).
  uf x. uf card_plus. uf cardinal_sum. tv.
  assert (card_plus Bnats Bnat = card_plus (cardinal Bnat) (cardinal Bnat)).
  app cardinal_sum_pr2. eqtrans Bnat. app equipotent_Bnats_Bnat.
  fprops. fprops. wr H. uf x.  rw H0. app sum2_infinite.
  fprops.  app infinite_Bnat.
Qed.

(** ** Oppposite *)

(** The oppositive of a number is obtained by swapping TPa and TPb in the second component, except for zero *)
Definition BZopp:= fun x =>
  Yo (Q x = TPa) (BZ_of_nat (P x))(BZm_of_nat (P x)).

Definition BZopp_fun := BL BZopp BZ BZ.

Lemma BZopp_zero : BZopp (BZ_zero) = BZ_zero.
Proof. uf BZopp. uf BZ_zero. uf BZ_of_nat.  aw. rww Y_if_not_rw. 
  uf BZm_of_nat. rww Y_if_rw. fprops.
Qed.

Hint Rewrite BZopp_zero : qw.

(** Oppositive maps Zp to Zm and Zps to Zms. It is an involution of Z  *)
Lemma BZopp_positive1 : forall x, inc x BZps -> inc (BZopp x) BZms.
Proof. ir. uf BZopp. ufi BZps H. awi H. ee. rw Y_if_not_rw. 
  uf BZm_of_nat. uf BZms. rww Y_if_not_rw.  aw. eee. ufi Bnats H0.
  Ztac. am.  TP_tac. 
Qed.

Lemma BZopp_positive2 : forall x, inc x BZp -> inc (BZopp x) BZm.
Proof. ir. uf BZm. nin (equal_or_not x  BZ_zero). rw H0. qw. fprops.
  app inc_tack_on_y. app BZopp_positive1. rw inc_BZps. au.
Qed.

Lemma BZopp_negative1 : forall x, inc x BZms -> inc (BZopp x) BZps.
Proof. ir. uf BZopp.  ufi BZms H. awi H. ee. rww Y_if_rw.  
  uf BZps. uf BZ_of_nat. aw. eee.
Qed. 

Lemma BZopp_negative2 : forall x, inc x BZm -> inc (BZopp x) BZp.
Proof. ir. ufi BZm H. rwi tack_on_inc H. nin H. cp (BZopp_negative1 H).
  rwi inc_BZps H0. nin H0. am. rw H. qw. app inc0_BZp.
Qed.

Lemma BZ_stable_opp: forall x, inc x BZ -> inc (BZopp x) BZ.
Proof. ir. rwi inc_BZ1 H. nin H. rw H. qw. qprops.  
  nin H. app sub_BZms_BZ. app  BZopp_positive1.
  app sub_BZps_BZ. app  BZopp_negative1.
Qed.

Hint Resolve BZ_stable_opp : qprops.

Lemma BZopp_P: forall x, P (BZopp x) = P x.
Proof. ir. uf BZopp. nin (equal_or_not (Q x) TPa). rww Y_if_rw. 
  rww BZ_of_nat_P. rww  Y_if_not_rw. rww BZm_of_nat_P.
Qed.

Lemma BZopp_involutive: forall x, inc x BZ -> BZopp (BZopp x) =x.
Proof. ir. set (y:= BZopp x). assert (y= BZopp x). tv. 
  assert (P y = P x). rw H0. app BZopp_P. 
  rwi inc_BZ1 H. nin H. rw H0. rw H. qw.
  nin H. uf BZopp. rww Y_if_rw. rw H1. ir. app inc_BZp_pr. qprops. 
  rw H0. uf BZopp. rw Y_if_not_rw. uf BZm_of_nat.  
  rw Y_if_not_rw. aw. app inc_BZps_nonzero. rw BZps_Q. fprops. am.
  uf BZopp. rww Y_if_not_rw. rw H1. app inc_BZm_pr. qprops.
  rw H0. uf BZopp. rww Y_if_rw. uf BZ_of_nat. aw. fprops. rww BZms_Q.
Qed.

Lemma BZopp_inj: forall a b, inc a BZ -> inc b BZ ->
  BZopp a = BZopp b -> a = b.
Proof.  ir. wr (BZopp_involutive H).  wr (BZopp_involutive H0).  rww H1.
Qed.

Lemma bijective_BZopp: bijective BZopp_fun.
Proof. red. assert (transf_axioms BZopp BZ BZ). red. ir. qprops.
  uf BZopp_fun. split. app bl_injective. ir. app BZopp_inj.
  app bl_surjective. ir. wr (BZopp_involutive H0). 
  exists (BZopp y). split. qprops. tv.
Qed.

Lemma BZopp_pr1: forall x, inc x Bnat ->
   BZopp (BZ_of_nat x) =  BZm_of_nat x.
Proof.  ir. uf BZopp. uf BZ_of_nat. aw. rww Y_if_not_rw. fprops.
Qed.

Lemma BZopp_pr2: forall x, inc x Bnat ->
   BZopp (BZm_of_nat x) =  BZ_of_nat x.
Proof.  ir. uf BZm_of_nat. nin (equal_or_not x card_zero).
  rww Y_if_rw. rw H0. qw. rww Y_if_not_rw. uf BZopp. aw. rww Y_if_rw.
Qed.

(** ** Absolute value *)

(** Absolute value is the number with a positive sign; It is the number of its opposite *)
Definition BZabs x := BZ_of_nat (P x).

Lemma BZ_abs_pos: forall x, inc x BZp -> BZabs x = x.
Proof. ir. uf BZabs. app inc_BZp_pr. 
Qed.

Lemma BZ_abs_neg: forall x, inc x BZms -> BZabs x = BZopp x.
Proof. ir. uf BZabs. uf BZopp. rww Y_if_rw. rww BZms_Q. 
Qed.

Lemma inc_BZabs_N: forall x, inc x BZ -> inc (BZabs x) BZp.
Proof. ir. uf BZp. uf BZabs. uf BZ_of_nat. aw. eee. app inc_BZ_P.
Qed. 

Lemma BZ_stable_abs: forall x, inc x BZ -> inc (BZabs x) BZ.
Proof. ir. app sub_BZp_BZ. app inc_BZabs_N.
Qed. 

Lemma BZabs_P: forall x, inc x BZ -> P (BZabs x) = P x.
Proof. ir. uf BZabs. uf BZ_of_nat. aw.
Qed.

Lemma BZabs_projection: forall x, inc x BZ -> BZabs (BZabs x) = BZabs x.
Proof. ir. ap BZ_abs_pos. app inc_BZabs_N.
Qed.
 
Lemma BZabs_opp: forall x, inc x BZ -> BZabs (BZopp x) = BZabs x.
Proof. ir. uf BZabs.  rww BZopp_P. 
Qed.

Lemma BZ_abs_zero : BZabs BZ_zero = BZ_zero.
Proof. ir. uf BZabs. rww BZ_zero_P. Qed.


Lemma BZ_abs_zero_rv : forall x,  inc x BZ -> BZabs x = BZ_zero 
  -> x = BZ_zero.
Proof. ir. ufi BZabs H0. ufi BZ_zero H0. cp (BZ_of_nat_inj H0).
  app BZ_zero_if_abs0.
Qed.

Hint Resolve BZ_stable_abs: qprops.

(** ** Oredering on Z*)

(** We  use the ordinal sum of the opposite ordering of Ns and N. This is clearly a total ordering *)
Definition BZ_ordering:=
  ordinal_sum2 (opposite_order (induced_order Bnat_order Bnats)) Bnat_order.

Lemma BZ_order_aux: 
  order (opposite_order (induced_order Bnat_order Bnats))
  & order Bnat_order.
Proof.  nin (Bnat_order_worder). 
  assert (sub Bnats (substrate Bnat_order)). rw Bnat_order_substrate. uf Bnats.
  app Z_sub. split. fprops. am.
Qed.

Lemma BZ_order: order BZ_ordering.
Proof. nin BZ_order_aux. uf BZ_ordering. app ordinal_sum2_order. 
Qed.

Lemma substrate_BZ_order: substrate BZ_ordering = BZ.
Proof. nin BZ_order_aux. uf BZ_ordering. rww ordinal_sum2_substrate.
  nin (Bnat_order_worder). clear H0.
  cp  Bnat_order_substrate.
  assert (sub Bnats (substrate Bnat_order)). rw H0. uf Bnats. app Z_sub.
  aw. rww H0. fprops. 
Qed.

Lemma BZ_totalorder: total_order BZ_ordering.
Proof. uf BZ_ordering. 
  cp (Bnat_order_worder). cp (worder_total H).  nin H. clear H1.
  cp Bnat_order_substrate.
  assert (sub Bnats (substrate Bnat_order)). rw H1. uf Bnats. app Z_sub.
  cp (total_order_sub H0 H2). cp (total_order_opposite H3).
  set (Bnatr := opposite_order (induced_order Bnat_order Bnats)) in *.
  assert (order Bnatr). nin H4. am. app ordinal_sum2_totalorder.  
Qed.


(** We define le, lt, ge and gt on Z, and state properties of this operation *)
Definition BZ_le x y:= inc x BZ & inc y BZ &
  ((Q x = TPa &  Q y = TPa & cardinal_le (P y) (P x))
    \/ (Q x = TPa & Q y = TPb)
    \/ (Q x = TPb & Q y = TPb & cardinal_le (P x) (P y))).
Definition BZ_lt x y:= BZ_le x y & x <> y.
Definition BZ_ge x y:= BZ_le y x.
Definition BZ_gt x y:= BZ_lt y x.

Lemma BZ_le_pr : forall x y, gle BZ_ordering x y = BZ_le x y.
Proof. ir. nin (BZ_order_aux). uf BZ_ordering. rww ordinal_sum2_gle.
  cp (ordinal_sum2_substrate H H0). cp substrate_BZ_order. ufi BZ_ordering  H2.
  rwi H2 H1. wr H1.
  assert (order (induced_order Bnat_order Bnats)). cp Bnat_order_substrate.
  assert (sub Bnats (substrate Bnat_order)). rw H3. uf Bnats. app Z_sub. fprops.
  ap iff_eq. ir. ee. uf BZ_le. ee. am. am. nin H6. left. ee. am.  am. 
  awi H8.  rwi Bnat_order_le H8. red in H8. ee; am.
  ufi BZ H5.  rwi canonical_du2_pr H5. nin H5. nin H9. ee; am. nin H9.
  TP_tac. ufi BZ H4. rwi canonical_du2_pr H4. nin H4. nin H9. ee; am. nin H9.
  TP_tac. am. nin H6. right. right. nin H6. ufi BZ H4.
  rwi canonical_du2_pr H4. nin H4.  nin H8. nin H8. contradiction. nin H8. 
  split. am. 
  ufi BZ H5. rwi canonical_du2_pr H5. nin H5. nin H10. nin H7. nin H10.
  contradiction. eee.  rwi Bnat_order_le H11. red in H11. ee; am.
  right. left. ee. am. ufi BZ H5. rwi canonical_du2_pr H5. nin H5. nin H8.
  ee. contradiction.  ee. am. 
  ir. red in H4. eee. cp (inc_BZ_P H4). cp (inc_BZ_P H5).
  nin H6.  left. eee. aw. rw Bnat_order_le. red. eee.
  ufi BZ H5. rwi canonical_du2_pr H5. nin H5. nin H11. ee. am. nin H11. TP_tac.
  ufi BZ H4. rwi canonical_du2_pr H4. nin H4. nin H11. ee. am. nin H11. TP_tac.
  nin H6. right. right. ee. am. rw H9. fprops. right. left. ee. rw H6. 
  fprops. rw H9. fprops. rw Bnat_order_le. red. eee.
Qed.

Lemma BZ_le_transitive: forall a b c,
  BZ_le a b -> BZ_le b c ->BZ_le a c.
Proof. intros a b c. wr BZ_le_pr.  wr BZ_le_pr. wr BZ_le_pr.
  cp  BZ_order. ir. order_tac.
Qed.


Lemma BZ_le_reflexive: forall a, inc a BZ -> BZ_le a a.
Proof.  ir. wr BZ_le_pr.  cp  BZ_order. order_tac. rww substrate_BZ_order.
Qed.

Lemma BZ_le_antisymmetric: forall a b,
  BZ_le a b -> BZ_le b a -> a = b.
Proof. intros a b. wr BZ_le_pr.  wr BZ_le_pr. cp  BZ_order. ir. order_tac.
Qed.

Lemma not_BZ_le_lt: forall a b, BZ_le a b -> BZ_lt b a -> False.
Proof. ir. nin H0. elim H1. app BZ_le_antisymmetric.
Qed.

Lemma BZ_lt_le_trans: forall a b c,
  BZ_lt a b -> BZ_le b c ->BZ_lt a c.
Proof. ir. split.  nin H. app (BZ_le_transitive H H0). 
  red. ir. wri H1 H0. elim (not_BZ_le_lt H0 H).
Qed.

Lemma BZ_le_lt_trans: forall a b c,
  BZ_le a b -> BZ_lt b c ->BZ_lt a c.
Proof. ir. nin H0. split. app (BZ_le_transitive H H0).  dneg. rwi H2 H.
  app  BZ_le_antisymmetric.
Qed.


Ltac BZo_tac := match goal with
  | Ha:BZ_le ?a ?b, Hb: BZ_le ?b ?c |- BZ_le ?a ?c  
     =>  ap (BZ_le_transitive Ha Hb)
  | Ha:BZ_lt ?a ?b, Hb: BZ_le ?b ?c |- BZ_lt ?a ?c  
     =>  ap (BZ_lt_le_trans Ha Hb)
  | Ha:BZ_le ?a ?b, Hb: BZ_lt ?b ?c |- BZ_lt ?a ?c  
     =>  ap (BZ_le_lt_trans Ha Hb)
  | Ha:BZ_lt ?a ?b, Hb: BZ_lt ?b ?c |- BZ_lt ?a ?c  
     =>  nin Ha; BZo_tac
  | Ha: BZ_le ?a ?b, Hb: BZ_lt ?b ?a |- _ 
    => elim (not_BZ_le_lt Ha Hb)
  | Ha:BZ_le ?x ?y, Hb:  BZ_le ?y ?x |- _ 
   => solve [ rw (BZ_le_antisymmetric Ha Hb) ; fprops ]
  | Ha: BZ_le ?a _ |- inc ?a BZ => nin Ha; am
  | Ha: BZ_le  _ ?a |- inc ?a BZ 
   => destruct Ha as [_ [Ha _ ]]; exact Ha
  | Ha: BZ_lt ?a _ |- inc ?a BZ => nin Ha; BZo_tac
  | Ha: BZ_lt  _ ?a |- inc ?a BZ => nin Ha; BZo_tac
end.

Lemma BZ_order_total: forall x y, inc x BZ -> inc y BZ -> 
  (BZ_le x y \/ BZ_le y x).
Proof. nin BZ_totalorder. ir. rwi substrate_BZ_order H0. 
  cp (H0 _ _ H1 H2). rwi BZ_le_pr H3.  ufi gge H3. rwi BZ_le_pr H3. am.
Qed.

Lemma BZ_order_total1: forall x y, inc x BZ -> inc y BZ -> 
  (BZ_le x y \/ BZ_lt y x).
Proof. ir. nin (equal_or_not x y). rw H1. left. app BZ_le_reflexive.
  nin (BZ_order_total H H0). au. right. split. am. au.
Qed.

Lemma BZ_order_total2: forall x y, inc x BZ -> inc y BZ -> 
  (x = y  \/ BZ_lt x y \/ BZ_lt y x).
Proof. ir. nin (equal_or_not x y). au. right.
  nin (BZ_order_total H H0). left. split. am.  am. right. split. am. au.
Qed.

Lemma BZ_le_pr1: forall x y, inc x BZp ->  inc y BZp -> 
  BZ_le x y = cardinal_le  (P x) (P y).
Proof. ir. cp (sub_BZp_BZ H).  cp (sub_BZp_BZ H0). cp (BZp_Q H). cp (BZp_Q H0). 
  uf BZ_le. app iff_eq. ir. ee. nin H7. ee. TP_tac. nin H7. ee. TP_tac. 
  ee. am. ir. eee. 
Qed.

Lemma BZ_le_pr2: forall x y, inc x BZp -> inc y BZms -> BZ_lt y x.
Proof. ir. red. cp (BZp_Q H). cp (BZms_Q H0). 
  split. uf BZ_le. ee; qprops. red. ir. rwi H3 H2. TP_tac.
Qed.

Lemma BZ_le_pr3: forall x y, inc x BZms -> inc y BZms ->
  BZ_le x y = cardinal_le  (P y) (P x).
Proof.  ir.  cp (BZms_Q H). cp (BZms_Q H0). uf BZ_le.
  app iff_eq. ir. ee. nin H5. ee. am. nin H5. ee. TP_tac. ee. TP_tac. 
  ir. ee; qprops.
Qed.

Lemma BZ_le_pr4: forall x y, inc x BZp -> inc y BZms -> ~ (BZ_le x y).
Proof. ir. red. ir. cp  (BZ_le_pr2 H H0).  BZo_tac.
Qed.

Lemma BZ_le_pr0: forall x y,
   BZ_le x y= 
   ((inc x BZms & inc y BZms & cardinal_le (P y) (P x))
    \/ (inc x BZms & inc y BZp)
    \/ (inc x BZp & inc y BZp & cardinal_le (P x) (P y))).
Proof. ir. ap iff_eq. ir. cp H. nin H. nin H1. rwi inc_BZ H. rwi inc_BZ H1.
  nin H; nin H1. rwii BZ_le_pr3 H0. au. au. elim (BZ_le_pr4 H H1). am.
  rwii BZ_le_pr1 H0. au. ir. nin H. ee. rww  BZ_le_pr3. nin H. ee.
  nin (BZ_le_pr2 H0 H). am. ee. rww BZ_le_pr1.
Qed.

Lemma BZ_le_pr5: forall x y, inc x BZp ->  inc y BZp -> 
  BZ_le x y = BZ_le (BZabs x) (BZabs y).
Proof. ir. rww BZ_abs_pos. rww BZ_abs_pos.
Qed.

Lemma BZ_lt_pr1: forall x y, inc x BZp ->  inc y BZp -> 
  BZ_lt x y = cardinal_lt  (P x) (P y).
Proof. ir. uf BZ_lt. rww BZ_le_pr1. uf cardinal_lt. app iff_eq.  ir. ee.
  am. dneg. wrr inc_BZp_pr. wrr (inc_BZp_pr H). ue. ir. eee. dneg. ue.
Qed.


Lemma BZ_positive_pr: forall x,  BZ_le BZ_zero x = inc x BZp.
Proof. ir. cp BZ_zero_Q. app iff_eq. ir. red in H0. ee. nin H2. ee. TP_tac.
  nin H2. ee. TP_tac. ee.
  rwi inc_BZ H1. nin H1. cp (BZms_Q H1). TP_tac. am.
  ir. red. ee; qprops. cp (BZp_Q H0). right. right.  ee. am. am. qw.
  cp (inc_BZ_P (sub_BZp_BZ H0)). fprops.
Qed.
 
Lemma BZ_positive_pr1: forall x,  BZ_lt BZ_zero x = inc x BZps.
Proof. ir. uf BZ_lt. rw BZ_positive_pr. rw inc_BZps. app iff_eq. ir. 
  nin H. au. ir. nin H. au.
Qed.

Lemma BZ_negative_pr: forall x,  BZ_lt x BZ_zero = inc x BZms.
Proof. ir. uf BZ_lt. app iff_eq. ir. ee. red in H. ee. rwi inc_BZ H. nin H. am.
  cp (BZp_Q H). nin H2. ee. TP_tac. nin H2. ee. TP_tac. ee. qwi H5.
  cp (zero_smallest2 H5). elim H0.  app BZ_zero_if_abs0. qprops.
  ir. assert (inc x BZ). qprops. cp (BZms_Q H).  cp BZ_zero_Q. 
  ee. red. ee. am. qprops.  au. red. ir. rwi H3 H1. TP_tac.
Qed.

Lemma BZ_negative_pr1: forall x,  BZ_le x BZ_zero = inc x BZm.
Proof. ir. nin (equal_or_not x BZ_zero). wr H. ap iff_eq. ir. rw H.
  uf BZm. fprops. ir. app BZ_le_reflexive. rw H. qprops.
  ap iff_eq. ir. assert (BZ_lt x BZ_zero). split; am.
  rwi  BZ_negative_pr H1. qprops. 
  ir. ufi BZm H0. rwi tack_on_inc H0. nin H0. wri BZ_negative_pr H0. nin H0. 
  am. rw H0. app BZ_le_reflexive. qprops.
Qed.

Lemma BZ_le_pr6: forall x y, inc x BZm -> inc y BZm ->
  BZ_le x y = cardinal_le  (P y) (P x).
Proof. ir. ufi BZm H0. ufi BZm H. rwi tack_on_inc H0. 
  rwi tack_on_inc H. nin H. nin H0. app BZ_le_pr3. rw H0.
  wri BZ_negative_pr H.  nin H. qw. ap iff_eq.
  ir. nin H. cp (inc_BZ_P H). fprops. ir. am. rw H. qw.
  nin H0. wri BZ_negative_pr H0. ap iff_eq. ir. BZo_tac. ir.
  cp (zero_smallest2 H1). rwi BZ_negative_pr H0. ufi BZms H0. awi H0. ee.
  ufi Bnats H3. Ztac. contradiction. rw H0. rw BZ_zero_P.
  ap iff_eq; ir. fprops. app BZ_le_reflexive. qprops.
Qed.

(** Opposite is an order isomorphism from (Z,<) to (Z,>) *)


Lemma BZ_le_opp: forall x y, BZ_le x y -> BZ_le (BZopp y) (BZopp x).
Proof. ir. assert (inc x BZ). BZo_tac. cp (BZopp_P x). rwi inc_BZ H0.
  assert (inc y BZ). BZo_tac. cp (BZopp_P y). rwi inc_BZ H2. 
  nin H0. nin H2.
  rwi (BZ_le_pr3 H0 H2) H. cp (BZopp_negative1 H0). cp (BZopp_negative1 H2).
  cp (sub_BZps_BZp H4). cp (sub_BZps_BZp H5). rww BZ_le_pr1. rw H1; rww H3.
  cp (BZopp_negative1 H0). wri BZ_positive_pr1 H4. cp (BZopp_positive2 H2).
  wri BZ_negative_pr1 H5.  nin H4. BZo_tac. 
  nin H2. red in H.  ee. ufi BZp H0. awi H0. ee. nin H5. ee. TP_tac.
  nin H5. ee. TP_tac. ee. cp (BZms_Q H2). TP_tac.
  cp (BZopp_positive2 H0).  cp (BZopp_positive2 H2).  
  rww BZ_le_pr6. rw H1; rw H3. wrr BZ_le_pr1.
Qed.

Lemma BZ_le_opp1: 
  order_isomorphism BZopp_fun BZ_ordering (opposite_order BZ_ordering).
Proof. ir. cp BZ_order. assert (order (opposite_order BZ_ordering)). fprops.
  red. uf BZopp_fun. aw. ee. am. am. app bijective_BZopp.  
  rww substrate_BZ_order. rww substrate_BZ_order. ir. aw. 
  rw BZ_le_pr. rw BZ_le_pr. ap iff_eq. ir. 
  app  BZ_le_opp. ir. wr (BZopp_involutive H1). wr (BZopp_involutive H2). 
  app  BZ_le_opp. red. ir. qprops. red. ir. qprops.
Qed.


Lemma non_zero_abs: forall b, inc b BZ -> b <> BZ_zero ->
  BZ_lt BZ_zero (BZabs b).
Proof. ir. cp (BZabs_P H). cp inc0_BZp. split. rw BZ_le_pr1. rw BZ_zero_P. 
  app zero_smallest. rw H1. cp (inc_BZ_P H). fprops. am.  app inc_BZabs_N. 
  dneg. app BZ_zero_if_abs0.  wr H1. wr H3. rww BZ_zero_P.
Qed. 

(** ** Addition *)


Definition BZplus x y:=
  let abs_sum := card_plus (P x) (P y) in
    let abs_diff1 := card_sub (P x) (P y) in
      let abs_diff2 := card_sub (P y) (P x) in
        Yo (inc x BZp & inc y BZp) (BZ_of_nat abs_sum)
        (Yo ( ~ inc x BZp & ~ inc y BZp) (BZm_of_nat abs_sum)
          (Yo (inc x BZp & ~ inc y BZp) 
            (Yo (cardinal_le (P y) (P x)) 
              (BZ_of_nat abs_diff1) (BZm_of_nat abs_diff2))
            (Yo (cardinal_le (P x) (P y)) 
              (BZ_of_nat abs_diff2) (BZm_of_nat abs_diff1)))).

Lemma BZ_plus_comm: forall x y, BZplus x y = BZplus y x.
Proof. ir. set (a:=  BZplus x y). assert (a = BZplus x y). tv.
  ufi BZplus H. uf BZplus. nin (p_or_not_p (inc y BZp & inc x BZp)). 
  rwii Y_if_rw H. rww Y_if_rw. rww card_plus_commutative. ee; am.
  rwii Y_if_not_rw H. rww Y_if_not_rw.
  nin (p_or_not_p (~ inc y BZp & ~ inc x BZp)). 
  rwii Y_if_rw H. rww Y_if_rw. rww card_plus_commutative. ee; am.
  rwii Y_if_not_rw H. rww Y_if_not_rw.
  nin (p_or_not_p (inc y BZp & ~ inc x BZp)).
  rwii Y_if_not_rw H. rww  Y_if_rw. red. ir. elim H0. eee.
  rwii Y_if_rw H. rww  Y_if_not_rw. nin (inc_or_not x BZp).  split. am.
  red. ir. elim H0. eee. nin (inc_or_not y BZp). elim H2. eee. elim H1. eee.
  dneg. eee. dneg. eee.
Qed.

Lemma BZ_plus_pp: forall x y, inc x BZp -> inc y BZp ->
  BZplus x y = BZ_of_nat (card_plus (P x) (P y)).
Proof. ir. uf BZplus. rww Y_if_rw.  split; am.
Qed.

Lemma BZ_plus_mm: forall x y, inc x BZms -> inc y BZms ->
  BZplus x y = BZm_of_nat (card_plus (P x) (P y)).
Proof. ir. uf BZplus. assert (~ inc x BZp). red. ir.
  elim (BZ_disjoint_neg_pos H H1).
  assert (~ inc y BZp). red. ir. elim (BZ_disjoint_neg_pos H0 H2).
  rw Y_if_not_rw. rw Y_if_rw. tv. au. red. ir. ee. au.
Qed.

Lemma BZ_plus_mp: forall x y, inc x BZms -> inc y BZp ->
  BZplus x y = BZplus y x.
Proof. ir.  assert (~ inc x BZp). red. ir. elim (BZ_disjoint_neg_pos H H1).
  set (a:=  BZplus x y). assert (a=BZplus x y). tv.
  ufi BZplus H2. rwii Y_if_not_rw H2. rwii Y_if_not_rw H2. rwi Y_if_not_rw H2.
  uf BZplus. rww Y_if_not_rw. rww Y_if_not_rw. rw Y_if_rw. am. split; am.
  red. ir. ee. au. red; ir. ee. au. red. ir. ee. au. red. ir. ee. au.
  red. ir. ee. au.
Qed.

Lemma BZ_plus_pm: forall x y, inc x BZp -> inc y BZms ->
  BZplus x y = (Yo (cardinal_le (P y) (P x)) 
    (BZ_of_nat (card_sub (P x) (P y)))
    (BZm_of_nat (card_sub (P y) (P x)))).
Proof. ir. assert (~ inc y BZp). red. ir. elim (BZ_disjoint_neg_pos H0 H1).
  uf BZplus. rw Y_if_not_rw. rw Y_if_not_rw. rw Y_if_rw. tv.
  split; am. red. ir. ee. au. red. ir. ee.  au.
Qed.

Lemma BZ_plus_pm1: forall x y, inc x BZp -> inc y BZms ->
  cardinal_le (P y) (P x) -> BZplus x y = BZ_of_nat(card_sub (P x) (P y)).
Proof. ir. rww BZ_plus_pm. rww Y_if_rw.
Qed. 

Lemma BZ_plus_pm2: forall x y, inc x BZp -> inc y BZms ->
  cardinal_lt (P x) (P y) -> BZplus x y = (BZm_of_nat (card_sub (P y) (P x))).
Proof. ir. rww BZ_plus_pm. rww Y_if_not_rw. red. ir.
  elim (not_card_le_lt H2 H1). 
Qed. 

Lemma BZ_stable_plus: forall x y, 
  inc x BZ -> inc y BZ -> inc (BZplus x y) BZ.
Proof. ir. uf BZplus. cp (inc_BZ_P H).  cp (inc_BZ_P H0).
  assert (inc (card_plus (P x) (P y)) Bnat). fprops. 
  nin (p_or_not_p (inc x BZp & inc y BZp)). rww Y_if_rw. app inc_BZ_of_nat.
  rww Y_if_not_rw.  nin (p_or_not_p (~ inc x BZp & ~ inc y BZp)).
  rww Y_if_rw. app inc_BZm_of_nat. rww Y_if_not_rw.  
  assert (inc (card_sub (P x) (P y)) Bnat). fprops.
  assert (inc (card_sub (P y) (P x)) Bnat). fprops.
  nin (p_or_not_p (inc x BZp & ~ inc y BZp)).  rww Y_if_rw.
  nin (p_or_not_p (cardinal_le (P y) (P x))).  rww Y_if_rw.
  app inc_BZ_of_nat. rww Y_if_not_rw. app inc_BZm_of_nat.
  rww Y_if_not_rw. nin (p_or_not_p (cardinal_le (P x) (P y))).  rww Y_if_rw.
  app inc_BZ_of_nat. rww Y_if_not_rw. app inc_BZm_of_nat.
Qed.


Lemma BZplus_0_l: forall x, 
  inc x BZ -> BZplus BZ_zero x  = x.
Proof. ir. cp (inc_BZ_P H). cp inc0_BZp. rwi inc_BZ H. nin H.  
  nin (inc_BZms_pr H). cp (BZopp_negative1 H). rww BZ_plus_pm2. rw BZ_zero_P. 
  rww minus_n_0C. rw BZ_zero_P.  split. app zero_smallest. fprops. au.
  rww BZ_plus_pp.  rw BZ_zero_P. rww zero_unit_suml. ap (inc_BZp_pr H). fprops.
Qed.

Lemma BZplus_0_r: forall x, 
  inc x BZ -> BZplus x BZ_zero = x.
Proof. ir. rww BZ_plus_comm.  ap BZplus_0_l. am. 
Qed.

Hint Rewrite BZplus_0_l BZplus_0_r: qw.

Lemma BZplus_opp_r: forall x, 
  inc x BZ -> BZplus x (BZopp x) = BZ_zero.
Proof. ir. assert (forall y, inc y BZps -> BZplus y (BZopp y) = BZ_zero).
  ir. cp (BZopp_positive1  H0). cp (sub_BZps_BZ H0). cp (sub_BZps_BZp H0).
  cp (BZopp_P y). cp (inc_BZ_P H2). rww BZ_plus_pm1. rw H4. rww minus_n_nC. 
  rw H4. fprops. cp H. rwi inc_BZ1 H. nin H.  rw H.  qw. qprops. 
  nin H. app H0. rww BZ_plus_comm. cp (BZopp_negative1 H). cp (H0 _ H2). 
  rwii BZopp_involutive H3. 
Qed.

Lemma BZplus_opp_l: forall x, 
  inc x BZ -> BZplus (BZopp x) x = BZ_zero.
Proof. ir. rww BZ_plus_comm. app BZplus_opp_r. 
Qed.

Hint Rewrite  BZplus_opp_l  BZplus_opp_r: qw.

Lemma BZ_plus_N_Ns: forall x y, inc x Bnat -> inc y Bnats ->
  inc (card_plus x y) Bnats.
Proof. ir. assert (inc y Bnat). ufi Bnats H0. Ztac. am. 
  assert (cardinal_le card_zero x).
  fprops.  assert (cardinal_lt card_zero y). split. fprops. ufi Bnats H0.
  Ztac. intuition. cp (finite_sum2_lt inc0_Bnat inc0_Bnat H H1 H2 H3).
  rwi zero_unit_suml H4. uf Bnats. Ztac. fprops. nin H4. intuition. fprops.
Qed.

Lemma BZp_stable_plus: forall x y, inc x BZp -> inc y BZp ->
  inc (BZplus x y) BZp.
Proof. ir. rww BZ_plus_pp. uf BZ_of_nat. uf BZp. aw. eee.
  cp (inc_BZ_P (sub_BZp_BZ H)).  cp (inc_BZ_P (sub_BZp_BZ H0)). fprops.
Qed.

Lemma BZps_stable_plus_r: forall x y, inc x BZp -> inc y BZps ->
  inc (BZplus x y) BZps.
Proof. ir. rwi inc_BZps H0. ee. rw inc_BZps. split. app BZp_stable_plus. 
  dneg. rwii BZ_plus_pp H2. ufi BZp H. cp H0. awi H. ee. ufi BZp H0. awi H0. 
  ee. assert (inc (P y) Bnats). uf Bnats.  Ztac. dneg. app BZ_zero_if_abs0.
  qprops. cp (BZ_plus_N_Ns H4 H8). ufi Bnats H9.  Ztac. 
  cp (f_equal P H2). rwi BZ_of_nat_P H12. rwi BZ_zero_P H12. contradiction.
Qed.

Lemma BZps_stable_plus_l: forall x y, inc x BZps -> inc y BZp ->
  inc (BZplus x y) BZps.
Proof. ir. rw BZ_plus_comm. app BZps_stable_plus_r. 
Qed.

Lemma BZps_stable_plus_rl: forall x y, inc x BZps -> inc y BZps ->
  inc (BZplus x y) BZps.
Proof. ir. app BZps_stable_plus_r. app sub_BZps_BZp.
Qed.

Lemma BZms_stable_plus_rl: forall x y, inc x BZms -> inc y BZms ->
  inc (BZplus x y) BZms.
Proof. ir. rww BZ_plus_mm. ufi BZms H. ufi BZms H0. awi H. awi H0. ee.
  uf BZm_of_nat. ufi Bnats H3. Ztac. cp (BZ_plus_N_Ns H5 H1).
  rw Y_if_not_rw. uf BZms. aw. eee. ufi Bnats H7. Ztac. am.
Qed.

Lemma BZms_stable_plus_r: forall x y, inc x BZm -> inc y BZms ->
  inc (BZplus x y) BZms.
Proof. ir. ufi BZm H. rwi tack_on_inc H. nin H. app BZms_stable_plus_rl.
  rw H. qw. qprops.
Qed.

Lemma BZms_stable_plus_l: forall x y, inc x BZms -> inc y BZm ->
  inc (BZplus x y) BZms.
Proof. ir. ufi BZm H0. rwi tack_on_inc H0. nin H0. app BZms_stable_plus_rl.
  rw H0. qw. qprops.
Qed.

Lemma BZm_stable_plus: forall x y, inc x BZm -> inc y BZm ->
  inc (BZplus x y) BZm.
Proof. ir. ufi BZm H.  rwi tack_on_inc H. nin H. cp (BZms_stable_plus_l H H0).
  app sub_BZms_BZm. rw H. qw. qprops.
Qed.

Lemma card_sub_non_zero1: forall a b, inc a Bnat -> inc b Bnat ->
  cardinal_lt a b -> card_sub b a <> card_zero.
Proof. ir. app card_sub_non_zero. srw. fprops. fprops.
Qed.

Lemma cardinal_le_total_order4: forall a b, 
  inc a Bnat  -> inc b Bnat ->
  cardinal_le a b \/ cardinal_lt b a.
Proof. ir. app cardinal_le_total_order2.  fprops. fprops.
Qed.

Lemma BZ_plus_opp_distr: forall x y, inc x BZ -> inc y BZ ->
  BZopp (BZplus x y) = BZplus (BZopp x) (BZopp y).
Proof. ir. assert (forall a b, inc a BZps -> inc b BZms -> 
  BZopp (BZplus a b) = BZplus (BZopp a) (BZopp b)). ir.
  nin (equal_or_not (P a) (P b)).
  assert (BZopp b = a). uf BZopp. ufi BZms H2. awi H2. ee. rw Y_if_rw.
  wr H3. app inc_BZp_pr. app sub_BZps_BZp. am. rw H4. rw BZplus_opp_l.
  wr H4.  rw BZplus_opp_l. qw. qprops. qprops.
  cp (BZopp_positive1 H1). cp (BZopp_negative1 H2).
  rw (BZ_plus_mp H4 (sub_BZps_BZp H5)). 
  assert (Hu:= inc_BZ_P (sub_BZms_BZ H2)).
  assert (Hv:= inc_BZ_P (sub_BZps_BZ H1)).
  cp (BZopp_P a). cp (BZopp_P b).
  nin (cardinal_le_total_order4 Hu Hv).
  rw (BZ_plus_pm1 (sub_BZps_BZp H1) H2 H8).
  assert (cardinal_lt (P b) (P a)). split. am. au.
  wri H6 H9; wri H7 H9; rw (BZ_plus_pm2 (sub_BZps_BZp H5) H4 H9). rw H6; rw H7.
  uf BZopp. uf BZ_of_nat.  aw. rww Y_if_not_rw. fprops.
  cp (card_sub_non_zero1 Hv Hu H8). 
  rw (BZ_plus_pm2 (sub_BZps_BZp H1) H2 H8). nin H8.
  wri H6 H8; wri H7 H8; rw (BZ_plus_pm1 (sub_BZps_BZp H5) H4 H8). rw H6; rw H7.
  uf BZm_of_nat. rw Y_if_not_rw. uf BZopp. aw. rww Y_if_rw. am. 
  cp H. cp H0. rwi inc_BZ1 H. nin H. rw H. qw. qprops. rwi inc_BZ1 H0. nin H0. 
  rw H0. qw. qprops.
  assert (forall a b, inc a BZps -> inc b BZps ->  BZopp (BZplus a b) = 
    BZplus (BZopp a) (BZopp b)). ir.
  rww BZ_plus_pp. rww BZ_plus_mm. rww BZopp_P. rww BZopp_P. uf BZ_of_nat.
  uf BZopp. aw. rww Y_if_not_rw. fprops. 
  ap (BZopp_positive1 H4). ap (BZopp_positive1 H5). app sub_BZps_BZp. 
  app sub_BZps_BZp. nin H; nin H0. app H4. app H1. 
  rww (BZ_plus_comm). rww H1. rww BZ_plus_comm. fprops. fprops.
  cp (BZopp_negative1 H). cp (BZopp_negative1 H0). cp (H4 _ _ H5 H6).
  rwi BZopp_involutive H7.  rwi BZopp_involutive H7. wr H7.
  rww BZopp_involutive. app BZ_stable_plus. qprops. qprops. am. am.
Qed.

Lemma BZ_plus_assoc: forall x y z,  inc x BZ -> inc y BZ -> inc z BZ ->
  BZplus x (BZplus y z) = BZplus (BZplus x y) z.
Proof. set (f:=fun x => Yo(inc x BZp) (J card_zero (P x)) (J (P x) card_zero)).
  set (g:=fun x=> Yo (cardinal_le (P x) (Q x))
    (BZ_of_nat(card_sub (Q x) (P x))) (BZm_of_nat (card_sub (P x) (Q x)))).
  assert (Ha:forall x, inc x BZ -> g (f x) = x). ir. cp (inc_BZ_P H).
  rwi inc_BZ H. uf f. nin H. rww Y_if_not_rw. uf g. aw. nin (inc_BZms_pr H). 
  rww Y_if_not_rw. rw minus_n_0C. am. am. dneg. apply zero_smallest2. am.
  red. ir. elim (BZ_disjoint_neg_pos H H1). rww Y_if_rw. uf g. aw.
  rww Y_if_rw.  rw minus_n_0C. app inc_BZp_pr. am. fprops.
  set (h:= fun x y => J (card_plus (P x) (P y)) (card_plus (Q x) (Q y))).
  (* pseudo-injectivity  of g *)
  assert (Hu:forall x y, inc x (product Bnat Bnat) -> inc y (product Bnat Bnat)
    -> (g x = g y) = (card_plus (P x) (Q y) = card_plus (P y) (Q x))).
  ir. awi H; awi H0. ee.
  nin (cardinal_le_total_order4 H3 H4). uf g. rww Y_if_rw.
  cp (card_sub_pr H4 H3 H5). set (a:= card_sub (Q x) (P x)) in *. wr H6.
  rw card_plus_associative. rww (card_plus_commutative (P y) (P x)).
  wr card_plus_associative. assert  ( (Q y = card_plus (P y) a) =
  (card_plus (P x) (Q y) = card_plus (P x) (card_plus (P y) a))).
  app iff_eq. ir. rw H7. tv. ir. apply plus_simplifiable_left with (P x).
  am. am. app Bnat_stable_plus. uf a. fprops. am. wr H7. clear H7.
  ap iff_eq. ir. nin (cardinal_le_total_order4 H1 H2). rwi Y_if_rw H7. 
  cp (card_sub_pr H2 H1 H8). wr H9. ufi BZ_of_nat H7.
  cp (f_equal P H7). awi H10. rww H10. am. rwi Y_if_not_rw H7. 
  nin H8. cp (card_sub_pr H1 H2 H8). set (b:=card_sub (P y) (Q y)) in *.
  ufi BZ_of_nat H7. ufi  BZm_of_nat H7. nin (equal_or_not b card_zero).
  rwi Y_if_rw H7. rwi H11 H10. awi H10. contradiction. fprops. am.
  rwi Y_if_not_rw H7. cp (f_equal Q H7). awi H12. TP_tac. am.
  red. ir. co_tac.
  ir. rw H7. rw Y_if_rw. rw card_plus_commutative. sy. rww card_sub_pr1.
  uf a. fprops. app sum_increasing3. fprops. uf a. fprops. uf g.
  rw Y_if_not_rw. nin H5. cp (card_sub_pr H3 H4 H5). 
  set (a:= card_sub (P x) (Q x)) in *. assert (inc a Bnat). uf a. fprops.
  assert (inc a Bnats). uf Bnats. Ztac. red. ir. rwi H9 H7. awi H7. elim H6. 
  am. fprops. wr H7. assert ((card_plus (Q y) a= P y) =
  (card_plus (card_plus (Q x) a) (Q y) = card_plus (P y) (Q x))).
  rw (card_plus_commutative (P y) (Q x)). wr card_plus_associative.
  rw card_plus_commutative. app iff_eq. ir. rw H10. tv. ir.
  apply plus_simplifiable_left with (Q x). am. fprops. am. am. wr H10. 
  clear H10. app iff_eq. ir. 
  nin (cardinal_le_total_order4 H1 H2). rwi Y_if_rw H10. 
  ufi BZm_of_nat H10. ufi  BZ_of_nat H10. nin (equal_or_not a card_zero).
  rw H12. aw. cp (card_sub_pr H2 H1 H11). rwi Y_if_rw H10. ufi BZ_zero H10.
  cp (f_equal P H10). ufi BZ_of_nat H14. awi H14. wri H14 H13. awi H13. 
  sy. am. fprops. am. fprops. rwi Y_if_not_rw H10. cp (f_equal Q H10).  
  awi H13.  TP_tac. am. am.  rwi Y_if_not_rw H10.
  assert (a = card_sub (P y) (Q y)). ufi BZm_of_nat H10.
  nin (equal_or_not a card_zero). rwi Y_if_rw H10. ufi BZ_zero  H10. 
  ufi BZ_of_nat H10. nin (equal_or_not (card_sub (P y) (Q y)) card_zero). ue. 
  rwi Y_if_not_rw H10. cp (f_equal Q H10). awi H14. TP_tac.
  rwi Y_if_not_rw H10. cp (f_equal Q H10). awi H14. TP_tac. am. am.
  rwi Y_if_not_rw H10. nin (equal_or_not (card_sub (P y) (Q y)) card_zero). 
  rwi Y_if_rw H10. ufi BZ_zero H10. ufi BZ_of_nat H10.  cp (f_equal Q H10).  
  awi H14. TP_tac. am. rwi Y_if_not_rw H10.  cp (f_equal P H10).  awi H14.
  am. am. am. rw H12. rww card_sub_pr.  nin H11. am. red. ir. co_tac.
  ir. wr H10. rw Y_if_not_rw. rw (card_plus_commutative  (Q y) a). 
  rww card_sub_pr1. red. ir. assert (cardinal_lt (Q y) (card_plus (Q y) a)). 
  assert (cardinal_lt card_zero a). split. app zero_smallest. fprops.
  ufi Bnats H9. Ztac. intuition. rw card_plus_commutative.
  cp (finite_sum3_lt (inc0_Bnat) H2 H8 H12). awi H13. am. fprops. co_tac.
  red. ir. co_tac.
  (* link with plus and the functions *)
  assert (Hb: forall x y, inc x BZp -> inc y BZms ->
    BZplus x y = g (h (f x) (f y))).
  ir. uf f. rww Y_if_rw. rww Y_if_not_rw.
  assert (is_cardinal (P x)). cp (inc_BZ_P (sub_BZp_BZ H)). fprops.
  assert (is_cardinal (P y)). cp (inc_BZ_P (sub_BZms_BZ H0)). fprops.
  nin (cardinal_le_total_order2 H2 H1). uf h. uf g. aw. rww Y_if_rw. 
  rww BZ_plus_pm1. uf h. aw. uf g. aw. rww BZ_plus_pm2. rww Y_if_not_rw. 
  red. ir. co_tac. red. ir. elim (BZ_disjoint_neg_pos H0 H1).
  assert (forall x y, inc x BZ -> inc y BZ -> BZplus x y = g (h (f x) (f y))).
  ir. rwi inc_BZ H; rwi inc_BZ H0. nin H; nin H0. uf f.
  rw Y_if_not_rw. rw Y_if_not_rw. rw BZ_plus_mm. 
  set (xy:=card_plus (P x) (P y)).
  assert (inc xy Bnats). ufi BZms H. ufi BZms H0. awi H. awi H0. ee.
  uf xy. app BZ_plus_N_Ns. ufi Bnats H3. Ztac. am.
  assert (f (BZm_of_nat xy) = J xy card_zero). uf f. rww Y_if_not_rw. 
  uf BZm_of_nat. rww Y_if_not_rw. aw.  ufi Bnats H1. Ztac. am. red. ir.
  ufi BZm_of_nat H2.  rwi Y_if_not_rw H2. ufi BZp H2. awi H2. ee. TP_tac.
  ufi Bnats H1. Ztac. am. uf h. aw. fold xy. wr H2. rww Ha. app inc_BZm_of_nat.
  ufi Bnats H1. Ztac. am. fprops. am. am. red. ir.
  elim (BZ_disjoint_neg_pos H0 H1). red. ir. elim (BZ_disjoint_neg_pos H H1).
  rw BZ_plus_comm. assert (h (f x) (f y) = h (f y) (f x)). uf h.
  rw card_plus_commutative. rww (card_plus_commutative (Q (f x)) (Q (f y))).
  rw H1. ap (Hb _ _ H0 H). rww Hb.
  uf f. rww Y_if_rw. rww Y_if_rw. uf h. aw. rw BZ_plus_pp. 
  set (xy:=card_plus (P x) (P y)). assert (inc xy Bnat).
  uf xy. cp (inc_BZ_P (sub_BZp_BZ H)). cp (inc_BZ_P (sub_BZp_BZ H0)). fprops.
  assert (inc (BZ_of_nat xy) BZp).  uf BZ_of_nat. uf BZp. aw. eee.   
  assert (f (BZ_of_nat xy) = J card_zero xy). uf f. rww Y_if_rw. 
  uf BZ_of_nat. aw. wr H3. rww Ha. qprops. am.  am. fprops.
  (* stability *)
  assert (Hf:forall x, inc x BZ -> inc (f x) (product Bnat Bnat)).
  ir. cp (inc_BZ_P H0). uf f. nin (inc_or_not x BZp).
  rww Y_if_rw. aw. eee. rww Y_if_not_rw. aw. eee. 
  assert (Hg:forall x, inc x (product Bnat Bnat) -> inc (g x) BZ). ir.
  awi H0. ee. uf g. nin (p_or_not_p (cardinal_le (P x) (Q x))).
  ir. rw Y_if_rw. app inc_BZ_of_nat. fprops. am. ir. rw Y_if_not_rw. 
  app inc_BZm_of_nat. fprops. am.
  ir. set (yz:=  BZplus y z). set (xy:= BZplus x y). cp (BZ_stable_plus H1 H2).
  cp (BZ_stable_plus H0 H1). fold xy in H4. fold yz in H3.
  rww H. rww H. set (NN:= product Bnat Bnat) in *.
  assert (Hh:forall u v, inc u NN -> inc v NN -> inc (h u v) NN). uf NN.
  uf h. ir. awi H5. awi H6. aw. ee. fprops. fprops. fprops.
  assert (yz = BZplus y z). tv. rwi (H _ _ H1 H2) H5. cp (Ha _ H3). rwi H5 H6.
  assert (g (f yz) = g (h (f y) (f z))). rww H5. 
  rwi (Hu _ _ (Hf _ H3) (Hh _ _ (Hf _ H1) (Hf _ H2))) H7. clear H6.
  ufi h H7. awi H7.
  assert (xy = BZplus x y). tv. rwi (H _ _ H0 H1) H6. cp (Ha _ H4).
  assert (g (f xy) = g (h (f x) (f y))). rww H8. 
  rwi (Hu _ _ (Hf _ H4) (Hh _ _ (Hf _ H0) (Hf _ H1))) H9. ufi h H9. awi H9.
  rw (Hu _ _ (Hh _ _ (Hf _ H0) (Hf _ H3)) (Hh _ _ (Hf _ H4) (Hf _ H2))). 
  uf h. rw pr1_pair. rw pr1_pair. rw pr2_pair. rw pr2_pair.
  set (Px:= P (f x)) in *. set (Qx:=Q (f x))  in *. 
  set (Py:= P (f y)) in *. set (Qy:=Q (f y))  in *. 
  set (Pz:= P (f z)) in *. set (Qz:=Q (f z))  in *. 
  set (Pyz:= P (f yz)) in *.  set (Qyz:= Q (f yz)) in *.
  set (Pxy:= P (f xy)) in *.  set (Qxy:= Q (f xy)) in *.
  assert (inc Px Bnat). uf Px. cp (Hf _ H0).  ufi NN H10. awi H10. ee; am.
  assert (inc Qx Bnat). uf Qx. cp (Hf _ H0).  ufi NN H11. awi H11. ee; am.
  assert (inc Py Bnat). uf Py. cp (Hf _ H1).  ufi NN H12. awi H12. ee; am.
  assert (inc Qy Bnat). uf Qy. cp (Hf _ H1).  ufi NN H13. awi H13. ee; am.
  assert (inc Pz Bnat). uf Pz. cp (Hf _ H2).  ufi NN H14. awi H14. ee; am.
  assert (inc Qz Bnat). uf Qz. cp (Hf _ H2).  ufi NN H15. awi H15. ee; am.
  assert (inc Pxy Bnat). uf Pxy. cp (Hf _ H4).  ufi NN H16. awi H16. ee; am.
  assert (inc Qxy Bnat). uf Qxy. cp (Hf _ H4).  ufi NN H17. awi H17. ee; am.
  assert (inc Pyz Bnat). uf Pyz. cp (Hf _ H3).  ufi NN H18. awi H18. ee; am.
  assert (inc Qyz Bnat). uf Qyz. cp (Hf _ H3).  ufi NN H19. awi H19. ee; am.
  apply plus_simplifiable_right with Qy. am. fprops. fprops.
  wr card_plus_associative. wr card_plus_associative.
  set (a:=(card_plus Pyz (card_plus (card_plus Qxy Qz) Qy))).
  assert (a=card_plus Pyz (card_plus (card_plus Qxy Qz) Qy)). tv.
  rwi card_plus_associative H20. rwi card_plus_associative H20.
  rwi (card_plus_commutative Pyz Qxy) H20.  wri card_plus_associative H20. 
  wri card_plus_associative H20. rwi (card_plus_commutative Qy Qz) H7.
  rwi H7 H20. rw H20. rw (card_plus_commutative Pxy Pz).
  wr (card_plus_associative  Py Pz Qyz). set (b:=card_plus Pz Qyz).
  rw (card_plus_associative  Qxy Py b). rw (card_plus_commutative  Qxy Py). 
  wr (card_plus_associative  Py Qxy b). rw card_plus_associative.
  rw card_plus_associative. wr H9. rw (card_plus_associative  Pxy Qx Qy).  
  rw (card_plus_associative (card_plus Pz Pxy) Qx Qyz).
  wr (card_plus_associative Pz Pxy Qx). set (c:= card_plus Pxy Qx). 
  rw (card_plus_commutative Pz c). wr (card_plus_associative c Pz Qyz).
  fold b. wr card_plus_associative.  wr card_plus_associative.
  rw (card_plus_commutative b Qy). tv.
Qed.


(** ** Multiplication *)

Definition BZmult x y :=
  let aux := BZ_of_nat (card_mult (P x) (P y)) in
    (Yo (inc x BZp & inc y BZp) aux
      (Yo (inc x BZms & inc y BZms) aux (BZopp aux))).

Definition BZ_mult_sign_aux x y:=  
  Yo (x = BZ_zero) TPb (Yo (y= BZ_zero) TPb (Yo (Q x = Q y) TPb TPa)).

Lemma BZ_mult_aux: forall x y, inc x BZ -> inc y BZ -> 
  inc (card_mult (P x) (P y)) Bnat.
Proof. ir. cp (inc_BZ_P H). cp (inc_BZ_P H0). fprops.
Qed.


Lemma BZmult_pm: forall x y, inc x BZp -> inc y BZms-> 
  BZmult x y = BZm_of_nat (card_mult (P x) (P y)).
Proof. ir. uf BZmult. rw Y_if_not_rw. rw Y_if_not_rw. rww BZopp_pr1.
  cp (inc_BZ_P (sub_BZp_BZ H)).  cp (inc_BZ_P (sub_BZms_BZ H0)). fprops.
  red. ir. ee. elim (BZ_disjoint_neg_pos H1 H). 
  red. ir. ee. elim (BZ_disjoint_neg_pos H0 H2). 
Qed.
 
Lemma BZmult_mp: forall x y, inc x BZms -> inc y BZp -> 
  BZmult x y = BZm_of_nat (card_mult (P x) (P y)).
Proof. ir. uf BZmult. rw Y_if_not_rw. rw Y_if_not_rw. rww BZopp_pr1.
  cp (inc_BZ_P (sub_BZp_BZ H0)).  cp (inc_BZ_P (sub_BZms_BZ H)). fprops.
  red. ir. ee. elim (BZ_disjoint_neg_pos H2 H0). 
  red. ir. ee. elim (BZ_disjoint_neg_pos H H1). 
Qed.

Lemma BZmult_mm: forall x y, inc x BZms -> inc y BZms-> 
  BZmult x y = BZ_of_nat (card_mult (P x) (P y)).
Proof. ir. uf BZmult. rw Y_if_not_rw. rww Y_if_rw. au. 
  red. ir. ee. elim (BZ_disjoint_neg_pos H H1).
Qed.

Lemma BZmult_pp: forall x y, inc x BZp -> inc y BZp -> 
  BZmult x y = BZ_of_nat (card_mult (P x) (P y)).
Proof. ir. uf BZmult. rww Y_if_rw. eee.
Qed.

Lemma BZ_stable_mult: forall x y, inc x BZ -> inc y BZ ->  inc (BZmult x y) BZ.
Proof. ir. cp (BZ_mult_aux H H0). uf BZmult.
  set (xy := (card_mult (P x) (P y)))  in *. cp (inc_BZ_of_nat H1).
  nin (p_or_not_p (inc x BZp & inc y BZp)). rww Y_if_rw. 
  rww Y_if_not_rw. nin (p_or_not_p (inc x BZms & inc y BZms)).
  rww Y_if_rw. rww Y_if_not_rw. qprops.
Qed.


Lemma BZ_mult_abs1: forall x y, inc x BZ -> inc y BZ ->  
  P (BZmult x y) = (card_mult (P x) (P y)).
Proof. ir. uf BZmult. set (xy := (card_mult (P x) (P y))).
  assert (P (BZ_of_nat xy) = xy). uf BZ_of_nat. aw. 
  nin (p_or_not_p (inc x BZp & inc y BZp)). rww Y_if_rw. 
  rww Y_if_not_rw. nin (p_or_not_p (inc x BZms & inc y BZms)).
  rww Y_if_rw. rww Y_if_not_rw. rww BZopp_P. 
Qed.

Lemma BZ_mult_comm: forall x y, 
  BZmult x y = BZmult y x.
Proof. ir. uf  BZmult. nin (p_or_not_p (inc x BZp & inc y BZp)).
  rw Y_if_rw. rw Y_if_rw. rww card_mult_commutative. eee. eee.
  rww Y_if_not_rw.  nin (p_or_not_p (inc x BZms & inc y BZms)).
  rw Y_if_rw. rww Y_if_not_rw.  rw Y_if_rw.  rww card_mult_commutative.
  eee.  dneg. eee. eee. rww Y_if_not_rw.  rww Y_if_not_rw.  rww Y_if_not_rw. 
  rww card_mult_commutative. dneg. eee. clear H0. dneg. eee.
Qed.

Lemma BZmult_0_l: forall x, inc x BZ -> BZmult BZ_zero x = BZ_zero.
Proof. ir. assert (card_mult (P BZ_zero) (P x) = card_zero).
  rw BZ_zero_P. cp (inc_BZ_P H). rw card_mult_commutative.
  app zero_prod_absorbing. 
  assert (BZ_of_nat (card_mult (P BZ_zero) (P x)) = BZ_zero). rw H0. tv.
  uf BZmult. nin (p_or_not_p (inc BZ_zero BZp & inc x BZp)).
  rww Y_if_rw. rww Y_if_not_rw.
  nin (p_or_not_p (inc BZ_zero BZms & inc x BZms)).  
  rww Y_if_rw. rww Y_if_not_rw. rw H1. qw.
Qed.

Lemma BZmult_0_r: forall x, inc x BZ -> BZmult x BZ_zero = BZ_zero.
Proof. ir. rww BZ_mult_comm.  app  BZmult_0_l. 
Qed.

Hint Rewrite  BZmult_0_l  BZmult_0_r: qw.

Lemma  BZmult_0_rl: forall x y, inc x BZ -> inc y BZ ->
  x <> BZ_zero -> y <> BZ_zero -> BZmult x y <> BZ_zero.
Proof. ir. red. ir. cp (f_equal P H3). rwii  BZ_mult_abs1 H4. 
  rwi BZ_zero_P H4. assert (P x <> card_zero). clear H2. dneg.
  app BZ_zero_if_abs0.  assert (P y <> card_zero). clear H5. dneg.
  app BZ_zero_if_abs0. elim (zero_cardinal_product2 H5 H6).  am.
Qed.

Lemma BZ_mult_abs2: forall x y, inc x BZ -> inc y BZ ->  
  BZmult x y = J (card_mult (P x) (P y)) (BZ_mult_sign_aux x y).
Proof. ir. cp (BZ_mult_abs1 H H0). 
  nin (equal_or_not x BZ_zero). cp ( BZmult_0_l H0). 
  uf BZ_mult_sign_aux. rww Y_if_rw. rw H2. rw H3.
  app pair_extensionality. uf BZ_zero. uf BZ_of_nat. fprops. fprops.
  rwi H2 H1. wr H1. rw H3. aw. aw. uf BZ_zero. uf BZ_of_nat. aw.
  nin (equal_or_not y BZ_zero). cp (BZmult_0_r H). 
  uf BZ_mult_sign_aux. rww Y_if_not_rw. rww Y_if_rw. rw H3. rw H4.
  app pair_extensionality. uf BZ_zero. uf BZ_of_nat. fprops. fprops.
  rwi H3 H1. wr H1. rw H4. aw. aw. qw. uf BZ_zero. uf BZ_of_nat. aw.
  uf BZ_mult_sign_aux.  rww Y_if_not_rw.  rww Y_if_not_rw.
  set (Ha:=BZ_stable_mult H H0).
  app pair_extensionality. 
  ufi BZ Ha. rwi canonical_du2_pr Ha. ee; am. fprops. rw H1. aw. aw.
  uf BZmult. nin (p_or_not_p (inc x BZp & inc y BZp)). 
  rw Y_if_rw. nin H4. uf BZ_of_nat. aw. ufi BZp H4. ufi BZp H5. awi H4. awi H5.
  ee. rww Y_if_rw. ue. am. rww Y_if_not_rw. 
  nin (p_or_not_p (inc x BZms & inc y BZms)). rww Y_if_rw. 
  uf BZ_of_nat. aw.  rww Y_if_rw. nin H5. ufi BZms H5. ufi BZms H6. 
  awi H5. awi H6. ee. ue.  cp (BZmult_0_rl H H0 H2 H3).
  assert (card_mult (P x) (P y) <> card_zero). dneg.
  app BZ_zero_if_abs0. rw H1. am.
  rww Y_if_not_rw.  rww Y_if_not_rw.
  uf BZopp. rw Y_if_not_rw. uf BZm_of_nat. rw Y_if_not_rw.  aw.
  uf BZ_of_nat. aw. uf BZ_of_nat. aw. fprops.
  clear Ha. rwi inc_BZ H. rwi inc_BZ H0. nin H. cp H. ufi BZms H. awi H. ee.
  rw H10.
  nin H0. elim H5. au. ufi BZp H0. awi H0. ee. rw H12. fprops.
  nin H0. ufi BZp H. awi H. ee.  ufi BZms H0. awi H0. ee.
  rw H9. rw H11. fprops. elim H4. au.
Qed.

Lemma BZ_mult_abs: forall x y, inc x BZ -> inc y BZ ->  
  BZabs (BZmult x y) = BZmult (BZabs x) (BZabs y).
Proof. ir. sy. rw BZmult_pp. rww BZ_mult_abs2. repeat rww BZabs_P. 
  uf BZabs. aw. app inc_BZabs_N.  app inc_BZabs_N.
Qed.

Lemma BZmult_1_l: forall x, inc x BZ -> BZmult BZ_one x = x.
Proof. ir. assert (is_cardinal (P x)). cp (inc_BZ_P H). fprops.
  assert (P BZ_one = card_one). uf BZ_one. app BZ_of_nat_P.
  assert (inc BZ_one BZp). uf BZ_one.  app inc_BZ_of_natp. fprops.
  rwi inc_BZ H. nin H. rww BZmult_pm. rw H1. aw. 
  nin (inc_BZms_pr H). am.  rww BZmult_pp. rw H1. aw. app (inc_BZp_pr H).
Qed.

Lemma BZmult_1_r: forall x, inc x BZ -> BZmult x BZ_one = x.
Proof. ir. rww BZ_mult_comm. ap BZmult_1_l. am. 
Qed.

Hint Rewrite  BZmult_1_r BZmult_1_l: qw.

Lemma inc_mone_BZms: inc BZ_mone BZms.
Proof. assert (card_one <> card_zero). cp card_one_not_zero. au.
  uf BZ_mone. uf BZms. uf BZm_of_nat. rw Y_if_not_rw. aw. eee.
  uf Bnats. Ztac. fprops. am.
Qed.

Lemma BZ_opp_mone: BZopp BZ_mone = BZ_one.
Proof. uf BZ_mone. rww BZopp_pr2. fprops.
Qed.

Lemma BZ_opp_one: BZopp BZ_one = BZ_mone.
Proof. wr BZ_opp_mone.  rww BZopp_involutive. qprops.
Qed.

Hint Resolve inc_mone_BZms inc_m1_BZ:qprops.

Lemma BZ_mult_m1_r: forall x, inc x BZ ->  BZmult x BZ_mone = BZopp x.
Proof. ir. assert (is_cardinal (P x)). cp (inc_BZ_P H). fprops.
  assert (P BZ_mone = card_one). uf BZ_mone. rww BZm_of_nat_P.
  cp inc_mone_BZms. rwi inc_BZ H. nin H. rww BZmult_mm. rw H1. aw. 
  uf BZopp. rw Y_if_rw. tv. ufi BZms H. awi H. ee. am.
  rww BZmult_pm. rw H1. aw.  uf BZopp.  rw Y_if_not_rw. tv.
  ufi BZp H. awi H. ee. rw H4. fprops.
Qed.

Lemma BZ_mult_m1_l: forall x, inc x BZ ->  BZmult BZ_mone x = BZopp x.
Proof. ir. rw BZ_mult_comm. rww  BZ_mult_m1_r.
Qed.

Lemma BZopp_mult_distrib_r: forall x y, inc x BZ -> inc y BZ ->
  BZopp (BZmult x y) = BZmult x (BZopp y).
Proof. ir. nin (equal_or_not y BZ_zero). rw H1. qw. 
  cp (BZ_mult_aux H H0). rwi inc_BZ H. rwi inc_BZ H0. nin H; nin H0.
  rww BZmult_mm. rww BZmult_mp. rw BZopp_P. rww BZopp_pr1. 
  app BZopp_negative2. app (sub_BZms_BZm).
  rww BZmult_mp. rww BZmult_mm. rw BZopp_P. rww BZopp_pr2. 
  app BZopp_positive1. rww inc_BZps. au.
  rww BZmult_pm. rww BZmult_pp.  rw BZopp_P. rww BZopp_pr2. 
  app BZopp_negative2. app (sub_BZms_BZm).
  rww BZmult_pp. rww BZmult_pm. rw BZopp_P. rww BZopp_pr1. 
  app BZopp_positive1.  rww inc_BZps. au.
Qed.

Lemma BZopp_mult_distrib_l: forall x y, inc x BZ -> inc y BZ ->
  BZopp (BZmult x y) = BZmult (BZopp x) y.
Proof. ir. rww BZ_mult_comm. rww BZopp_mult_distrib_r. rww BZ_mult_comm. 
Qed.

Lemma BZmult_opp_comm: forall x y, inc x BZ -> inc y BZ ->
  BZmult x (BZopp y) = BZmult (BZopp x) y.
Proof. ir. wrr BZopp_mult_distrib_l. rww BZopp_mult_distrib_r.
Qed.

Lemma BZmult_opp_opp: forall x y, inc x BZ -> inc y BZ ->
  BZmult (BZopp x) (BZopp y) = BZmult x y.
Proof. ir. rw (BZmult_opp_comm (BZ_stable_opp H) H0). rww BZopp_involutive.
Qed.


Lemma BZ_mult_assoc: forall x y z,  inc x BZ -> inc y BZ -> inc z BZ ->
  BZmult x (BZmult y z) = BZmult (BZmult x y) z.
Proof. ir. 
  assert (Ht: forall x y, inc x BZ -> inc y BZ ->
  BZmult (BZopp x) y = BZopp (BZmult x y)). ir. rww BZ_mult_comm.
  wr BZopp_mult_distrib_r. rww BZ_mult_comm. am. am. fprops.
  assert (Ha:forall x y z,  inc x BZp -> inc y BZp-> inc z BZp ->
  BZmult x (BZmult y z) = BZmult (BZmult x y) z). 
  ir. rw (BZmult_pp H2 H3). rw (BZmult_pp H3 H4). 
  cp (BZ_mult_aux (sub_BZp_BZ H3) (sub_BZp_BZ H4)).  
  rw (BZmult_pp H2 (inc_BZ_of_natp H5)).
  cp (BZ_mult_aux (sub_BZp_BZ H2) (sub_BZp_BZ H3)).  
  rw (BZmult_pp  (inc_BZ_of_natp H6) H4). rw BZ_of_nat_P. rw BZ_of_nat_P.
  rww card_mult_associative.
  assert (Hb:forall x y z,  inc x BZp -> inc y BZp -> inc z BZ ->
  BZmult x (BZmult y z) = BZmult (BZmult x y) z). ir. rwi inc_BZ H4. nin H4. 
  cp (BZopp_negative1 H4). cp (sub_BZps_BZp H5). cp (Ha _ _ _ H2 H3 H6). 
  assert (inc x0 BZ). qprops. assert (inc y0 BZ). qprops.
  assert (inc z0 BZ). qprops.  set (BZ_stable_mult H9 H10).
  set (BZ_stable_mult H8 H9). do 3 wrii BZopp_mult_distrib_r H7.
  app BZopp_inj; qprops. app BZ_stable_mult.  app BZ_stable_mult. app Ha.
  assert (Hc:forall x y z,  inc x BZp -> inc y BZ -> inc z BZ ->
    BZmult x (BZmult y z) = BZmult (BZmult x y) z). ir. rwi inc_BZ H3. nin H3. 
  cp (BZopp_negative1 H3). cp (sub_BZps_BZp H5).
  assert (inc x0 BZ).  qprops. assert (inc y0 BZ). qprops.
  set (BZ_stable_mult H8 H4).  set (BZ_stable_mult H7 H8).
  cp (Hb _ _ _ H2 H6 H4). rwii Ht H9; qprops.
  do 2 wrii BZopp_mult_distrib_r H9. rwii Ht H9. 
  app BZopp_inj. app BZ_stable_mult. app BZ_stable_mult. app Hb.
  rwi inc_BZ H. nin H.  cp (BZopp_negative1 H). cp (sub_BZps_BZp H2).
  assert (inc x BZ). qprops. cp (Hc _ _ _ H3 H0 H1). 
  set (BZ_stable_mult H4 H0).  set (BZ_stable_mult H0 H1).
  do 3 rwii Ht H5.  app BZopp_inj. app BZ_stable_mult. app BZ_stable_mult. 
  app Hc.
Qed.

Lemma BZmult_plus_distr_r : forall n m p,
  inc n BZ -> inc m BZ -> inc p BZ  ->
  BZmult n  (BZplus m  p) = BZplus (BZmult n m)  (BZmult n  p).
Proof. assert (forall n m p,   inc n BZps -> inc m BZp -> inc p BZ  ->
  BZmult n  (BZplus m  p) = BZplus (BZmult n m)  (BZmult n  p)).
  ir. cp (sub_BZps_BZp H). rwi inc_BZ H1. nin H1.
  cp (BZopp_negative1 H1). assert (Ha:=inc_BZ_P (sub_BZp_BZ H0)).
  assert (Hb:=inc_BZ_P (sub_BZms_BZ H1)). assert (Hc:=inc_BZ_P (sub_BZp_BZ H2)).
  assert (is_cardinal (P m)). fprops. assert (is_cardinal (P p)). fprops.
  assert(Hd: P n <> card_zero).  ufi BZps H. awi H. ee. ufi Bnats H6. Ztac. am.
  nin (cardinal_le_total_order2 H5 H4). rw (BZ_plus_pm1 H0 H1 H6).
  cp (card_sub_pr Ha Hb H6). set (mp:=card_sub (P m) (P p)) in *.
  rw (BZmult_pp H2 H0). rw (BZmult_pm H2 H1). 
  assert (cardinal_le (card_mult (P n) (P p)) (card_mult (P n) (P m))).
  app product_increasing2.  fprops.
  set (u:= card_mult (P n) (P p)) in *. assert (inc u Bnat). uf u. fprops.
  assert (inc u Bnats). uf u. uf Bnats. Ztac. app zero_cardinal_product2. 
  ufi BZms H1. awi H1. ee. ufi Bnats H10. Ztac. am.
  assert (inc (BZm_of_nat u) BZms). ufi Bnats H10. Ztac. app inc_BZm_of_natms.
  cp (card_sub_pr (Bnat_stable_mult Hc Ha) H9 H8). 
  wri (BZm_of_nat_P u) H8.  wri (BZ_of_nat_P (card_mult (P n) (P m))) H8.
  rw (BZ_plus_pm1 (inc_BZ_of_natp (Bnat_stable_mult Hc Ha)) H11 H8). 
  rw BZm_of_nat_P. rw BZ_of_nat_P.  
  assert (inc mp Bnat). uf mp. app Bnat_stable_sub. 
  rw (BZmult_pp H2 (inc_BZ_of_natp H13)). rw BZ_of_nat_P. uf mp. app uneq.
  apply plus_simplifiable_left with u.  am. fprops.
  app Bnat_stable_sub. fprops. rw H12. uf u. wr cardinal_distrib_prod_sum3. 
  fold mp. rww H7. 
  rw (BZ_plus_pm2 H0 H1 H6). nin H6.
  cp (card_sub_pr Hb Ha H6). set (mp:=card_sub (P p) (P m)) in *.
  rw (BZmult_pp H2 H0). rw (BZmult_pm H2 H1). 
  assert (cardinal_lt (card_mult (P n) (P m)) (card_mult (P n) (P p))).
  app finite_prod2_lt.  fprops. split;am.  
  set (u:= card_mult (P n) (P p)) in *. assert (inc u Bnat). uf u. fprops.
  assert (inc u Bnats). uf Bnats.  Ztac. uf u. app zero_cardinal_product2. 
  ufi BZms H1. awi H1. ee. ufi Bnats H11. Ztac. am.
  assert (inc (BZm_of_nat u) BZms). ufi Bnats H11. Ztac. app inc_BZm_of_natms.
  cp H9. nin H9. cp (card_sub_pr H10 (Bnat_stable_mult Hc Ha) H9). 
  wri (BZm_of_nat_P u) H13. wri (BZ_of_nat_P (card_mult (P n) (P m))) H13.
  rw (BZ_plus_pm2 (inc_BZ_of_natp (Bnat_stable_mult Hc Ha)) H12 H13). 
  rw BZm_of_nat_P. rw BZ_of_nat_P. rw BZmult_pm. app uneq.
  apply plus_simplifiable_left with (card_mult (P n) (P m)). fprops. 
  rw BZm_of_nat_P. assert (inc mp Bnat). uf mp. fprops. fprops. fprops.
  rw BZm_of_nat_P.  wr cardinal_distrib_prod_sum3.  rww H15. uf mp.
  rww card_sub_pr. am. app  inc_BZm_of_natms. uf mp. fprops. uf mp.
  cp (card_sub_pr Hb Ha H6). red. ir. rwi H17 H16. awi H16. elim H7. am.
  fprops.
  rw (BZ_plus_pp H0 H1). rw (BZmult_pp H2 H0). rw (BZmult_pp H2 H1).  
  rw BZmult_pp. rw BZ_plus_pp. rw BZ_of_nat_P.  rw BZ_of_nat_P.  rw BZ_of_nat_P.
  rww cardinal_distrib_prod_sum3. app inc_BZ_of_natp. app BZ_mult_aux.
  qprops. qprops. app inc_BZ_of_natp. app BZ_mult_aux. qprops. qprops. am.
  app inc_BZ_of_natp. cp (inc_BZ_P (sub_BZp_BZ H0)). 
  cp (inc_BZ_P (sub_BZp_BZ H1)). fprops.
  assert (forall n m p, inc n BZps -> inc m BZ -> inc p BZ  ->
  BZmult n  (BZplus m  p) = BZplus (BZmult n m)  (BZmult n  p)). ir.
  ir. rwi inc_BZ H1. nin H1. 
  cp (BZ_stable_opp H2). cp (BZopp_negative1 H1). cp (sub_BZps_BZp H4). 
  cp (H _ _ _ H0 H5 H3). assert (inc n BZ). qprops. assert (inc m BZ). qprops.
  wrii BZ_plus_opp_distr H6. cp (BZ_stable_plus H8 H2).
  cp (BZ_stable_mult H7 H2).  cp (BZ_stable_mult H7 H8).
  do 3 wrii BZopp_mult_distrib_r H6. wrii BZ_plus_opp_distr H6. 
  app BZopp_inj. app BZ_stable_mult.  app BZ_stable_plus. app H.
  ir. rwi inc_BZ1 H1. nin H1. rw H1. qw. qprops.  app BZ_stable_plus.
  nin H1. app H0.
  assert (forall x y, inc x BZ -> inc y BZ ->
    BZmult (BZopp y) x = BZopp (BZmult y x)). ir. rww BZ_mult_comm.
  rww (BZ_mult_comm y x). sy. app BZopp_mult_distrib_r. 
  assert (inc n BZ). qprops.  cp (BZ_stable_plus H2 H3).
  cp (BZopp_negative1 H1). cp (H0 _ _ _ H7 H2 H3). 
  do 3 (rwii H4 H8). cp (BZ_stable_mult H5 H2).  cp (BZ_stable_mult H5 H3).
  wrii BZ_plus_opp_distr H8. 
  app BZopp_inj. app BZ_stable_mult.  app BZ_stable_plus. 
Qed.

Lemma BZmult_plus_distr_l : forall n m p,
  inc n BZ -> inc m BZ -> inc p BZ  ->
  BZmult (BZplus m  p) n = BZplus (BZmult m n)  (BZmult p n).
Proof. ir. rw (BZ_mult_comm m n). rw (BZ_mult_comm p n).
  wrr BZmult_plus_distr_r. app BZ_mult_comm. 
Qed.


Definition BZminus x y := BZplus x (BZopp y).
Definition BZsucc x  := BZplus x BZ_one.
Definition BZpred x  := BZminus x BZ_one.


Lemma BZ_stable_minus: forall x y, 
  inc x BZ -> inc y BZ -> inc (BZminus x y) BZ.
Proof. ir. uf BZminus. app BZ_stable_plus. qprops. Qed.

Lemma BZ_stable_succ: forall x, inc x BZ -> inc (BZsucc x) BZ.
Proof. ir. uf BZsucc. app BZ_stable_plus. qprops. Qed.

Lemma BZ_stable_pred: forall x, inc x BZ -> inc (BZpred x) BZ.
Proof. ir. uf BZpred. app BZ_stable_minus. qprops. Qed.


Lemma BZ_succ_pred: forall x, inc x BZ -> BZsucc (BZpred x) = x.
Proof. ir. uf BZsucc. uf BZpred. uf BZminus. wr BZ_plus_assoc.
  qw. qprops. am. qprops. qprops.
Qed.

Lemma BZ_pred_succ: forall x, inc x BZ -> BZpred (BZsucc x) = x.
Proof. ir. uf BZsucc. uf BZpred. uf BZminus. wr BZ_plus_assoc.
  qw. qprops. am. qprops. qprops.
Qed.

Lemma BZ_succ_N: forall x, inc x Bnat ->
  BZsucc (BZ_of_nat x) = BZ_of_nat (succ x).
Proof. ir. uf BZsucc.  rww BZ_plus_pp. uf BZ_one. rww BZ_of_nat_P. 
  rww BZ_of_nat_P. app inc_BZ_of_natp. uf BZ_one. app inc_BZ_of_natp. fprops.
Qed.

Lemma BZ_prec_N: forall x, inc x Bnats ->
  BZpred (BZ_of_nat x) = BZ_of_nat (predc x).
Proof. ir. ufi Bnats H.  Ztac. assert (is_finite_c x). wrr inc_Bnat. 
  cp (predc_pr H2 H1). ee. 
  assert (inc (predc x) Bnat). rww inc_Bnat. set (y:= predc x).
  rw H4. wrr (BZ_succ_N H5). rww BZ_pred_succ. app inc_BZ_of_nat.
Qed.

Lemma BZminus_plus: forall n m, inc n BZ -> inc m BZ ->
  BZminus (BZplus n m) n = m.
Proof. ir. uf BZminus. cp (BZ_stable_opp H). rww (BZ_plus_comm n m). 
  wrr BZ_plus_assoc. qw.
Qed.

Lemma BZplus_minus: forall n m, inc n BZ -> inc m BZ ->
  BZplus n (BZminus m n) = m.
Proof. ir. uf BZminus. cp (BZ_stable_opp H). rww (BZ_plus_comm m (BZopp n)). 
  rww BZ_plus_assoc. qw.
Qed.

Lemma BZplus_minus_ea: forall n m p, inc n BZ -> inc m BZ ->inc p BZ ->
  n = BZplus m p -> p = BZminus n m.
Proof. ir. rw H2. sy. app BZminus_plus.
Qed.

Lemma BZminus_diag: forall a, inc a BZ -> BZminus a a = BZ_zero.
Proof. ir. uf BZminus. qw.
Qed.

Lemma BZ_minus_diag_rw: forall a b, inc a BZ -> inc b BZ ->
  BZminus a b = BZ_zero -> a = b.
Proof. ir.  cp (f_equal (BZplus b) H1). rwi BZplus_minus H2. qwi H2.
  am. am. am. am.
Qed.

Lemma BZminus_0_r: forall a, inc a BZ -> BZminus a BZ_zero = a.
Proof. ir. uf BZminus. qw.
Qed.

Lemma BZminus_0_l: forall a, inc a BZ -> BZminus BZ_zero a = BZopp a.
Proof. ir. uf BZminus. qw. qprops.
Qed.

Lemma BZminus_plus_simpl_l:  forall n m p, inc n BZ -> inc m BZ ->inc p BZ ->
  BZminus (BZplus p n) (BZplus p m) = BZminus n m.
Proof. ir. uf BZminus. rww BZ_plus_opp_distr. rw (BZ_plus_comm p n).
  cp (BZ_stable_opp H1).
  rww BZ_plus_assoc. wrr (BZ_plus_assoc H H1 H2). qw. app BZ_stable_plus.
  qprops.
Qed.

Lemma BZminus_plus_simpl_r:  forall n m p, inc n BZ -> inc m BZ ->inc p BZ ->
  BZminus (BZplus n p) (BZplus m p) = BZminus n m.
Proof. ir.  rw (BZ_plus_comm n p).  rw (BZ_plus_comm m p).
  app  BZminus_plus_simpl_l.
Qed.

Lemma BZminus_plus_comm: forall n m p, inc n BZ -> inc m BZ ->inc p BZ ->
  BZminus (BZplus n m) p = BZplus (BZminus n p) m.
Proof. ir. uf BZminus. rw (BZ_plus_comm n m). cp (BZ_stable_opp H1).
  wrr BZ_plus_assoc. rww BZ_plus_comm. 
Qed.

Lemma BZminus_succ_l:  forall n m, inc n BZ -> inc m BZ ->
  BZsucc (BZminus n m) = BZminus (BZsucc n) m.
Proof. ir. uf BZsucc. sy. app BZminus_plus_comm. qprops.
Qed.

Lemma BZsucc_discr: forall a, inc a BZ -> a <> BZsucc a.
Proof. ir. assert (inc BZ_one BZ). qprops.
  red. uf BZsucc. ir. cp (BZminus_plus H H0). wri H1 H2.
  ufi BZminus H2. qwi H2. symmetry in H2. ufi BZ_one H2. ufi BZ_zero H2.
  cp (f_equal P H2). rwi BZ_of_nat_P H3.  rwi BZ_of_nat_P H3.
  elim (card_one_not_zero H3). am.
Qed.

Lemma BZplus_reg_r : forall n m p, inc n BZ -> inc m BZ -> inc p BZ ->
  BZplus m n = BZplus p n -> m = p.
Proof. ir. wr (BZminus_plus H H0). rww BZ_plus_comm. rww H2. rww BZ_plus_comm. 
  app BZminus_plus.
Qed.

Lemma BZplus_reg_l : forall n m p, inc n BZ -> inc m BZ -> inc p BZ ->
  BZplus n m = BZplus n p -> m = p.
Proof. ir. rwi BZ_plus_comm H2. rwi (BZ_plus_comm n p) H2.  
  app (BZplus_reg_r H H0 H1 H2).
Qed.

Lemma BZmult_minus_distr_r : forall n m p,
  inc n BZ -> inc m BZ -> inc p BZ  ->
  BZmult n  (BZminus m  p) = BZminus (BZmult n m)  (BZmult n  p).
Proof. uf BZminus. ir. rww  BZmult_plus_distr_r. wrr BZopp_mult_distrib_r. 
  qprops.
Qed.

Lemma BZmult_minus_distr_l : forall n m p,
  inc n BZ -> inc m BZ -> inc p BZ  ->
  BZmult (BZminus m p) n = BZminus (BZmult m n)  (BZmult p n).
Proof. ir. rww (BZ_mult_comm m n). rww (BZ_mult_comm p n).
  wrr BZmult_minus_distr_r. app BZ_mult_comm. 
Qed.

Lemma BZ_minus_inj: forall a b, inc a BZ -> inc b BZ ->
  BZminus a b = BZ_zero -> a = b.
Proof. ir. cp (BZplus_minus H0 H). rwi H1 H2. qwi H2. sy; am. am.
Qed.

Lemma BZmult_reg_r : forall n m p,
  inc n BZ -> inc m BZ -> inc p BZ  ->  n <> BZ_zero ->
  BZmult m n = BZmult p n -> m=p.
Proof. ir. cp (BZmult_minus_distr_l H H0 H1). rwi H3 H4. rwi BZminus_diag H4.
  nin (equal_or_not  (BZminus m p) BZ_zero). app BZ_minus_inj.
  cp (BZmult_0_rl (BZ_stable_minus H0 H1) H H5 H2). contradiction. 
  app BZ_stable_mult.
Qed.

Lemma BZmult_reg_l : forall n m p,
  inc n BZ -> inc m BZ -> inc p BZ  -> n <> BZ_zero ->
  BZmult n m = BZmult n p -> m=p.
Proof. ir. rwi (BZ_mult_comm n m) H3. rwi (BZ_mult_comm n p) H3.
  ap  (BZmult_reg_r H H0 H1 H2 H3). 
Qed.


Lemma BZmult_1_inversion_l: forall a b, inc a BZ -> inc b BZ ->
  BZmult a b = BZ_one -> (a=b & (a = BZ_one  \/ a = BZ_mone)).
Proof. assert (HH:card_zero = card_one ->False). ir.
  symmetry in H. elim (card_one_not_zero H).
  assert (forall a b, inc a Bnats -> inc b Bnat ->
  card_mult a b = card_one -> (a = b & a = card_one)).
  ir. ufi Bnats H. nin (Z_all H).  nin (equal_or_not b card_zero). rwi H4 H1. 
  rwi zero_prod_absorbing H1. elim (HH H1).
  assert (cardinal_le card_one a). app one_small_cardinal. fprops.
  assert (cardinal_le card_one b). app one_small_cardinal. fprops.
  assert (cardinal_le b b). fprops. cp (product_increasing2 H5 H7). rwi H1 H8.
  rwi one_unit_prodl H8.  cp (cardinal_antisymmetry1 H8 H6).
  rwi H9 H1. rwi one_unit_prodr H1. rw H1. rw H9. au. fprops. fprops.
  assert (forall a b, inc a BZps -> inc b BZ ->
    BZmult a b = BZ_one -> (a = b & a = BZ_one)).
  ir. rwi inc_BZ H1. nin H1.  rwi (BZmult_pm (sub_BZps_BZp H0) H1) H2.
  ufi  BZm_of_nat H2. rwi Y_if_not_rw H2. ufi BZ_one H2.
  cp (f_equal Q H2). ufi BZ_of_nat H3. awi H3. TP_tac.
  app zero_cardinal_product2.  ufi BZps H0. awi H0. ee. ufi Bnats H3. Ztac. am.
  ufi BZms H1. awi H1. ee. ufi Bnats H3. Ztac. am.
  rwi (BZmult_pp (sub_BZps_BZp H0) H1) H2. 
  ufi BZps H0. awi H0.  ufi BZp H1. awi H1. uf BZ_one. uf BZ_of_nat.
  nin H0. nin H3. nin H1. nin H5. ufi BZ_one H2.
  cp (f_equal P H2). rwi BZ_of_nat_P H7. rwi BZ_of_nat_P H7.
  nin (H _ _ H3 H5 H7). split. app pair_extensionality. ue. wr H9. wr H4. aw.  
  ir. rwi inc_BZ H1. nin H1. cp (sub_BZms_BZ H1). 
  wri (BZmult_opp_opp H4 H2) H3. cp (BZopp_negative1 H1). 
  cp (H0 _ _ H5 (BZ_stable_opp H2) H3). ee. app BZopp_inj. right.
  cp (BZopp_involutive H4). rwi H7 H8.  wr H8. 
  uf BZopp. uf BZ_one. uf BZ_of_nat. aw. rww Y_if_not_rw. fprops. 
  nin (equal_or_not a BZ_zero). rwi H4 H3. qwi H3. 
  ufi BZ_zero H3; ufi BZ_one H3. cp (BZ_of_nat_inj H3). contradiction. am.
  assert (inc a BZps). rww inc_BZps. eee. nin (H0  _ _ H5 H2). au. am.
Qed.

Lemma BZplus_diag_eq_mult_2: forall n, inc n BZ ->
  BZplus n n = BZmult n BZ_two.
Proof. ir. rww BZ_mult_comm. sy. 
  assert (Ha: forall n, inc n BZp -> BZmult BZ_two n = BZplus n n).
  ir. rww BZ_plus_pp. rww BZmult_pp. uf BZ_two. rw BZ_of_nat_P. 
  rww two_times_n. uf BZ_two.  app inc_BZ_of_natp. fprops.
  rwi inc_BZ H. nin H.  cp (sub_BZms_BZ H). cp (BZ_stable_opp H0).
  cp (BZopp_negative1 H). wr (BZopp_involutive (sub_BZms_BZ H)). 
  wrr BZ_plus_opp_distr. wrr BZopp_mult_distrib_r. rww Ha. app sub_BZps_BZp.
  uf BZ_two. app  inc_BZ_of_nat. fprops. app Ha.
Qed.

Lemma BZmult_succ_r : forall n m, inc n BZ -> inc m BZ ->
  BZmult n (BZsucc m) = BZplus (BZmult n  m) n.
Proof. ir. uf BZsucc. rww BZmult_plus_distr_r. qw. qprops.
Qed.

Lemma BZmult_succ_l : forall n m, inc n BZ -> inc m BZ ->
  BZmult (BZsucc n) m = BZplus (BZmult n m) m.
Proof. ir. rww BZ_mult_comm. rww  BZmult_succ_r. rww  BZ_mult_comm. 
Qed.

(** ** The sign function *)
Definition BZsign x:= Yo (P x = card_zero) BZ_zero
  (Yo (Q x = TPb) BZ_one BZ_mone).

Lemma BZ_stable_sign: forall x,  inc (BZsign x) BZ.
Proof. ir. uf BZsign.  nin (equal_or_not (P x) card_zero).
  rww Y_if_rw. qprops. rww Y_if_not_rw. nin (equal_or_not (Q x) TPb).
  rww Y_if_rw. qprops. rww Y_if_not_rw. qprops.
Qed.

Hint Resolve  BZ_stable_sign: qprops.

Lemma BZ_sign_pos: forall x, inc x BZps -> BZsign x = BZ_one.
Proof. ir. uf BZsign. ufi BZps H. awi H. ee. ufi Bnats H0. Ztac.
  rww Y_if_not_rw. rww Y_if_rw.  
Qed.

Lemma BZ_sign_neg: forall x, inc x BZms -> BZsign x = BZ_mone.
Proof. ir. uf BZsign. ufi BZms H. awi H. ee. ufi Bnats H0. Ztac.
  rww Y_if_not_rw. rww Y_if_not_rw.  rw H1. fprops.
Qed.

Lemma BZ_sign_zero: BZsign BZ_zero = BZ_zero.
Proof. uf BZsign. rw BZ_zero_P. rw BZ_zero_Q. rww Y_if_rw.
Qed.


Lemma BZ_abs_sign: forall x, inc x BZ -> x = BZmult (BZsign x) (BZabs x).
Proof. ir. cp (BZ_stable_abs H).
  rwi inc_BZ1 H. nin H. rw H. rw BZ_sign_zero.  qw. ue.
  nin H. rww BZ_sign_pos. qw. rww BZ_abs_pos. app sub_BZps_BZp. 
  rww BZ_sign_neg. rww BZ_mult_m1_l.
  rww BZ_abs_neg. rww BZopp_involutive. qprops.
Qed.

Lemma BZ_sign_abs: forall x, inc x BZ -> BZmult x (BZsign x) = (BZabs x).
Proof. ir. cp H. cp (BZ_stable_sign x).
  rwi inc_BZ1 H. nin H. rw H. rw BZ_abs_zero. qw. ue.
  nin H. rww BZ_sign_pos. qw. rww BZ_abs_pos. app sub_BZps_BZp. 
  rww BZ_sign_neg. rww BZ_mult_m1_r. rww BZ_abs_neg. 
Qed.

Lemma BZ_sign_trichotomy: forall a, inc a BZ ->
  BZsign a = BZ_one \/ BZsign a = BZ_mone  \/ BZsign a = BZ_zero.
Proof. ir. rwi inc_BZ1 H. nin H. rw H. rw BZ_sign_zero. au.
  nin H. rww BZ_sign_pos. au. rww BZ_sign_neg. au.
Qed. 

Lemma BZ_mult_sign: forall x y,  inc x BZ -> inc y BZ ->  
  BZsign (BZmult x y) = BZmult (BZsign x) (BZsign y).
Proof. ir. cp (BZ_mult_abs2 H H0). cp (f_equal BZsign H1).
  ufi BZ_mult_sign_aux H2.
  nin (equal_or_not x BZ_zero). rw H3. rw BZ_sign_zero. qw.
  rww BZ_sign_zero. qprops. rwii Y_if_not_rw H2.
  nin (equal_or_not y BZ_zero). rw H4. rw BZ_sign_zero. qw.
  rww BZ_sign_zero. qprops. rwii Y_if_not_rw H2. rw H2. 
  assert (card_mult (P x) (P y) <> card_zero). cp (BZmult_0_rl H H0 H3 H4).
  dneg. rwi H6 H1. app BZ_zero_if_abs0. app BZ_stable_mult. rw H1. aw.
  set (t :=  BZmult (BZsign x) (BZsign y)). uf BZsign. aw. rww Y_if_not_rw. 
  uf t. rwi inc_BZ H. nin H. rww BZ_sign_neg.
  rww BZ_mult_m1_l. rwi inc_BZ H0. ufi BZms H. awi H. ee. rw H7.
  nin H0. rww BZ_sign_neg. rw BZ_opp_mone. rww Y_if_rw. rww Y_if_rw. 
  ufi BZms H0. awi H0. ee. ue. rw BZ_sign_pos. rw BZ_opp_one.
  rww Y_if_not_rw. rw Y_if_not_rw. fprops. ufi BZp H0. awi H0. ee. rw H9. 
  fprops. rww inc_BZps. au. qprops. rw BZ_sign_pos. qw.
  ufi BZp H. awi H. ee. rw H7. rwi inc_BZ H0. nin H0.
  rww BZ_sign_neg. rww Y_if_not_rw. rww Y_if_not_rw.  fprops.
  ufi BZms H0. awi H0. ee. ue. rw BZ_sign_pos.  rww Y_if_rw. rww Y_if_rw. 
  ufi BZp H0. awi H0. ee. sy; am. rw inc_BZps. au. qprops. rw inc_BZps. au.
Qed.


Lemma BZ_le_minus: forall a b, inc a BZ -> inc b BZ -> 
  (BZ_le a b = inc (BZminus b a) BZp).
Proof. assert (forall a b, inc a BZp -> inc b BZp -> 
  (BZ_le a b = inc (BZminus b a) BZp)).
  ir. cp (inc_BZ_P (sub_BZp_BZ H)). cp (inc_BZ_P (sub_BZp_BZ H0)). 
  assert (Ha: is_cardinal (P a)).  fprops.
  assert (Hb: is_cardinal (P b)). fprops.
  uf BZminus. rww BZ_le_pr1. nin (equal_or_not a BZ_zero). rw H3.
  rw BZ_zero_P. rw BZopp_zero. qw. app iff_eq. ir.
  app zero_smallest. qprops. assert (inc a BZps). rww inc_BZps. au.
  cp (BZopp_positive1 H4). cp (BZopp_P a). 
  set (ma := BZopp a) in *.
  nin (cardinal_le_total_order2 Ha Hb).
  wri H6 H7. rw (BZ_plus_pm1 H0 H5 H7). ap iff_eq. ir. app inc_BZ_of_natp.
  app Bnat_stable_sub. ue. ir. ue.
  app iff_eq. ir. co_tac. ir. wri H6 H7. 
  rwi (BZ_plus_pm2 H0 H5 H7) H8. ufi BZm_of_nat H8.
  nin (equal_or_not  (card_sub (P ma) (P b)) card_zero).
  nin H7. wri H6 H1. cp (card_sub_pr H1 H2 H7). rwi H9 H11. awi H11.
  contradiction. am. rwi Y_if_not_rw H8. ufi BZp H8. awi H8. ee. TP_tac. am.
  assert (forall a b,
      inc a BZp -> inc b BZms -> BZ_le a b = inc (BZminus b a) BZps).
  ir. app iff_eq. ir. nin (BZ_le_pr2 H0 H1). elim H4. BZo_tac. ir.
  ir. ufi BZminus H2. nin (equal_or_not a BZ_zero). rwi H3 H2.
  qwii H2.  elim (BZ_disjoint_neg_spos H1 H2). qprops.
  assert (inc a BZps). rw inc_BZps. au. cp (BZopp_positive1 H4). 
  elim (BZ_disjoint_neg_spos (BZms_stable_plus_rl H1 H5) H2). 
  ir. rwi inc_BZ H1. nin H1. 
  uf BZminus. cp (BZ_plus_opp_distr H2 (sub_BZms_BZ H1)).
  cp (BZopp_negative1 H1). cp (sub_BZps_BZp H4). cp (BZ_stable_opp H2).
  rwi inc_BZ H6. app iff_eq. ir. cp (BZ_le_opp H7). nin H6.
  cp (BZopp_negative1 H6). rwi BZopp_involutive H9. 
  app BZp_stable_plus. app sub_BZps_BZp. am.
  rwi (H _ _ H6 H5) H8. ufi BZminus H8. rwi BZopp_involutive H8.
  rww BZ_plus_comm. qprops.
  ir. assert (inc (BZminus (BZopp a) (BZopp b)) BZp).
  uf BZminus. rww BZopp_involutive. rww BZ_plus_comm. qprops.
  nin H6. cp (BZopp_negative1 H6). rwi BZopp_involutive H9. 
  cp (sub_BZps_BZp H9). nin (BZ_le_pr2 H10 H1). am. am.
  wri (H _ _ H6 H5) H8. cp (BZ_le_opp H8).
  rwi BZopp_involutive H9. rwi BZopp_involutive H9. am. am. qprops.
  rwi inc_BZ H2. nin H2. rww H0. ap iff_eq. ir.
  app sub_BZps_BZp. ir. rw  inc_BZps. split. am. red. ir.
  cp (BZ_minus_inj (sub_BZms_BZ H2)  (sub_BZp_BZ H1) H4).
  rwi H5 H2. elim (BZ_disjoint_neg_pos H2 H1). app H.
Qed.

Lemma BZ_le_minus1: forall a b, inc a BZ -> inc b BZ -> 
  (BZ_le a b = BZ_le BZ_zero (BZminus b a)).
Proof. ir. rww BZ_le_minus. set (x:= BZminus b a). assert (inc x BZ).
  uf x. app BZ_stable_minus. rww BZ_positive_pr.
Qed.

Lemma BZ_lt_minus: forall a b, inc a BZ -> inc b BZ -> 
  (BZ_lt a b = inc (BZminus b a) BZps).
Proof. ir. uf BZ_lt. app iff_eq. ir. rw inc_BZps. ee. wrr BZ_le_minus.
  dneg. sy.  app BZ_minus_inj. ir. rwi inc_BZps H1. ee. rww BZ_le_minus.
  dneg. rw H3. app BZminus_diag.
Qed.

Lemma BZ_lt_minus1: forall a b, inc a BZ -> inc b BZ -> 
  (BZ_lt a b = BZ_lt BZ_zero (BZminus b a)).
Proof. ir. rww BZ_lt_minus. set (x:= BZminus b a). assert (inc x BZ).
  uf x. app BZ_stable_minus. rww BZ_positive_pr1.
Qed.
Lemma BZ_opp_minus: forall a b, inc a BZ -> inc b BZ ->
  BZopp (BZminus a b) = BZminus b a.
Proof. ir. uf BZminus. rww BZ_plus_opp_distr. rww BZopp_involutive.
  app BZ_plus_comm. qprops.
Qed.


Lemma BZ_lt_minus2: forall a b, inc a BZ -> inc b BZ -> 
  (BZ_lt a b = inc (BZminus a b) BZms).
Proof. ir. rww BZ_lt_minus. wr (BZ_opp_minus H0 H).
  set (c:=  (BZminus b a)). assert (inc c BZ). uf c. app BZ_stable_minus.
  app iff_eq. ir. app BZopp_positive1.
  ir. cp (BZopp_negative1 H2). rwi BZopp_involutive H3. am. am.
Qed.

Lemma BZ_sum_increasing_left: forall a b c, inc a BZ -> inc b BZ -> inc c BZ -> 
  (BZ_le a b = BZ_le (BZplus c a) (BZplus c b)).
Proof. ir. rww BZ_le_minus. cp (BZ_stable_plus H1 H). cp (BZ_stable_plus H1 H0).
  rww BZ_le_minus. uf BZminus. rww (BZ_plus_opp_distr H1 H).
  rw (BZ_plus_comm c b). cp (BZ_stable_opp H). cp (BZ_stable_opp H1).
  rww BZ_plus_assoc. wr (BZ_plus_assoc H0 H1 H5). qw. app BZ_stable_plus.
Qed.

Lemma BZ_sum_increasing_right: forall a b c, inc a BZ -> inc b BZ -> inc c BZ
  ->  (BZ_le a b = BZ_le (BZplus a c) (BZplus b c)).
Proof. ir. rw (BZ_plus_comm a c). rw (BZ_plus_comm b c).
  app BZ_sum_increasing_left.
Qed.

Lemma BZ_sum_strict_increasing_left: forall a b c,
  inc a BZ -> inc b BZ -> inc c BZ -> 
  (BZ_lt a b = BZ_lt (BZplus c a) (BZplus c b)).
Proof. ir. rww BZ_lt_minus. cp (BZ_stable_plus H1 H). cp (BZ_stable_plus H1 H0).
  rww BZ_lt_minus. uf BZminus. rww (BZ_plus_opp_distr H1 H).
  rw (BZ_plus_comm c b). cp (BZ_stable_opp H). cp (BZ_stable_opp H1).
  rww BZ_plus_assoc. wr (BZ_plus_assoc H0 H1 H5). qw. app BZ_stable_plus.
Qed.

Lemma BZ_sum_strict_increasing_right: forall a b c,
  inc a BZ -> inc b BZ -> inc c BZ
  ->  (BZ_lt a b = BZ_lt (BZplus a c) (BZplus b c)).
Proof. ir. rw (BZ_plus_comm a c). rw (BZ_plus_comm b c).
  app BZ_sum_strict_increasing_left.
Qed.

Lemma BZ_induction_pos: forall (a:Set) (r:EP), inc a BZ ->
  (r a) -> (forall n,  BZ_le a n -> r n -> r (BZsucc n))
  ->  (forall n, BZ_le a n -> r n).
Proof. ir. set (s:= fun n => r (BZplus (BZ_of_nat n) a)). assert (inc n BZ).  
  red in H2; ee; am. assert (Ha:inc (BZminus n a) BZp). wrr BZ_le_minus.
  wr (BZplus_minus H H3). rww BZ_plus_comm. wr (inc_BZp_pr Ha).
  change (s (P(BZminus n a))). app cardinal_c_induction_v. uf s. qw.
  ir. uf s. wrr BZ_succ_N.  cp (inc_BZ_of_nat H4).
  assert (BZplus (BZsucc (BZ_of_nat n0)) a = BZsucc (BZplus (BZ_of_nat n0) a)). 
  uf BZsucc. cp inc1_BZ.  wrr BZ_plus_assoc. 
  rw (BZ_plus_comm BZ_one a). rww BZ_plus_assoc. rw H7. app H1.
  rww BZ_le_minus. rww BZ_plus_comm. rw (BZminus_plus H H6).
  app inc_BZ_of_natp. app BZ_stable_plus. app inc_BZ_P. app BZ_stable_minus.
Qed.


Lemma BZ_induction_neg: forall (a:Set) (r:EP), inc a BZ ->
  (r a) -> (forall n, BZ_le n a -> r n -> r (BZpred n))
  ->  (forall n, BZ_le n a -> r n).
Proof. ir. set (b:= BZopp a). set (s := fun n => r (BZopp n)).
  assert (inc b BZ). uf b. qprops. assert (s b). uf s. uf b. 
  rww BZopp_involutive. assert (forall n,  BZ_le b n -> s n -> s (BZsucc n)).
  uf s. ir. assert (BZopp (BZsucc n0)= BZpred (BZopp n0)). uf BZpred.
  uf BZsucc. rw BZ_plus_opp_distr. uf BZminus. tv. red in H5; ee; am. qprops.
  rw H7. app H1. wrr (BZopp_involutive H). app BZ_le_opp.
  assert (inc n BZ). red in H2; ee; am. wrr (BZopp_involutive H6).
  change (s (BZopp n)). cp (BZ_le_opp H2). app (BZ_induction_pos _ H3 H4 H5).
Qed.

Lemma BZ_ind1: forall (a:Set) (p:EP), inc a BZ ->
  p a ->
  (forall x, BZ_le a x-> p x -> p (BZsucc x)) ->
  (forall x, BZ_le x a -> p x -> p (BZpred x)) -> 
  forall n, inc n BZ -> p n.
Proof. ir. nin (BZ_order_total H H3). 
  ap (BZ_induction_pos _ H H0 H1 H4). ap (BZ_induction_neg _ H H0 H2 H4). 
Qed.

Lemma BZ_ind: forall (p:EP),
  p BZ_zero ->
  (forall x, inc x BZ -> p x -> p (BZsucc x)) ->
  (forall x, inc x BZ -> p x -> p (BZpred x)) -> 
  forall n, inc n BZ -> p n.
Proof. ir. apply BZ_ind1 with (a:=BZ_zero). qprops. am.  ir. app H0.
  nin H3; ee; am. ir. app H1. nin H3; ee; am. am.
Qed.


Definition Z_of_BZ n := Yt (inc n BZp) 
  (Z_of_nat (cardinal_nat (P n)))
  (Zopp (Z_of_nat (cardinal_nat (P n)))).

Definition BZ_of_Z (n:Z) :=
  match n with | Z0 => BZ_zero
    | Zpos p => BZ_of_nat (nat_to_B (nat_of_P p))
    | Zneg p => BZm_of_nat (nat_to_B (nat_of_P p))
end.

Lemma Z_of_BZ_pos: forall n, inc n BZp -> 
  Z_of_BZ n = Z_of_nat (cardinal_nat (P n)).
Proof. ir. uf Z_of_BZ. rww (Yt_if_rw H (A:=Z)).
Qed.

Lemma Z_of_BZ_neg: forall n, inc n BZms -> 
  Z_of_BZ n = Zopp (Z_of_nat (cardinal_nat (P n))).
Proof. ir. uf Z_of_BZ. rww Yt_if_not_rw. 
  red; ir. app (BZ_disjoint_neg_pos H H0).
Qed.

Lemma Z_of_BZ_zero: Z_of_BZ BZ_zero = 0%Z.
Proof. rw  Z_of_BZ_pos. rw BZ_zero_P. rw cardinal_nat_zero.  auto.
  uf BZ_zero. app inc_BZ_of_natp. fprops.
Qed.

Lemma Z_of_BZ_one: Z_of_BZ BZ_one = 1%Z.
Proof. uf BZ_one. rw Z_of_BZ_pos. rw BZ_of_nat_P. rw cardinal_nat_one. 
  auto. app inc_BZ_of_natp. fprops.
Qed.

Lemma Z_of_BZ_mone: Z_of_BZ BZ_mone = Zopp 1%Z.
Proof. uf BZ_mone. rw Z_of_BZ_neg.  rw BZm_of_nat_P. rw cardinal_nat_one. 
  auto. qprops. 
Qed.

Lemma BZ_of_Z_inc: forall x, inc (BZ_of_Z x) BZ.
Proof. ir. uf BZ_of_Z. elim x. qprops. ir.  app inc_BZ_of_nat. fprops. 
  ir. app inc_BZm_of_nat. fprops.
Qed.

Hint Resolve BZ_of_Z_inc: qprops.

Lemma positive_non_zero: forall p: positive,
  inc (nat_to_B (nat_of_P p)) Bnats.
Proof. ir.  uf Bnats. Ztac. fprops. red. ir. 
  wri nat_B_0 H. cp (nat_B_inj _ _ H). cp (lt_O_nat_of_P p). rwi H0 H1.
  elim (lt_irrefl 0). am.
Qed.

Lemma BZ_of_Z_pr: forall x, Z_of_BZ (BZ_of_Z x) = x.
Proof. ir. uf BZ_of_Z. elim x. 
  rw (Z_of_BZ_pos (inc0_BZp)). rw BZ_zero_P. rw cardinal_nat_zero. auto.
  ir.  rw Z_of_BZ_pos. aw. sy. rw BZ_of_nat_P. aw.
  ap Zpos_eq_Z_of_nat_o_nat_of_P. app inc_BZ_of_natp. fprops.
  ir. rw Z_of_BZ_neg. rw BZm_of_nat_P.  aw.  wr Zpos_eq_Z_of_nat_o_nat_of_P.
  auto. app inc_BZm_of_natms. fprops. aw. cp (positive_non_zero p).
  ufi Bnats H. Ztac. am.
Qed.

Lemma Z_of_BZ_opp: forall x, inc x BZ -> Z_of_BZ (BZopp x) = Zopp (Z_of_BZ x).
Proof. ir. rwi inc_BZ1 H. nin H. rw H. qw. rw Z_of_BZ_zero. autoa.
  nin H. cp (BZopp_positive1 H).  rw (Z_of_BZ_neg H0). 
  rw BZopp_P. rw (Z_of_BZ_pos (sub_BZps_BZp H)). tv. fprops.
  cp (BZopp_negative1 H).  rw (Z_of_BZ_neg H). 
  rw Zopp_involutive. rww (Z_of_BZ_pos  (sub_BZps_BZp H0)). rww BZopp_P.
Qed.

Lemma BZ_opp_pos3: forall m,  BZopp (BZ_of_nat m) = BZm_of_nat m.
Proof. ir. uf BZopp. rww Y_if_not_rw. rww BZ_of_nat_P.
  uf BZ_of_nat. aw. fprops.
Qed.

Lemma BZ_opp_neg3: forall m,  BZopp (BZm_of_nat m) = BZ_of_nat m.
Proof. ir. uf BZm_of_nat. nin (equal_or_not m card_zero). rw H.
  rww Y_if_rw. qw. rw Y_if_not_rw. uf BZopp. aw. rww Y_if_rw. am.
Qed.

Lemma BZ_Z_opp: forall x, BZ_of_Z (Zopp x) = BZopp (BZ_of_Z x).
Proof. ir. induction x. simpl. qw. simpl. rww  BZ_opp_pos3.
  simpl. rww BZ_opp_neg3.
Qed.

Lemma BZ_of_Z_pr1: forall x, inc x BZ -> BZ_of_Z (Z_of_BZ x) = x.
Proof. assert (forall x, inc x BZp ->  BZ_of_Z (Z_of_BZ x) = x). ir.
  rww Z_of_BZ_pos.  set (y:= cardinal_nat (P x)). wrr inc_BZp_pr.
  wr (nat_to_B_pr (inc_BZ_P (sub_BZp_BZ H))). fold y. 
  nin y. tv. set (z:= nat_to_B (S y)). simpl. uf z.
  rww nat_of_P_o_P_of_succ_nat_eq_succ.
  ir. cp H0.  rwi inc_BZ H0. nin H0. cp (BZopp_negative1 H0).
  cp (H _ (sub_BZps_BZp H2)). rwi Z_of_BZ_opp H3. rwi BZ_Z_opp H3.
  app BZopp_inj. qprops. am. app H. 
Qed.

Lemma BZ_Z_inj: forall a b, BZ_of_Z a = BZ_of_Z b -> a = b.
Proof. ir. wr (BZ_of_Z_pr a).  wr (BZ_of_Z_pr b). rww H.
Qed.

Lemma BZ_Z_surjective: forall a, inc a BZ -> exists b,  BZ_of_Z b = a.
Proof. ir. exists (Z_of_BZ a). rww (BZ_of_Z_pr1 H).
Qed.

Lemma BZ_Z_0: BZ_of_Z 0 = BZ_zero.
Proof. tv. Qed.

Lemma BZ_Z_1: BZ_of_Z 1 = BZ_one.
Proof. simpl. rww succ_zero. Qed.

Lemma BZ_Z_m1: BZ_of_Z (- 1)%Z = BZ_mone.
Proof. simpl. rww succ_zero. Qed.

Lemma BZ_of_Z_neg: forall p: positive,
  BZ_of_Z (Zneg p) = BZopp (BZ_of_Z (Zpos p)).
Proof. ir. assert (Zneg p = Zopp (Zpos p)). auto. rw H. rww BZ_Z_opp.
Qed.

Lemma BZ_Z_mult: forall a b,  BZ_of_Z (a*b)= BZmult (BZ_of_Z a) (BZ_of_Z b).
Proof. assert (Ha: forall p p0: positive,
  BZ_of_Z (Zpos p * Zpos p0) = BZmult (BZ_of_Z (Zpos p)) (BZ_of_Z (Zpos p0))).
  ir. simpl. set (a:= nat_to_B (nat_of_P p)). assert (inc a Bnat). uf a. fprops.
  set (b:= nat_to_B (nat_of_P p0)). assert (inc b Bnat). uf b. fprops.
  rww BZmult_pp. rw BZ_of_nat_P. rw BZ_of_nat_P. rw nat_of_P_mult_morphism. aw. 
  app  inc_BZ_of_natp.  app  inc_BZ_of_natp.
  assert (Hb:forall (a:Z) (b:positive), Zmult a (Zneg b) = Zopp (a * Zpos b)). 
  ir. rw Zopp_mult_distr_r. auto.
  assert (Hc:forall (a:positive) (b : Z), 
    BZ_of_Z ((Zpos a) * b) = BZmult (BZ_of_Z (Zpos a)) (BZ_of_Z b)).
  ir. nin b. rw BZ_Z_0. qw. qprops. 
  app Ha. rww BZ_of_Z_neg. wrr BZopp_mult_distrib_r. rw Hb. rww BZ_Z_opp.
  rww Ha. qprops. qprops.
  ir. nin a. qw. qprops. app Hc. 
  rww BZ_of_Z_neg.  wrr BZopp_mult_distrib_l.
  rw Zmult_comm. rw Hb. assert ((- (b * Zpos p))%Z = Zopp (Zpos p * b)).
  rww Zmult_comm. rw H. rw BZ_Z_opp. rww Hc. qprops. qprops.
Qed.

Lemma BZ_Z_plus: forall a b,  BZ_of_Z (a+b)= BZplus (BZ_of_Z a) (BZ_of_Z b).
Proof. assert (Ha: forall a b: positive,  
  BZ_of_Z (Zpos a + Zpos b) = BZplus (BZ_of_Z (Zpos a)) (BZ_of_Z (Zpos b))).
  ir. simpl. rw nat_of_P_plus_morphism. aw. rww BZ_plus_pp. 
  rw BZ_of_nat_P. rw BZ_of_nat_P. tv. app inc_BZ_of_natp. fprops.
  app inc_BZ_of_natp. fprops.
  assert (Hb: forall a b: positive,  
  BZ_of_Z (Zneg a + Zneg b) = BZplus (BZ_of_Z (Zneg a)) (BZ_of_Z (Zneg b))).
  ir. rw BZ_of_Z_neg. rw BZ_of_Z_neg. wr BZ_plus_opp_distr. wr Ha. wr BZ_Z_opp.
  rw Zopp_plus_distr. autoa. qprops. qprops.
  assert (Hc: forall a b: positive,  
  BZ_of_Z (Zpos a + Zneg b) = BZplus (BZ_of_Z (Zpos a)) (BZ_of_Z (Zneg b))).
  ir. rw BZ_of_Z_neg.
  elim (Dcompare ((a ?= b)%positive Eq)). ir. cp (Pcompare_Eq_eq _ _ H).
  simpl.  rw H. rw H0. rw BZplus_opp_r. rw BZ_Z_0. tv.  app inc_BZ_of_nat.
  fprops.
  ir. nin H. cp (nat_of_P_lt_Lt_compare_morphism _ _ H). simpl. rw H.
  awi H0. set (u:= nat_to_B (nat_of_P a)) in *.
  set (v:= nat_to_B (nat_of_P b)) in *.  assert (inc v Bnat). uf v. fprops.
  assert (inc v Bnats). uf Bnats. Ztac. red. ir. rwi H2 H0.
  app (zero_smallest1 H0). assert (inc (BZ_of_nat v) BZps). uf BZps.
  uf BZ_of_nat. aw. eee. cp (BZopp_positive1 H3). 
  set (vo:= BZopp (BZ_of_nat v)) in *.
  assert (inc u Bnat). uf u. fprops.
  wri (BZ_of_nat_P u) H0.  assert (v = P vo). uf vo. 
  rw (BZopp_P (BZ_of_nat v)) . rww BZ_of_nat_P. rwi H6 H0.
  rw BZ_of_Z_neg. rww BZ_plus_pm2. simpl.
  rw (nat_of_P_minus_morphism _ _ (ZC2 _ _ H)). rw BZ_of_nat_P. aw.
  simpl. wr H6. rww BZ_opp_pos3. app inc_BZ_of_natp.
  simpl.  rw H. cp (nat_of_P_gt_Gt_compare_morphism _ _ H). red in H0. awi H0.
  set (u:= nat_to_B (nat_of_P a)) in *. set (v:= nat_to_B (nat_of_P b)) in *. 
  assert (inc v Bnats). uf v. app positive_non_zero. simpl.
  assert (inc u Bnat). uf u. fprops. cp (inc_BZ_of_natp H2).
  wri (BZ_of_nat_P u) H0.  assert (inc (BZ_of_nat v) BZps). uf BZps.
  uf BZ_of_nat. aw. eee. cp (BZopp_positive1 H4). 
  set (vo:= BZopp (BZ_of_nat v)) in *.
  assert (v = P vo). uf vo. rw (BZopp_P (BZ_of_nat v)). rww BZ_of_nat_P. 
  nin H0. rwi H6 H0. rw (BZ_plus_pm1 H3 H5 H0).  wr H6. rw BZ_of_nat_P. 
  rww (nat_of_P_minus_morphism _ _ H). aw. 
  ir. nin a. rw BZ_Z_0. rw Zplus_0_l. rww BZplus_0_l. qprops. 
  nin b. rw BZ_Z_0. rw Zplus_0_r. rww BZplus_0_r. qprops. app Ha. app Hc.
  nin b. rw BZ_Z_0. rw Zplus_0_r. rww BZplus_0_r. qprops.
  rw Zplus_comm. rw BZ_plus_comm. app Hc. qprops.
Qed.

Lemma BZ_Z_minus: forall a b,  BZ_of_Z (a-b)= BZminus (BZ_of_Z a) (BZ_of_Z b).
Proof. ir. uf Zminus. rw BZ_Z_plus. rww BZ_Z_opp.
Qed.

Lemma BZ_Z_pred: forall a,  BZ_of_Z (Zpred a)= BZpred (BZ_of_Z a).
Proof. ir. uf Zpred. uf BZpred. rw BZ_Z_plus. rww BZ_Z_m1. uf BZminus. 
  rww BZ_opp_one.
Qed.

Lemma BZ_Z_succ: forall a,  BZ_of_Z (Zsucc a)= BZsucc (BZ_of_Z a).
Proof. ir. uf Zsucc. uf BZsucc. rw BZ_Z_plus. rww BZ_Z_1.
Qed.

(** More comparision *)

Lemma BZ_lt_opp: forall x y, BZ_lt x y -> BZ_lt (BZopp y) (BZopp x).
Proof. ir.  nin H. red. split. app  BZ_le_opp. red. ir. elim H0.
  sy. red in H. ee. app BZopp_inj. 
Qed.

Lemma BZ_lt_opp1: forall x y, inc x BZ -> inc y BZ ->
  BZ_lt x y = BZ_lt (BZopp y) (BZopp x).
Proof. ir. app iff_eq. ir. app  BZ_lt_opp. ir. wrr (BZopp_involutive H).
  wrr (BZopp_involutive H0). app  BZ_lt_opp.
Qed.

Lemma BZ_sum_increasing0: forall a b c d, BZ_le a c -> BZ_le b d ->
  BZ_le (BZplus a b) (BZplus c d).
Proof. ir. assert (inc a BZ). BZo_tac. assert (inc c BZ). BZo_tac. 
  assert (inc b BZ). BZo_tac. assert (inc d BZ). BZo_tac. 
  assert  (BZ_le (BZplus a b) (BZplus a d)). 
  wrr BZ_sum_increasing_left. 
  assert (BZ_le (BZplus a d) (BZplus c d)). 
  wrr BZ_sum_increasing_right. BZo_tac.
Qed.

Lemma BZ_sum_increasing1: forall a b c d, BZ_le a c -> BZ_lt b d ->
  BZ_lt (BZplus a b) (BZplus c d).
Proof. ir. assert (inc a BZ). BZo_tac. assert (inc c BZ). BZo_tac. 
  assert (inc b BZ). nin H0. BZo_tac. assert (inc d BZ). nin H0. BZo_tac. 
  assert  (BZ_lt (BZplus a b) (BZplus a d)). 
  wrr BZ_sum_strict_increasing_left. 
  assert (BZ_le (BZplus a d) (BZplus c d)). 
  wrr BZ_sum_increasing_right. BZo_tac.
Qed.

Lemma BZ_sum_increasing2: forall a b c d, BZ_lt a c -> BZ_le b d ->
  BZ_lt (BZplus a b) (BZplus c d).
Proof. ir. rw BZ_plus_comm. rw (BZ_plus_comm c d).
  app BZ_sum_increasing1.
Qed.

Lemma BZ_sum_increasing3: forall a b c d,  BZ_lt a c -> BZ_lt b d ->
  BZ_lt (BZplus a b) (BZplus c d).
Proof. ir. nin H. app BZ_sum_increasing1.
Qed.


Lemma BZ_sum_increasing4: forall a c d, BZ_le a c -> BZ_le BZ_zero d ->
  BZ_le a (BZplus c d).
Proof. ir. cp  (BZ_sum_increasing0 H H0). qwi H1. am. BZo_tac.
Qed.

Lemma BZ_sum_increasing5: forall a c d, BZ_le a c -> BZ_lt BZ_zero d ->
  BZ_lt a (BZplus c d).
Proof. ir. cp  (BZ_sum_increasing1 H H0). qwi H1. am. BZo_tac.
Qed.

Lemma BZ_sum_increasing6: forall a c d, BZ_lt a c -> BZ_le BZ_zero d ->
  BZ_lt a (BZplus c d).
Proof. ir. cp  (BZ_sum_increasing2 H H0). qwi H1. am. nin H. BZo_tac.
Qed.

Lemma BZ_sum_increasing7: forall a c d,  BZ_lt a c -> BZ_lt BZ_zero d ->
  BZ_lt a (BZplus c d).
Proof. ir. cp  (BZ_sum_increasing3 H H0). qwi H1. am. nin H. BZo_tac.
Qed.


Lemma BZ_sum_increasing8: forall a b c , BZ_le a c -> BZ_le b BZ_zero ->
  BZ_le (BZplus a b) c.
Proof. ir. cp (BZ_sum_increasing0 H H0). qwi H1. am. BZo_tac.
Qed.

Lemma BZ_sum_increasing9: forall a b c, BZ_le a c -> BZ_lt b BZ_zero ->
  BZ_lt (BZplus a b) c.
Proof. ir. cp (BZ_sum_increasing1 H H0). qwi H1. am. BZo_tac.
Qed.

Lemma BZ_sum_increasing10: forall a b c, BZ_lt a c -> BZ_le b BZ_zero ->
  BZ_lt (BZplus a b) c.
Proof. ir. cp (BZ_sum_increasing2 H H0). qwi H1. am. nin H. BZo_tac.
Qed.

Lemma BZ_sum_increasing12: forall a b c,  BZ_lt a c -> BZ_lt b BZ_zero ->
  BZ_lt (BZplus a b) c.
Proof. ir. cp (BZ_sum_increasing3 H H0). qwi H1. am. nin H. BZo_tac.
Qed.

Lemma BZ_sum_increasing13: forall a b, inc a BZ -> inc b BZp ->
  BZ_le a (BZplus a b).
Proof. ir. wri BZ_positive_pr H0.
  cp (BZ_le_reflexive H). app BZ_sum_increasing4.
Qed.

Lemma BZ_sum_increasing14: forall a b, inc a BZ -> inc b BZps ->
  BZ_lt a (BZplus a b).
Proof. ir. wri BZ_positive_pr1 H0.
  cp (BZ_le_reflexive H). app BZ_sum_increasing5.
Qed.


Lemma BZ_sum_increasing15: forall a b, inc a BZ -> inc b BZm ->
  BZ_le (BZplus a b) a.
Proof. ir. wri BZ_negative_pr1 H0.
  cp (BZ_le_reflexive H). app BZ_sum_increasing8.
Qed.

Lemma BZ_sum_increasing16: forall a b, inc a BZ -> inc b BZms ->
  BZ_lt (BZplus a b) a.
Proof. ir. wri BZ_negative_pr H0.
  cp (BZ_le_reflexive H). app BZ_sum_increasing9.
Qed.

Lemma BZ_lt_succ: forall n, inc n BZ -> BZ_lt n (BZsucc n).
Proof. ir. uf BZsucc. rww BZ_lt_minus. uf BZsucc. rww BZminus_plus.
  app inc1_BZps. qprops. app BZ_stable_plus. qprops.
Qed.

Lemma BZ_lt_pred: forall n, inc n BZ -> BZ_lt (BZpred n) n.
Proof. ir. assert (inc (BZpred n) BZ). uf BZpred. app BZ_stable_minus.
  qprops. wr (BZ_succ_pred H). rw BZ_pred_succ. app BZ_lt_succ. am.
Qed.

Lemma BZp_stable_prec: forall a, inc a BZ -> inc (BZsucc a) BZps ->
  inc a BZp.
Proof. ir. cp (BZ_pred_succ H). set (b := BZsucc a) in *. wr H1.
  assert (inc (P b) Bnats). ufi BZps H0. awi H0. eee. 
  rwi inc_BZps H0. nin H0.  cp (inc_BZp_pr H0). wr H4. rww BZ_prec_N.
  app inc_BZ_of_natp. ufi Bnats H2. Ztac.  assert (is_finite_c (P b)). fprops.
  cp (predc_pr H7 H6). ee. bw. 
Qed.


Lemma BZ_lt_succ1: forall a b, inc a BZ -> inc b BZ ->
  BZ_lt a (BZsucc b) = BZ_le a b.
Proof. ir.  app iff_eq. ir.
  rwii BZ_lt_minus H1. wrii BZminus_succ_l H1. rww BZ_le_minus. 
  app  BZp_stable_prec. app BZ_stable_minus. uf BZsucc. app BZ_stable_plus. 
  qprops. ir. cp (BZ_lt_succ H0). BZo_tac.
Qed.

Lemma BZ_order_isomorphism_prop: forall f,
 (order_isomorphism f BZ_ordering BZ_ordering) =
 (exists u, inc u BZ & f = BL (fun z => BZplus z u) BZ BZ).
Proof. ir. set (p := fun a => BL (fun z : Set => BZplus z a) BZ BZ).
  assert (forall a, inc a BZ -> transf_axioms (fun z => BZplus z a) BZ BZ).
  ir. red. ir. app BZ_stable_plus.
  assert (forall a, inc a BZ -> bijective (p a)). ir. uf p.
  app bl_bijective. app H. ir. app (BZplus_reg_r H0 H1 H2).
  ir. exists (BZminus y a). split. app BZ_stable_minus.
  rw BZ_plus_comm. app BZplus_minus.
  assert (forall a b c, inc a BZ -> BZ_le b c -> BZ_le (W b (p a)) (W c (p a))).
  ir. uf p. rw bl_W. rw bl_W.
  app BZ_sum_increasing0. app BZ_le_reflexive. app H. BZo_tac. app H.
  BZo_tac.
  assert (forall a, inc a BZ -> order_isomorphism (p a)BZ_ordering BZ_ordering).
  ir. red. ee. app BZ_order. app BZ_order. app H0. uf p. aw. 
  rww substrate_BZ_order. uf p. aw.
  rww substrate_BZ_order. ir. cp H3; cp H4. ufi p H3. ufi p H4. awi H3. awi H4.
  cp (H0 _ H2).
  assert (inc  (W x (p a))  (target (p a))). ap inc_W_target. fct_tac. tv.
  assert (inc  (W y (p a))  (target (p a))). ap inc_W_target. fct_tac. tv.
  rww BZ_le_pr. rww BZ_le_pr.  app iff_eq. ir. app H1. ir.
  nin (BZ_order_total H3 H4). am. cp (H1 _  _ _ H2 H11).
  assert (W y (p a) = W x (p a)). BZo_tac. nin H7. nin H7. 
  rw (H15 _ _ H6 H5 H13).  app BZ_le_reflexive. 
  ap iff_eq. ir.
  red in H3. rwi substrate_BZ_order H3. ee.
  assert (forall a b, BZ_le a b -> BZ_le (W a f) (W b f)). ir.
  wr BZ_le_pr. wrr H8. rww BZ_le_pr. wr H6. BZo_tac. wr H6. BZo_tac.
  assert (forall a b, BZ_lt a b -> BZ_lt (W a f) (W b f)). ir.
  nin H10. split. app H9. dneg. red in H5. ee. nin H5. app H14.
  wr H6. BZo_tac. wr H6. BZo_tac.
  assert (inc (W BZ_zero f) BZ). 
  rw H7. app inc_W_target. fct_tac. wr H6. qprops. exists (W BZ_zero f). ee. 
  am. change (f = p (W BZ_zero f)). cp (H2 _ H11). 
  set (g := p (W BZ_zero f)) in *.  red in H12. rwi substrate_BZ_order H12. 
  ee. app function_exten. fct_tac. fct_tac.  ue. ue.
  wr H6. app BZ_ind. uf g. uf p. rw bl_W. qw. app H. qprops.
  ir. assert (inc (BZsucc x) BZ).  app BZ_stable_succ.
  assert (W (BZsucc x) g = BZplus (W x f) BZ_one). uf g. uf p.
  rw bl_W. uf BZsucc. rw (BZ_plus_comm x BZ_one). wrr BZ_plus_assoc.
  ufi g H19. ufi p H19. rwi bl_W H19. wrr H19. rww BZ_plus_comm.
  app H. am. qprops. app H. am. rw H21.
  assert (inc (BZplus (W x f) BZ_one) BZ). app BZ_stable_plus.
  rw H7. app inc_W_target. fct_tac.  wrr H6. qprops.
  nin H5. rwi H7 H22. nin (surjective_pr2 H23 H22). nin H24.
  wri H6 H24. nin (BZ_order_total1 H24 H18).  cp (H9 _ _ H26). rwi H25 H27.
  assert (inc (W x f) BZ). BZo_tac. cp (BZ_lt_succ H28).
  ufi BZsucc H29. BZo_tac. assert (BZ_le (BZsucc x) x0). wr BZ_lt_succ1.
  uf BZsucc. app  BZ_sum_increasing2. app BZ_le_reflexive. qprops.  am. am.
  nin (equal_or_not (BZsucc x) x0). rww H28. 
  assert (BZ_lt (BZsucc x) x0). split; am. cp (H10 _ _ H29). rwi H25 H30.
  cp (BZ_lt_succ  H18).  cp (H10 _ _ H31). 
  change (BZ_lt (W (BZsucc x) f) (BZsucc (W x f))) in H30.
  rwi BZ_lt_succ1 H30. BZo_tac. BZo_tac. BZo_tac.
  ir. assert (inc (BZpred x) BZ).  app BZ_stable_pred.
  assert (W (BZpred x) g = BZplus (W x f) BZ_mone). uf g. uf p.
  rw bl_W. uf BZpred. uf BZminus. rw BZ_opp_one. 
  rw (BZ_plus_comm x BZ_mone). wrr BZ_plus_assoc. rww BZ_plus_comm.
  ufi g H19. ufi p H19. rwi bl_W H19. wrr H19. app H. am. qprops. 
  app H. am. rw H21.
  assert (Ha : inc  (W x f)  BZ). rw H7. app inc_W_target. fct_tac. wrr H6. 
  assert (Hb:= inc1_BZ). assert (Hc:= inc_m1_BZ).
  assert (inc (BZplus (W x f) BZ_mone) BZ). app BZ_stable_plus. 
  nin H5. rwi H7 H22. nin (surjective_pr2 H23 H22). nin H24.
  wri H6 H24. nin (BZ_order_total1 H18 H24).  cp (H9 _ _ H26). rwi H25 H27.
  assert (inc  (BZplus (W x f) BZ_mone) BZ). BZo_tac. cp (BZ_lt_succ H28).
  assert (BZ_lt  (W x f) (BZsucc (BZplus (W x f) BZ_mone))). BZo_tac.
  ufi BZsucc H30. wrii BZ_plus_assoc H30. wri BZ_opp_one H30. qwi H30. nin H30.
  elim H31. tv. am. am. 
  wri (BZ_succ_pred  H18) H26. rwii BZ_lt_succ1 H26.
  nin (equal_or_not x0 (BZpred x)). wrr H27. 
  assert (BZ_lt x0 (BZpred x)). split; am. cp (H10 _ _ H28). rwi H25 H29.
  assert (inc (BZpred x) BZ). app BZ_stable_pred. cp (BZ_lt_succ H30).  
  rwii BZ_succ_pred  H31. cp (H10 _ _ H31). 
  assert (inc  (W (BZpred x) f) BZ). BZo_tac. 
  assert (inc  (BZplus (W x f) BZ_mone) BZ). BZo_tac. 
  rwi (BZ_sum_strict_increasing_right H34 H33 inc1_BZ) H29.
  wrii BZ_plus_assoc H29. wri BZ_opp_one H29. qwi H29. 
  change (BZ_lt (W x f) (BZsucc (W (BZpred x) f))) in H29.
  rwii BZ_lt_succ1 H29.  BZo_tac. am. am. 
  ir. nin H3. nin H3. change (f = p x) in H4.  rw H4. app H2.
Qed.

Lemma BZ_le_abs: forall n, inc n BZ -> BZ_le n (BZabs n).
Proof. ir. cp H. rwi inc_BZ H. nin H. cp (inc_BZabs_N H0).
  nin (BZ_le_pr2 H1 H). am. rww BZ_abs_pos. app BZ_le_reflexive.
Qed.

Lemma BZ_le_triangular: forall n m, inc n BZ -> inc m BZ -> 
  BZ_le (BZabs (BZplus n m)) (BZplus (BZabs n) (BZabs m)).
Proof. assert (forall n m, inc n BZp -> inc m BZp -> 
  BZ_le (BZabs (BZplus n m)) (BZplus (BZabs n) (BZabs m))).
  ir. rw (BZ_abs_pos H). rw (BZ_abs_pos H0).
  cp (BZp_stable_plus H H0). rw  (BZ_abs_pos H1). app BZ_le_reflexive. qprops.
  assert (forall n m, inc n BZp -> inc m BZ -> 
  BZ_le (BZabs (BZplus n m)) (BZplus (BZabs n) (BZabs m))).
  ir. rwi inc_BZ H1. nin H1. 
  rw (BZ_abs_pos H0). rw (BZ_abs_neg H1).
  assert (inc (P n) Bnat). app inc_BZ_P. qprops.
  assert (inc (P m) Bnat). app inc_BZ_P. qprops.
  nin (cardinal_le_total_order4 H3 H2).
  assert (inc (BZ_of_nat (card_sub (P n) (P m))) BZp). 
  app inc_BZ_of_natp. fprops. 
  rw (BZ_plus_pm1 H0 H1 H4). rww BZ_abs_pos.
  assert (cardinal_le (P n) (card_plus (P n) (P m))). app sum_increasing3.
  fprops. fprops. cp (sub_increasing2 H2 H3 H2 H6 H4).
  assert (BZ_le (BZ_of_nat (card_sub (P n) (P m))) n). rw BZ_le_pr1.
  rww BZ_of_nat_P. am. am. assert (BZ_le n (BZplus n (BZopp m))). 
  app BZ_sum_increasing13. qprops.
  app sub_BZps_BZp. app BZopp_negative1. BZo_tac. 
  assert (inc (BZm_of_nat (card_sub (P m) (P n))) BZms).
  app inc_BZm_of_natms. fprops. app card_sub_non_zero1.
  rw (BZ_plus_pm2 H0 H1 H4). rww BZ_abs_neg. rw BZ_opp_neg3.
  assert (cardinal_le (P m) (card_plus (P m) (P n))). app sum_increasing3.
  fprops. fprops. nin H4. cp (sub_increasing2 H3 H2 H3 H6 H4).
  assert (inc (BZopp m) BZp). app BZopp_negative2. app sub_BZms_BZm.
  assert (BZ_le (BZ_of_nat (card_sub (P m) (P n))) (BZopp m)). rw BZ_le_pr1.
  rww BZ_of_nat_P. rw BZopp_P. am. fprops. app inc_BZ_of_natp. fprops. am.
  assert (BZ_le (BZopp m) (BZplus n (BZopp m))). rw BZ_plus_comm.
  app BZ_sum_increasing13. qprops. BZo_tac. app H.
  ir. rwi inc_BZ H1. nin  H1. cp (sub_BZms_BZ H1). cp (BZ_stable_plus H3 H2).
  cp (H0 _ _ (sub_BZps_BZp (BZopp_negative1 H1)) (BZ_stable_opp H2)).
  wri BZ_plus_opp_distr H5.  do 3 rwii BZabs_opp H5.  am. am. app H0.
Qed.

Lemma BZp_stable_mult: forall a b, inc a BZp -> inc b BZp->
  inc (BZmult a b) BZp.
Proof. ir. rww BZmult_pp. app inc_BZ_of_natp. app BZ_mult_aux.
  qprops. qprops.
Qed.

Lemma BZps_stable_mult: forall a b, inc a BZps -> inc b BZps ->
  inc (BZmult a b) BZps.
Proof. ir. rwi inc_BZps H.  rwi inc_BZps H0.  ee.
  rw inc_BZps. split.  app  BZp_stable_mult.  app BZmult_0_rl.
  qprops. qprops.
Qed.


Lemma BZm_unstable_mult: forall a b, inc a BZm -> inc b BZm ->
  inc (BZmult a b) BZp.
Proof. ir. cp inc0_BZp. ufi BZm H. rwi tack_on_inc H. nin H.
  ufi BZm H0. rwi tack_on_inc H0. nin H0.
  rww BZmult_mm. app inc_BZ_of_natp. app BZ_mult_aux.
  qprops. qprops. rw H0. qw. qprops. rw H. qw. qprops.
Qed.

Lemma BZms_unstable_mult: forall a b, inc a BZms -> inc b BZms ->
  inc (BZmult a b) BZps.
Proof. ir. rw inc_BZps. split.  rww BZmult_mm. app inc_BZ_of_natp. 
  app BZ_mult_aux. qprops. qprops. app BZmult_0_rl. qprops. qprops.
  red. ir. elim (BZ_disjoint_neg_0 H H1).
  red. ir. elim (BZ_disjoint_neg_0 H0 H1).
Qed.

Lemma BZpm_stable_mult: forall a b, inc a BZp -> inc b BZm ->
  inc (BZmult a b) BZm.
Proof. ir.  ufi BZm H0. rwi tack_on_inc H0. nin H0. rww BZmult_pm.
  app inc_BZm_of_natm.  app BZ_mult_aux. qprops. qprops. rw H0.
  qw. uf BZm. fprops. qprops.
Qed.


Lemma BZpms_stable_mult: forall a b, inc a BZps -> inc b BZms ->
  inc (BZmult a b) BZms.
Proof. ir. rwi inc_BZps H.  nin H. assert (inc (BZmult a b) BZ).
  app BZ_stable_mult; qprops.
  rwi inc_BZ H2. nin H2. am. assert (inc b BZm). app sub_BZms_BZm.
  cp (BZpm_stable_mult H H3). assert (inc (BZmult a b) BZps). rw inc_BZps. 
  split. am. assert (b <> BZ_zero). red. ir. elim (BZ_disjoint_neg_0 H0 H5).
  app BZmult_0_rl. qprops. qprops. elim (BZ_disjoint_pos_neg H5 H4).
Qed.

Lemma BZps_stable_mult1: forall a b,
  inc a BZ -> inc b BZ -> inc (BZmult a b) BZps ->
  ((inc a BZps = inc b BZps) & (inc a BZms = inc b BZms)).
Proof. ir. cp H; cp H0. rwi inc_BZ1 H.  nin H. rwi H H1. qwi H1.
  rwi inc_BZps H1.  nin H1. elim H4. tv. am.
  rwi inc_BZ1 H0.  nin H0. rwi H0 H1. qwi H1.
  rwi inc_BZps H1. nin H1. elim H4. tv. am. nin H; nin H0.  split. app iff_eq.
  app iff_eq. ir. elim (BZ_disjoint_neg_spos H4 H).
  ir.  elim (BZ_disjoint_neg_spos H4 H0).
  cp (BZpms_stable_mult H H0).  elim (BZ_disjoint_neg_spos H4 H1).
  cp (BZpms_stable_mult H0 H). rwi BZ_mult_comm H4. 
  elim (BZ_disjoint_neg_spos H4 H1).
  split. ap iff_eq. ir. elim (BZ_disjoint_neg_spos H H4).
  ir.  elim (BZ_disjoint_neg_spos H0 H4). app iff_eq; ir.
Qed.


Lemma BZ_prod_increasing1: forall a b c, inc c BZp ->
  BZ_le a b -> BZ_le  (BZmult a c)  (BZmult b c).
Proof. ir. assert (inc a BZ). BZo_tac. assert (inc b BZ). BZo_tac.
  cp (sub_BZp_BZ H). rw BZ_le_minus. wrr BZmult_minus_distr_l. 
  app BZp_stable_mult. wrr BZ_le_minus.  app BZ_stable_mult. 
  app BZ_stable_mult.
Qed.

Lemma BZ_prod_increasing2: forall a b c, inc c BZps ->
  BZ_lt a b -> BZ_lt  (BZmult a c)  (BZmult b c).
Proof. ir. assert (inc a BZ). nin H0; BZo_tac. assert (inc b BZ). nin H0. 
  BZo_tac. cp (sub_BZps_BZ H). rw BZ_lt_minus. wrr BZmult_minus_distr_l. 
  app BZps_stable_mult. wrr BZ_lt_minus.  app BZ_stable_mult. 
  app BZ_stable_mult.
Qed.


Lemma BZ_prod_increasing3: forall a b c, inc c BZm ->
  BZ_le a b -> BZ_ge  (BZmult a c)  (BZmult b c).
Proof. ir. assert (inc a BZ). BZo_tac. assert (inc b BZ). BZo_tac.
  uf BZ_ge. cp (sub_BZm_BZ H). rw BZ_le_minus. wrr BZmult_minus_distr_l. 
  app BZm_unstable_mult.
  wrr BZ_opp_minus. app BZopp_positive2. wrr BZ_le_minus. 
  app BZ_stable_mult.  app BZ_stable_mult.
Qed.

Lemma BZ_prod_increasing4: forall a b c, inc c BZms ->
  BZ_lt a b -> BZ_gt  (BZmult a c)  (BZmult b c).
Proof. ir. assert (inc a BZ). nin H0; BZo_tac. assert (inc b BZ). nin H0. 
  BZo_tac.  cp (sub_BZms_BZ H). uf BZ_gt. rw BZ_lt_minus. 
  wrr BZmult_minus_distr_l. 
  app BZms_unstable_mult. wrr BZ_opp_minus. app BZopp_positive1. 
  wrr BZ_lt_minus. app BZ_stable_mult.  app BZ_stable_mult.
Qed.

Lemma BZ_prod_increasing5: forall a b c d, inc b BZp -> inc c BZp ->
  BZ_le a b -> BZ_le c d ->
  BZ_le (BZmult a c)  (BZmult b d).
Proof. ir. cp (BZ_prod_increasing1 H0 H1).  cp (BZ_prod_increasing1 H H2).
  rwi BZ_mult_comm H4. rwi (BZ_mult_comm d b)  H4. BZo_tac.
Qed.

Lemma BZ_prod_increasing6: forall a b c d, inc b BZps -> inc c BZps ->
  BZ_lt a b -> BZ_lt c d ->
  BZ_lt (BZmult a c)  (BZmult b d).
Proof. ir. cp (BZ_prod_increasing2 H0 H1).  cp (BZ_prod_increasing2 H H2).
  rwi BZ_mult_comm H4. rwi (BZ_mult_comm d b)  H4. BZo_tac.
Qed.

Lemma injective_P_BZp: forall a b, inc a BZp -> inc b BZp ->
  P a = P b ->  a = b.
Proof. ir. ufi BZp H; ufi BZp H0. awi H. awi H0. ee. app pair_extensionality.
  ue.
Qed.


Lemma BZ_prod_increasing7: forall a b c d, inc a BZp -> inc c BZp ->
  BZ_lt a b -> BZ_lt c d ->
  BZ_lt (BZmult a c)  (BZmult b d).
Proof.  assert (forall a b, inc a BZp -> BZ_lt a b -> inc b BZps).
  ir. wri BZ_positive_pr H.  wr BZ_positive_pr1. BZo_tac.
  ir. cp (H _  _ H0 H2).  cp (H _  _ H1 H3).
  nin (equal_or_not c BZ_zero). rw H6. qw. rw BZ_positive_pr1. 
  app BZps_stable_mult. qprops. app  BZ_prod_increasing6. 
  rw inc_BZps. au.
Qed.

Lemma BZ_prod_increasing8: forall a b c, inc b BZp ->  inc c BZps ->
  BZ_le a b -> 
  BZ_le a  (BZmult b c).
Proof. ir. assert (inc BZ_one BZp). uf BZ_one. app inc_BZ_of_natp. fprops.
  assert (BZ_le  BZ_one c). rw BZ_le_pr1. uf BZ_one. rw BZ_of_nat_P.
  ufi BZps H0. awi H0. ee. ufi Bnats H3. Ztac. app one_small_cardinal.
  fprops. am. app sub_BZps_BZp. cp (BZ_prod_increasing5 H H2 H1 H3).
  qwi H4. am. BZo_tac.
Qed.

Lemma BZ_prod_increasing9: forall a c, inc a BZp ->  inc c BZps ->
  BZ_le a (BZmult a c).
Proof. ir. cp (BZ_le_reflexive (sub_BZp_BZ H)).
  ap (BZ_prod_increasing8 H H0 H1).
Qed.

Lemma BZ_prod_increasing10:  forall a b c, inc b BZp ->  inc c BZps ->
  BZ_lt a b -> 
  BZ_lt a  (BZmult b c).
Proof. ir. assert (inc BZ_one BZp). uf BZ_one. app inc_BZ_of_natp. fprops.
  assert (BZ_le  BZ_one c). rw BZ_le_pr1. uf BZ_one. rw BZ_of_nat_P.
  ufi BZps H0. awi H0. ee. ufi Bnats H3. Ztac. app one_small_cardinal.
  fprops. am. app sub_BZps_BZp. nin (equal_or_not c  BZ_one).
  rw H4. qw. qprops. assert (BZ_lt BZ_one c). split; au.
  assert (inc a BZ). nin H1. BZo_tac. rwi inc_BZ1 H6. nin H6.
  assert (inc b BZps). nin H1. rw inc_BZps.  split. am. ue.
  assert (inc  (BZmult b c) BZps). app BZps_stable_mult. rw H6.
  rww BZ_positive_pr1. nin H6. 
  cp (BZ_prod_increasing7 (sub_BZps_BZp H6) H2 H1 H5). qwi H7. am.
  qprops. wri BZ_negative_pr H6.
  cp (BZp_stable_mult H (sub_BZps_BZp H0)). wri BZ_positive_pr H7. BZo_tac.
Qed.


Lemma BZ_prod_increasing11: forall a b c, inc c BZps ->
  inc a BZ -> inc b BZ ->
  BZ_le a b = BZ_le  (BZmult a c)  (BZmult b c).
Proof. ir. assert (inc c BZp). rwi inc_BZps H. nin H. am. 
  assert (inc c BZ). qprops.
  rww BZ_le_minus. rww BZ_le_minus. 
  wrr BZmult_minus_distr_l. set (d := BZminus b a). ap iff_eq. ir. 
  app BZp_stable_mult. ir. assert (inc d BZ). uf d. app BZ_stable_minus. 
  rwi inc_BZ H5.
  nin H5. cp (BZpms_stable_mult H H5). rwi BZ_mult_comm H6.
  elim (BZ_disjoint_neg_pos H6 H4).  am. 
  app BZ_stable_mult.  app BZ_stable_mult. 
Qed.

Lemma BZ_prod_increasing12: forall a b c, inc c BZps ->
  inc a BZ -> inc b BZ ->
  BZ_lt a b = BZ_lt  (BZmult a c)  (BZmult b c).
Proof.  ir. cp H. assert (inc c BZ). qprops.
  rww BZ_lt_minus. rww BZ_lt_minus. 
  wrr BZmult_minus_distr_l. set (d := BZminus b a). ap iff_eq. ir. 
  app BZps_stable_mult. ir. assert (inc d BZ). uf d.  app BZ_stable_minus. 
  rwi inc_BZ1 H5.
  nin H5. rwi H5 H4. qwi H4. rwi inc_BZps H4. ee. elim H6. tv. am.
  nin H5. am. cp (BZpms_stable_mult H H5). rwi BZ_mult_comm H6.
  elim (BZ_disjoint_neg_spos H6 H4).  app BZ_stable_mult. app BZ_stable_mult. 
Qed.

(** ** Division *)

Definition BZdivision_prop a b q r := 
  a = BZplus (BZmult b q) r & BZ_lt r (BZabs b) & inc r BZp.

Definition BZquo a b :=
  let  q:= BZ_of_nat (card_quo (P a) (P b)) in 
    Yo (b = BZ_zero) BZ_zero
    (Yo (Q a= TPb) (Yo (Q b = TPb) q (BZopp q))
      (Yo (Bnat_divides (P b) (P a)) (Yo (Q b = TPb) (BZopp q) q)
        (Yo (Q b = TPb) (BZopp (BZsucc q)) (BZsucc q)))).

Definition BZrem a b :=
  let r:= BZ_of_nat (card_rem (P a) (P b)) in 
    Yo (b = BZ_zero) BZ_zero
    (Yo (Q a= TPb)  r
      (Yo (Bnat_divides (P b) (P a)) BZ_zero (BZminus (BZabs b) r))).
Definition BZ_divides b a := 
  inc a BZ & inc b BZ & b <> BZ_zero & BZrem a b = BZ_zero.

Lemma BZ_stable_quo: forall a b, inc a BZ -> inc b BZ  ->
  inc (BZquo a b) BZ.
Proof. ir. cp (inc_BZ_P H). cp (inc_BZ_P H0). uf BZquo.
  nin (equal_or_not b BZ_zero). rww Y_if_rw. qprops.  rww Y_if_not_rw.
  assert (P b <> card_zero). dneg. app BZ_zero_if_abs0.
  cp (inc_quotient_bnat H1 H2 H4). cp (inc_BZ_of_nat H5).
  nin (equal_or_not (Q a) TPb). rw Y_if_rw. 
  nin (equal_or_not (Q b) TPb). rww Y_if_rw. rw Y_if_not_rw. qprops. am. am.
  rww Y_if_not_rw. nin (p_or_not_p (Bnat_divides (P b) (P a))).
  rww Y_if_rw.  nin (equal_or_not (Q b) TPb).   rww Y_if_rw. qprops.
  rww Y_if_not_rw. rww Y_if_not_rw. 
  cp (BZ_stable_succ H6). nin (equal_or_not (Q b) TPb).   
  rww Y_if_rw. qprops.  rww Y_if_not_rw.
Qed.

Lemma BZ_stable_rem: forall a b, inc a BZ -> inc b BZ ->
  inc (BZrem a b) BZp.
Proof. ir. cp (inc_BZ_P H). cp (inc_BZ_P H0).  uf BZrem.
  nin (equal_or_not b BZ_zero). rww Y_if_rw. app inc0_BZp. rww Y_if_not_rw.
  assert (P b <> card_zero). dneg. app BZ_zero_if_abs0.
  cp (inc_remainder_bnat H1 H2 H4). cp (inc_BZ_of_natp H5).
  uf BZrem. nin (equal_or_not (Q a) TPb). rw Y_if_rw.  am. am. rww Y_if_not_rw.
  nin (p_or_not_p (Bnat_divides (P b) (P a))). rww Y_if_rw. app inc0_BZp.   
  rww Y_if_not_rw. wr BZ_le_minus. rw BZ_le_pr1. rw BZ_of_nat_P.
  rw BZabs_P. nin (Bnat_division H1 H2 H4). ee. nin H11. nin H12. am. am.
  am. app inc_BZabs_N. qprops. qprops.
Qed.

Lemma BZ_stable_rem1: forall a b, inc a BZ -> inc b BZ -> 
  inc (BZrem a b) BZ.
Proof. ir. cp (BZ_stable_rem H H0). qprops.
Qed.
Lemma BZp_stable_quo: forall x y, inc x BZp -> inc y BZp ->
  inc (BZquo x y) BZp.
Proof. ir. uf BZquo.  nin (equal_or_not y  BZ_zero). rww Y_if_rw.
  app inc0_BZp.  rww Y_if_not_rw. rw (BZp_Q H). rww Y_if_rw.
  rw (BZp_Q H0). rww Y_if_rw. app inc_BZ_of_natp.
  cp (inc_BZ_P (sub_BZp_BZ H)). cp (inc_BZ_P (sub_BZp_BZ H0)). fprops.
  app inc_quotient_bnat. dneg. app BZ_zero_if_abs0. qprops.
Qed.

Hint Resolve BZ_stable_quo BZ_stable_rem  BZ_stable_rem1: qprops.

Lemma BZrem_small: forall a b, inc a BZ -> inc b BZ -> b <> BZ_zero ->
  BZ_lt (BZrem a b) (BZabs b).
Proof. ir. cp (inc_BZ_P H). cp (inc_BZ_P H0).
  assert (P b <> card_zero). dneg. app BZ_zero_if_abs0.
  nin (Bnat_division H2 H3 H4). ee.  red in H7. ee. rw BZ_lt_pr1. rw BZabs_P.
  uf BZrem. rww Y_if_not_rw. nin (equal_or_not (Q a) TPb). rw Y_if_rw. 
  rww BZ_of_nat_P. am. rww Y_if_not_rw.
  nin (p_or_not_p (Bnat_divides (P b) (P a))). rww Y_if_rw. rw BZ_zero_P.  
  split. app zero_smallest. fprops. au. rww Y_if_not_rw. 
  uf BZminus. set (c:= card_rem (P a) (P b)) in *. 
  assert (c = P (BZopp (BZ_of_nat c))).  rw BZopp_P. rw BZ_of_nat_P. tv.
  assert (c <> card_zero). dneg. uf Bnat_divides. am.
  assert (P (BZabs b) = P b). rww BZabs_P.  
  assert (inc (BZabs b) BZp). app inc_BZabs_N.
  assert (inc (BZ_of_nat c) BZps). rw inc_BZps. split. app inc_BZ_of_natp.
  dneg. cp (f_equal P H15). rwi BZ_of_nat_P H16. rwi BZ_zero_P H16. am. 
  cp (BZopp_positive1 H15).
  assert (cardinal_lt (P (BZopp (BZ_of_nat c))) (P (BZabs b))). wr H11. rww H13.
  nin H17. rw (BZ_plus_pm1 H14 H16 H17). wr H11.  rw H13.
  rw BZ_of_nat_P. split. app sub_le_symmetry. nin H8. am. red. ir.
  nin H8. cp (card_sub_pr H3 H5 H8). rwi H19 H21.
  assert (card_plus card_zero (P b) = P b). fprops. app zero_unit_suml.
  fprops. assert (card_plus c (P b) =  card_plus card_zero (P b) ). ue.
  elim H12. app (plus_simplifiable_right H3 H5 inc0_Bnat H23). am. 
  qprops. app inc_BZabs_N.
Qed.


Lemma card_division_exact: forall q b, inc q Bnat -> inc b Bnat ->
  b <> card_zero ->
  (card_quo (card_mult q b) b = q &  card_rem (card_mult q b) b = card_zero).
Proof. ir.  set (a:= card_mult q b). assert (inc a Bnat). uf a. fprops.
  cp (Bnat_division H2 H0 H1). nin H3. nin H4.
  assert (division_prop a b q card_zero).
  red. split. rw zero_unit_sumr. rww card_mult_commutative. fprops. 
  split. app zero_smallest. fprops. au.
  cp (division_unique H2 H0 H4 H3 H inc0_Bnat H1 H5 H6). ee; am.
Qed.

Lemma BZ_division_exact: forall b q, inc q BZ -> inc b BZ -> b <> BZ_zero ->
  (BZquo (BZmult q b) b = q &  BZrem (BZmult q b) b = BZ_zero).
Proof. ir. cp (inc_BZ_P H). cp (inc_BZ_P H0).
  assert (P b <> card_zero). dneg. app BZ_zero_if_abs0. uf BZquo. uf BZrem.
  rw BZ_mult_abs1.
  cp (card_division_exact H2 H3 H4). nin H5. rw H5. rw H6.
  assert (Bnat_divides (P b) (card_mult (P q) (P b))). red. am.
  split. rww Y_if_not_rw. rwi inc_BZ H. nin H. rwi inc_BZ H0. awi H0. nin H0.
  rw Y_if_rw. rw Y_if_not_rw. rw BZopp_pr1. cp (inc_BZms_pr H). eee. am.
  cp (BZms_Q H0). TP_tac. rww BZmult_mm. uf BZ_of_nat. aw. rw Y_if_not_rw. 
  rww Y_if_rw. rw Y_if_rw. rw BZopp_pr1. cp (inc_BZms_pr H). eee. am. app BZp_Q.
  rw BZmult_mp. uf BZm_of_nat. rw Y_if_not_rw. aw. fprops.
  app zero_cardinal_product2. ufi BZms H. awi H. ee. ufi Bnats H8. Ztac. am.
  am. am. nin (equal_or_not (P q) card_zero). rw H8. fold BZ_zero.
  rw BZopp_zero. cp (BZ_zero_if_abs0 (sub_BZp_BZ H) H8).
  assert (Yo (Q b = TPb) BZ_zero BZ_zero = q). rw H9.
  nin (equal_or_not (Q b) TPb). rww Y_if_rw.  rww Y_if_not_rw. rw H10.
  rw H9. qw. rww Y_if_rw.
  rwi inc_BZ H0. nin H0. rw Y_if_not_rw. rww Y_if_rw. rww Y_if_not_rw.
  app inc_BZp_pr. cp (BZms_Q H0). rw H9. fprops.  rw BZmult_pm. uf BZm_of_nat.
  rw Y_if_not_rw. aw. fprops. app zero_cardinal_product2. am. am.
  rww Y_if_rw. rww Y_if_rw. app inc_BZp_pr. rww (BZp_Q H0).
  rw BZmult_pp. uf BZ_of_nat. aw. am. am.
  fold BZ_zero.  rww Y_if_not_rw.  nin (equal_or_not  (Q (BZmult q b)) TPb). 
  rww Y_if_rw. rww Y_if_not_rw.  rww Y_if_rw. am. am.
Qed.


Lemma nat_division_of_zero: forall b, inc b Bnat -> b <> card_zero ->
  (card_quo card_zero b = card_zero & Bnat_divides b card_zero).
Proof. ir.  cp (Bnat_division inc0_Bnat H H0). nin H1. nin H2.
  assert (division_prop card_zero b card_zero card_zero).
  red. split. rw zero_unit_sumr. rw zero_prod_absorbing. tv. fprops. 
  split. app zero_smallest. fprops. au.
  cp (division_unique inc0_Bnat H H2 H1 inc0_Bnat inc0_Bnat  H0 H3 H4). ee; am.
Qed.



Lemma BZquo_non_zero: forall  a b, inc a BZ -> inc b BZ -> b <> BZ_zero ->
  let  q:= BZ_of_nat (card_quo (P a) (P b)) in 
    BZquo a b= 
    (Yo (Q a= TPb) (Yo (Q b = TPb) q (BZopp q))
      (Yo (Bnat_divides (P b) (P a)) (Yo (Q b = TPb) (BZopp q) q)
        (Yo (Q b = TPb) (BZopp (BZsucc q)) (BZsucc q)))).
Proof.  ir. uf BZquo. rww Y_if_not_rw. Qed.

Lemma BZrem_non_zero:  forall  a b, inc a BZ -> inc b BZ -> b <> BZ_zero ->
  let r:= BZ_of_nat (card_rem (P a) (P b)) in 
    BZrem a b = (Yo (Q a= TPb)  r
      (Yo (Bnat_divides (P b) (P a)) BZ_zero (BZminus (BZabs b) r))).
Proof.  ir. uf BZrem. rww Y_if_not_rw. Qed.

Lemma BZdivision_opp_b: forall a b, inc a BZ -> inc b BZ ->
  (BZrem a (BZopp b) = BZrem a b & BZquo a (BZopp b) = BZopp (BZquo a b)).
Proof. ir. nin (equal_or_not b BZ_zero). uf BZrem. rw Y_if_rw. rww Y_if_rw.
  split. tv. uf BZquo.  rw Y_if_rw. rww Y_if_rw. rw BZopp_zero. tv.
  rw H1. rww BZopp_zero. rw H1. rww BZopp_zero. assert (BZopp b <> BZ_zero).
  dneg. app BZ_zero_if_abs0. cp (f_equal P H2).  rwi BZopp_P H3. qwi H3. am. 
  split.
  rw (BZrem_non_zero H H0 H1). rw (BZrem_non_zero H (BZ_stable_opp H0) H2).
  rw BZopp_P. rww BZabs_opp. 
  rw (BZquo_non_zero H H0 H1). rw (BZquo_non_zero H (BZ_stable_opp H0) H2).
  clear H2. rw BZopp_P. set (q:=BZ_of_nat (card_quo (P a) (P b))) in *.
  assert (Ha: inc q BZ). 
  cp (inc_BZ_P H). cp (inc_BZ_P H0).
  assert (P b <> card_zero). dneg. app BZ_zero_if_abs0.
  nin (Bnat_division H2 H3 H4). ee.  red in H7. ee. uf q. app inc_BZ_of_nat.
  cp H. rwi inc_BZ H2. nin H2. assert (Q a = TPa). app BZms_Q. 
  rw Y_if_not_rw. nin (p_or_not_p (Bnat_divides (P b) (P a))).
  rw Y_if_rw. cp H0. rwi inc_BZ H5. nin H5. cp (BZopp_negative1 H5).
  ufi BZps H6. awi H6. ee. rww Y_if_rw. rww Y_if_not_rw. rww Y_if_rw.
  ufi BZms H5. awi H5. ee.  rww Y_if_not_rw. TP_tac. TP_tac. 
  assert (inc b BZps). rw inc_BZps. ee. am. am. cp (BZopp_positive1 H6).
  cp (BZms_Q H7). rww Y_if_not_rw. rww Y_if_not_rw. rww Y_if_rw.
  rww Y_if_rw. sy. app BZopp_involutive. ap (BZp_Q H5).  TP_tac. TP_tac. am.
  rw Y_if_not_rw. cp H0. rwi inc_BZ H5. nin H5. cp (BZopp_negative1 H5).
  cp (BZps_Q H6). rww Y_if_rw. rww Y_if_not_rw. rww Y_if_not_rw.
  cp (BZms_Q H5). rww Y_if_not_rw.  TP_tac. TP_tac. 
  assert (inc b BZps). rw inc_BZps. ee. am. am. cp (BZopp_positive1 H6).
  cp (BZms_Q H7). rww Y_if_not_rw. rww Y_if_not_rw. rww Y_if_not_rw.
  rww Y_if_rw. sy. app BZopp_involutive. uf BZsucc.  app BZ_stable_plus. 
  qprops. app (BZp_Q H5). TP_tac. TP_tac.
  am. TP_tac. cp  (BZp_Q H2). rw Y_if_rw.
  cp H0. rwi inc_BZ H4. nin H4. cp (BZopp_negative1 H4).
  ufi BZps H5. awi H5. ee. rww Y_if_rw. rww Y_if_rw. rww Y_if_not_rw.
  sy. app BZopp_involutive. cp (BZms_Q H4). TP_tac. 
  assert (inc b BZps). rw inc_BZps. ee. am. am. cp (BZopp_positive1 H5).
  cp (BZms_Q H6). rww Y_if_not_rw. rww Y_if_rw. rww Y_if_rw. app (BZp_Q H4). 
  TP_tac. am.
Qed.

Lemma BZdivision_opp_a1: forall a b, inc a BZ -> inc b BZ -> b <> BZ_zero ->
  Bnat_divides (P b) (P a) ->
  (BZrem (BZopp a) b = BZ_zero & BZrem a b = BZ_zero 
    & BZquo (BZopp a) b = BZopp (BZquo a b)).
Proof. ir. cp H2. red in H2. rww BZrem_non_zero. rww BZrem_non_zero.
  rw BZopp_P. rw H2. fold BZ_zero. 
  split.  nin (equal_or_not (Q (BZopp a)) TPb). rww Y_if_rw. 
  rww Y_if_not_rw. rww Y_if_rw. split.  nin (equal_or_not (Q a) TPb).
  rww Y_if_rw.  rww Y_if_not_rw. rww Y_if_rw. rww BZquo_non_zero. 
  rww BZquo_non_zero. rw BZopp_P. set (q:= BZ_of_nat (card_quo (P a) (P b))).
  assert (Ha: inc q BZ). cp (inc_BZ_P H). cp (inc_BZ_P H0).
  assert (P b <> card_zero). dneg. app BZ_zero_if_abs0.
  nin (Bnat_division H4 H5 H6). ee.  red in H9. ee. uf q. app inc_BZ_of_nat.
  cp H. rwi inc_BZ H4. nin H4. assert (Q a = TPa). ufi BZms H4. awi H4. eee.
  assert (Q (BZopp a) = TPb).  cp (BZopp_negative1 H4). ufi BZps H6. awi H6. ee.
  am. rw Y_if_rw.
  cp H0. rwi inc_BZ H7. nin H7. rw Y_if_not_rw. rw Y_if_not_rw. rw Y_if_rw.
  rw Y_if_not_rw. tv. cp (BZms_Q H7). TP_tac. am.  TP_tac.  
  cp (BZms_Q H7). TP_tac. cp (BZp_Q H7). rww Y_if_rw. rww Y_if_not_rw. 
  rww Y_if_rw.  rww Y_if_rw. rw BZopp_involutive. tv. am. TP_tac. am.
  cp (BZp_Q H4). nin (equal_or_not (P a) card_zero). assert (a = BZ_zero). 
  uf BZ_zero.  ufi BZp H4. awi H4. ee.  uf BZ_of_nat. wr H6. wr H5. aw. rw H7.
  rw BZopp_zero. wr H7. rw H5. rw Y_if_rw. assert (q = BZ_zero).
  uf q. rw H6. assert (P b <> card_zero). dneg.  app BZ_zero_if_abs0.
  nin (nat_division_of_zero (inc_BZ_P H0) H8). rww H9. 
  nin (equal_or_not (Q b) TPb).
  rw Y_if_rw. rw H8. rww BZopp_zero. am. rww Y_if_not_rw. rw H8.
  rww BZopp_zero.  rww BZopp_zero. tv.  assert (inc a BZps).
  rw inc_BZps. split. am. dneg. rw H7. rw BZ_zero_P. tv.
  cp (BZopp_positive1 H7). cp (BZms_Q H8). rww Y_if_not_rw. rw Y_if_rw.
  cp H0. rwi inc_BZ H10. nin H10. cp (BZms_Q H10). rw Y_if_not_rw. 
  rww Y_if_rw. rw Y_if_not_rw.
  rww BZopp_involutive.  TP_tac.  TP_tac. cp (BZp_Q H10). 
  rww Y_if_rw. rww Y_if_rw. rww Y_if_rw. am.  TP_tac. qprops. qprops. 
Qed.

Lemma BZdivision_opp_a2: forall a b, inc a BZ -> inc b BZ -> b <> BZ_zero
  -> ~ Bnat_divides (P b) (P a) ->
  BZrem (BZopp a) b = BZminus (BZabs b) (BZrem a b).
Proof. ir. rww BZrem_non_zero. rww BZrem_non_zero. cp (inc_BZ_P H). 
  cp (inc_BZ_P H0). assert (P b <> card_zero). clear H2. dneg. 
  app BZ_zero_if_abs0. fprops. nin (Bnat_division H3 H4 H5). nin H7. 
  red in H8. nin H8. rw BZopp_P.  set (r:= card_rem (P a) (P b)) in *.
  set (q:= card_quo (P a) (P b)) in *.
  set (rr :=BZ_of_nat r). 
  assert (inc rr BZ). uf rr. app inc_BZ_of_nat.
  assert (a <> BZ_zero). red. ir. cp (f_equal P H11). rwi BZ_zero_P H12.
  rwi H12 H2. elim H2. nin (nat_division_of_zero H4 H5).  am. 
  set (c := (Yo (Bnat_divides (P b) (P a)) BZ_zero (BZminus (BZabs b) rr))).
  rwi inc_BZ H. nin H. rw Y_if_rw. rw Y_if_not_rw.  uf c.
  rw Y_if_not_rw. uf BZminus. rw BZ_plus_opp_distr. rw BZ_plus_assoc.
  rw BZopp_involutive. rw BZplus_opp_r. qw. tv. qprops. am.
  qprops. qprops. qprops. qprops. qprops. am. ufi BZms H. awi H. ee. rw H13.
  fprops. cp (BZopp_negative1 H). ufi BZps H12. awi H12. ee. am.
  rw Y_if_not_rw. rw Y_if_rw. uf c. rw Y_if_not_rw. tv. am.
  ufi BZp H. awi H. ee. am.  assert (inc a BZps). rw inc_BZps. ee; am.
  cp (BZopp_positive1 H12). ufi BZms H13.  awi H13. ee. rw H15. fprops. qprops.
Qed.

Lemma BZdivision_opp_a3: forall a b, inc a BZ -> inc b BZps
  -> ~ Bnat_divides (P b) (P a) ->
  BZquo (BZopp a) b = BZopp (BZsucc (BZquo a b)).
Proof. ir. rwi inc_BZps H0. nin H0. rww BZquo_non_zero. rww BZquo_non_zero.
  cp (inc_BZ_P H). cp (inc_BZ_P (sub_BZp_BZ H0)).
  assert (P b <> card_zero). clear H1. dneg. app BZ_zero_if_abs0. qprops.
  nin (Bnat_division H3 H4 H5). nin H7. red in H8. nin H8. uf BZrem. uf BZquo.
  rw BZopp_P.  set (r:= card_rem (P a) (P b)) in *.
  set (q:= card_quo (P a) (P b)) in *. set (rq := BZ_of_nat q).
  assert (inc rq BZ). uf rq. app inc_BZ_of_nat.
  assert (a <> BZ_zero). red. ir. cp (f_equal P H11). rwi BZ_zero_P H12.
  rwi H12 H1. elim H1. nin (nat_division_of_zero H4 H5). am. 
  cp (BZp_Q H0). rwi inc_BZ H. nin H. rw Y_if_rw. rww Y_if_rw. rw Y_if_not_rw.
  rw Y_if_not_rw.  rww Y_if_rw. uf BZsucc. cp inc1_BZ. 
  cp (BZ_stable_plus H10 H13). rw BZ_plus_opp_distr. 
  rw BZopp_involutive. wr BZ_plus_assoc. rw BZplus_opp_r. qw.  am.
  am. am. qprops. am. qprops.  am. am. cp (BZms_Q H).
  TP_tac. cp (BZopp_negative1 H). ap (BZps_Q H13). rw Y_if_not_rw. 
  rw Y_if_not_rw. rww Y_if_rw. rww Y_if_rw. rww Y_if_rw. ap (BZp_Q H). am.
  assert (inc a BZps). rw inc_BZps.  eee. cp (BZopp_positive1 H13).
  cp (BZms_Q H14).  TP_tac. qprops. qprops. qprops.
Qed.

Lemma BZdivision_opp_a4: forall a b, inc a BZ -> inc b BZ -> b <> BZ_zero
  -> ~ Bnat_divides (P b) (P a) ->
  BZquo (BZopp a) b = BZopp (BZplus (BZsign b) (BZquo a b)).
Proof. ir. rwi inc_BZ H0. nin H0. cp (BZopp_negative1 H0).
  cp (sub_BZms_BZ H0). wri (BZopp_P b) H2. cp (BZdivision_opp_a3 H H3 H2).
  cp (BZdivision_opp_b H H4). nin H6. rwi H7 H5.
  cp (BZdivision_opp_b (BZ_stable_opp H) H4). nin H8. rwi H9 H5.
  set (u:= BZquo (BZopp a) b) in *. assert (inc u BZ). uf u. fprops.
  qprops. rw BZ_sign_neg. ufi BZsucc H5.  rwi BZ_plus_comm H5.
  assert (inc  (BZquo a b) BZ). qprops. rw BZ_plus_opp_distr.
  rw BZ_opp_mone. assert (inc  (BZplus BZ_one (BZopp (BZquo a b))) BZ).
  app BZ_stable_plus.  qprops. qprops.
  app BZopp_inj. qprops. qprops. am. assert (inc b BZps). rw inc_BZps. au.
  rww BZdivision_opp_a3. rww BZ_sign_pos. uf BZsucc. rw BZ_plus_comm. tv. 
Qed.

Lemma BZdivision: forall a b, inc a BZ -> inc b BZ -> b <> BZ_zero ->
  (inc (BZquo a b) BZ & inc (BZrem a b) BZp & 
  (BZdivision_prop a b (BZquo a b) (BZrem a b))).
Proof. ir. cp (BZ_stable_rem H H0). cp (BZ_stable_quo H H0).
  cp (BZrem_small H H0 H1). eee. red. eee.
  assert (forall a b, inc a BZp -> inc b BZp ->  b <> BZ_zero ->
    a = BZplus (BZmult b (BZquo a b)) (BZrem a b)).
  ir. rww BZquo_non_zero. rww BZrem_non_zero. assert (Q a0 = TPb). 
  ufi BZp H5. awi H5. ee. am.
  assert (Q b0 = TPb).  ufi BZp H6. awi H6. ee. am.
  cp (inc_BZ_P (sub_BZp_BZ H5)). cp (inc_BZ_P (sub_BZp_BZ H6)).
  assert (P b0 <> card_zero). dneg. app BZ_zero_if_abs0. qprops.
  nin (Bnat_division H10 H11 H12). ee. red in H15. ee.
  rww Y_if_rw. rww Y_if_rw.  rww Y_if_rw.  rw BZmult_pp. rw BZ_of_nat_P.
  rw BZ_plus_pp.  rw BZ_of_nat_P.  rw BZ_of_nat_P.  wr H15. sy. 
  app inc_BZp_pr. app inc_BZ_of_natp. fprops. app inc_BZ_of_natp. am.
  app inc_BZ_of_natp. qprops. qprops. qprops. qprops.
  assert (forall a b, inc a BZp -> inc b BZ ->  b <> BZ_zero ->
    a = BZplus (BZmult b (BZquo a b)) (BZrem a b)).
  ir. rwi inc_BZ H7. nin H7. cp (BZopp_negative1 H7). cp H9.
  rwi inc_BZps H10. nin H10. cp (H5 _ _  H6 H10 H11).
  nin (BZdivision_opp_b (sub_BZp_BZ H6)  (sub_BZms_BZ H7)).
  rwi H13 H12. rwi H14 H12. rwi BZmult_opp_opp H12. am. qprops. qprops. app H5.
  rwi inc_BZ H. nin H.  cp (BZopp_negative1 H). cp (sub_BZps_BZp H7). 
  cp (H6 _ _ H8 H0 H1).
  nin (p_or_not_p (Bnat_divides (P b) (P a))).
  cp (BZdivision_opp_a1 (sub_BZms_BZ H) H0 H1 H10). ee. rw H12.
  rw BZplus_0_r. rwi H11 H9. qwi H9.  rwi H13 H9.
  app BZopp_inj. qprops.  app BZ_stable_mult. rw H9. 
  rww BZopp_mult_distrib_r.  app BZ_stable_mult.   qprops. app BZ_stable_mult. 
  rwi (BZdivision_opp_a2 (sub_BZms_BZ H) H0 H1 H10) H9. 
  rwi (BZdivision_opp_a4 (sub_BZms_BZ H) H0 H1 H10) H9.
  assert (inc (BZsign b) BZ). qprops. rwii BZ_plus_opp_distr H9. 
  assert (inc (BZopp (BZsign b)) BZ). qprops.
  assert (inc (BZopp (BZquo a b)) BZ). qprops. rwii BZmult_plus_distr_r H9. 
  assert (inc (BZabs b) BZ). qprops.
  cp (BZ_stable_mult H0 H3).
  assert (BZmult b (BZopp (BZsign b)) = BZopp (BZabs b)).
  wrr BZopp_mult_distrib_r. rww BZ_sign_abs. rwi H16 H9.
  wrii BZopp_mult_distrib_r H9. set (u := BZopp (BZmult b (BZquo a b))) in H9.
  ufi BZminus H9.  assert (inc u BZ). uf u.  cp (BZ_stable_mult H0 H3). qprops.
  assert (inc (BZplus (BZopp (BZabs b)) u) BZ).  app BZ_stable_plus.  qprops.
  assert (inc (BZopp (BZrem a b)) BZ). qprops. rwii BZ_plus_assoc H9.
  assert (BZplus (BZplus (BZopp (BZabs b)) u) (BZabs b) = u).
  rw (BZ_plus_comm  (BZopp (BZabs b)) u). wrr BZ_plus_assoc.
  rw (BZ_plus_comm  (BZopp (BZabs b)) (BZabs b)). 
  change (BZplus (BZabs b)(BZopp (BZabs b))) with (BZminus (BZabs b) (BZabs b)).
  rw BZminus_diag. qw. qprops. qprops. rwi H20 H9.
  app BZopp_inj. qprops.  app BZ_stable_plus.  qprops. 
  rw H9. rw BZ_plus_opp_distr. tv. qprops. qprops.  app H6.
Qed.


Lemma BZdivision_prop_alt: forall a b q r, inc a BZ -> inc b BZ ->
  inc q BZ -> inc r BZ -> b <> BZ_zero ->
  BZdivision_prop a b q r = (BZ_le (BZmult b q) a
  & BZ_lt a (BZmult b (BZplus (BZsign b) q))
  & r = BZminus a  (BZmult b q)).
Proof. ir. rww BZmult_plus_distr_r. rww BZ_sign_abs.
  uf BZdivision_prop. set (w:= BZmult b q). assert (inc w BZ). uf w. 
  ap (BZ_stable_mult H0 H1).
  rww BZ_le_minus. assert (inc (BZabs b) BZ). qprops. rww  BZ_lt_minus.
  assert (inc (BZplus (BZabs b) w) BZ). app BZ_stable_plus.
  rww  BZ_lt_minus.
  assert (BZminus (BZplus (BZabs b) w) a = BZminus (BZabs b) (BZminus a w)).
  rw BZ_plus_comm. rw BZminus_plus_comm.  rw BZ_plus_comm. 
  wrr (BZ_opp_minus H H4). am. am. am. rw H7.
  ap iff_eq. ir.  nin H8. nin H9.  cp (BZminus_plus H4 H2).  wri H8 H11.
  rw H11. eee. ir. nin H8. nin H9. ee. rw H10. sy. app BZplus_minus. ue. ue.
  qprops.
Qed.

Lemma BZdivision_unique:forall a b q r q' r', inc a BZ -> inc b BZ ->
  inc q BZ -> inc r BZ -> inc q' BZ -> inc r' BZ -> b <> BZ_zero ->
  BZdivision_prop a b q r -> BZdivision_prop a b q' r' ->
  (q = q' & r =r').
Proof. ir. rwi (BZdivision_prop_alt H H0 H1 H2 H5) H6.
  rwi (BZdivision_prop_alt H H0 H3 H4 H5) H7. assert (q = q'). ee.
  cp (BZ_le_lt_trans H6 H8). cp (BZ_le_lt_trans H7 H10).
  assert (inc (BZplus (BZsign b) q') BZ). app BZ_stable_plus. qprops.
  assert (inc (BZplus (BZsign b) q) BZ). app BZ_stable_plus. qprops.
  assert (inc (BZmult b q) BZ). app BZ_stable_mult.
  assert (inc (BZmult b q') BZ). app BZ_stable_mult.
  assert (inc (BZmult b (BZplus (BZsign b) q')) BZ). app BZ_stable_mult.
  assert (inc (BZmult b (BZplus (BZsign b) q)) BZ). app BZ_stable_mult.
  rwii BZ_lt_minus H12.  rwii BZ_lt_minus H13.
  wrii BZmult_minus_distr_r H12. wrii BZmult_minus_distr_r H13.
  assert (inc (BZminus (BZplus (BZsign b) q') q) BZ). app BZ_stable_minus.
  assert (inc (BZminus (BZplus (BZsign b) q) q') BZ). app BZ_stable_minus.
  cp (BZps_stable_mult1 H0 H20 H12). cp (BZps_stable_mult1 H0 H21 H13).
  cp H0. rwi inc_BZ1 H24. nin H24. contradiction. nin H24. cp H24. cp H24. ee.
  rwi H22 H24. rwi  H23 H25. wrii BZ_lt_minus H24.  wrii BZ_lt_minus H25. 
  rwii BZ_sign_pos H24.  rwii BZ_sign_pos H25. 
  rwi BZ_plus_comm H24. change (BZ_lt q (BZsucc q')) in H24. 
  rwi BZ_lt_succ1 H24.
  rwi BZ_plus_comm H25. change (BZ_lt q' (BZsucc q)) in H25. 
  rwi BZ_lt_succ1 H25. BZo_tac. am. am. am. am. 
  ee. cp H24. cp H24. rwi H26 H24. rwi H25 H27.
  wrii BZ_lt_minus2 H24.  wrii BZ_lt_minus2 H27. 
  rwii BZ_sign_neg H24.  rwii BZ_sign_neg H27.
  set (s':= BZplus BZ_mone q') in *.   set (s:= BZplus BZ_mone q).
  assert (q = BZsucc s). uf s. rw BZ_plus_comm. wr BZ_opp_one.
  change (q = BZsucc (BZpred q)). rww BZ_succ_pred. rwi H29 H24. 
  rwi BZ_lt_succ1 H24.
  assert (q' = BZsucc s'). uf s'. rw BZ_plus_comm. wr BZ_opp_one.
  change (q' = BZsucc (BZpred q')). rww BZ_succ_pred. rwi H30 H27. 
  rwi BZ_lt_succ1 H27. fold s in H27. assert (s = s'). BZo_tac.
  rw H29. rw H30. rww H31. app BZ_stable_plus. qprops. uf s'.
  app BZ_stable_plus. qprops. uf s'. app BZ_stable_plus. qprops. uf s. 
  app BZ_stable_plus. qprops. split. am. ee. rw H12. rw H10; rww H8.
Qed.

Lemma BZ_division_zero: forall b, inc b BZ ->
  (BZquo BZ_zero b = BZ_zero &  BZrem BZ_zero b = BZ_zero).
Proof. ir. nin (equal_or_not b BZ_zero).  rw H0.  split.
  uf BZquo. rww Y_if_rw. uf BZrem.  rww Y_if_rw.
  cp (BZ_division_exact inc0_BZ H H0). qwi H1.  am.  am. 
Qed.

Lemma BZdivision_unique1:forall a b q r, inc a BZ -> inc b BZ ->
  inc q BZ -> inc r BZ -> b <> BZ_zero ->
  BZdivision_prop a b q r ->
  (q = BZquo a b & r  = BZrem a b).
Proof. ir. cp (BZdivision H H0 H3). nin H5. nin H6.
  cp (BZdivision_unique H H0 H1 H2 H5 (sub_BZp_BZ H6) H3 H4 H7). am.
Qed.



Lemma BZdivides_pr: forall a b,
  BZ_divides b a -> a = BZmult b  (BZquo a b).
Proof. ir. red in H. ee. cp (BZdivision H H0 H1). ee. nin H5. ee. rwi H2 H5.
  rwii BZplus_0_r H5. app BZ_stable_mult. 
Qed.

Lemma BZdivides_pr1: forall a b, inc a BZ -> inc b BZ ->
  b <> BZ_zero ->  BZ_divides b (BZmult b a).
Proof. ir. cp (BZ_division_exact H H0 H1). eee. uf BZ_divides.
  rww BZ_mult_comm. eee. app BZ_stable_mult.
Qed.

Lemma BZdivides_pr2: forall a b q, inc q BZ -> inc b BZ ->
  b <> BZ_zero -> a = BZmult b q -> q = BZquo a b. 
Proof. ir. assert (BZdivision_prop a b q BZ_zero). split.
  rww BZplus_0_r. app BZ_stable_mult. split. app  non_zero_abs. app inc0_BZp.
  assert (inc  (BZmult b q) BZ). app BZ_stable_mult.  wri H2 H4.
  nin (BZdivision_unique1 H4 H0 H inc0_BZ H1 H3). am.
Qed.

Lemma BZdivides_pr0: forall a b, 
  BZ_divides b a ->  Bnat_divides (P b) (P a).
Proof. uf BZ_divides. uf BZrem. uf Bnat_divides. ir. ee.
  set (r := card_rem (P a) (P b)) in *. 
  rwi Y_if_not_rw H2. rwi inc_BZ H. nin H. rwi Y_if_not_rw H2. 
  nin (equal_or_not r card_zero).
  am. rwi Y_if_not_rw H2.  cp (inc_BZ_P (sub_BZms_BZ H)). cp (inc_BZ_P H0).
  assert (P b <> card_zero). clear H3. dneg. app BZ_zero_if_abs0. 
  cp (Bnat_division H4 H5 H6). ee. nin H9. fold r in H10.
  assert (inc (BZ_of_nat r) BZ). app inc_BZ_of_nat.
  cp (BZ_minus_diag_rw (BZ_stable_abs H0) H11 H2). cp (f_equal P H12).
  rwi BZabs_P H13. rwi BZ_of_nat_P H13. nin H10. elim H14; au. am. am.
  cp (BZms_Q H). TP_tac. rwi Y_if_rw H2.  cp (f_equal P H2).
  rwi BZ_of_nat_P H3. qwi H3. am. ap (BZp_Q H). am. 
Qed.

Lemma BZdivides_pr3: forall a b, inc a BZ -> inc b BZ -> b <> BZ_zero ->
  BZ_divides b a =  Bnat_divides (P b) (P a).
Proof. ir. app iff_eq. ir. app  BZdivides_pr0.
  uf BZ_divides. uf BZrem. uf Bnat_divides. ir. eee.
  rww  Y_if_not_rw. rw H2.
  nin (equal_or_not (Q a) TPb). rww Y_if_rw. 
  rww Y_if_not_rw.  rww Y_if_rw.
Qed.

Lemma BZnon_zero_opp: forall x,
  inc x BZ -> (x <> BZ_zero) = (BZopp x <> BZ_zero).
Proof. ir.  app iff_eq. ir. dneg. app BZ_zero_if_abs0. 
  cp (f_equal P H1). rwii BZopp_P H2. qwi H2. am. ir. dneg. rw H1. qw.
Qed.

Lemma BZ_divides_opp1: forall a b, inc b BZ ->
  BZ_divides b a = BZ_divides (BZopp b) a.
Proof. ir. cp (BZ_stable_opp H).
  app iff_eq; ir; cp H1; nin H1; ee; rwii BZdivides_pr3 H2; rww BZdivides_pr3.
  rww BZopp_P. wrr BZnon_zero_opp.  rwi BZopp_P H2. am. rww BZnon_zero_opp.
Qed.

Lemma BZ_divides_opp2: forall a b, inc a BZ ->
  BZ_divides b a = BZ_divides b (BZopp a).
Proof. ir. cp (BZ_stable_opp H).
  app iff_eq; ir; cp H1; nin H1; ee; rwii BZdivides_pr3 H2; rww BZdivides_pr3.
  rww BZopp_P. rwi BZopp_P H2. am.
Qed.

Lemma BZ_quotient_opp2: forall a b, 
  BZ_divides b a -> BZquo (BZopp a) b = BZopp (BZquo a b).
Proof. ir. cp (BZdivides_pr0 H). red in H. ee.
  cp (BZdivision_opp_a1 H H1 H2 H0). ee. am.
Qed.

Lemma BZone_divides_all: forall a, inc a BZ -> BZ_divides BZ_one a.
Proof. ir. cp BZ_one_not_zero. assert (inc BZ_one BZ). qprops.
  cp (BZdivides_pr1 H H1 H0).  qwii H2.
Qed.

Lemma BZmone_divides_all: forall a, inc a BZ -> BZ_divides BZ_mone a.
Proof. ir. wr BZ_opp_one. wr BZ_divides_opp1. app  BZone_divides_all. 
 qprops.
Qed.

Lemma BZquotient_by_one: forall a, inc a BZ -> BZquo a BZ_one = a.
Proof. ir. sy. app BZdivides_pr2. qprops. app  BZ_one_not_zero. qw.
Qed.

Lemma BZquotient_by_mone: forall a, inc a BZ -> BZquo a BZ_mone = BZopp a.
Proof. ir. sy. app BZdivides_pr2. qprops. qprops. app  BZ_mone_not_zero.
  wr BZ_opp_one. rww BZmult_opp_opp. qw. qprops.
Qed.

Lemma BZdivides_pr4: forall a b q, 
  BZ_divides b a -> q = BZquo a b -> a = BZmult b q.
Proof. ir. cp (BZdivides_pr H). ue.
Qed.

Lemma BZdivides_pr5: forall b q,  inc b BZ -> inc q BZ -> b <> BZ_zero ->
  BZquo (BZmult b q) b = q.
Proof. ir. sy. app BZdivides_pr2.  
Qed.

Lemma BZdivision_itself: forall a, inc a BZ -> a <> BZ_zero ->
  (BZ_divides a a & BZquo a a = BZ_one).
Proof. ir. assert (a = BZmult a BZ_one). qw.  ee. 
  cp (BZdivides_pr1 inc1_BZ H H0). wrii H1 H2. 
  cp (BZdivides_pr5 H inc1_BZ H0). wrii H1 H2. 
Qed.

Lemma BZdivision_opp: forall a, inc a BZ -> a <> BZ_zero ->
  (BZ_divides a (BZopp a) & BZquo (BZopp a) a = BZ_mone).
Proof. ir. cp (BZ_mult_m1_r H).
  cp (BZdivides_pr1 inc_m1_BZ H H0). rwi H1 H2. 
  cp (BZdivides_pr5 H inc_m1_BZ H0). rwi H1 H3.  au.
Qed.

Lemma BZdivides_itself: forall a,  inc a BZ -> a <> BZ_zero ->
  BZ_divides a a.
Proof. ir. nin (BZdivision_itself H H0). am.
Qed.
  
Lemma BZquo_itself: forall a,  inc a BZ -> a <> BZ_zero ->
  BZquo a a = BZ_one.
Proof. ir. nin (BZdivision_itself H H0). am.
Qed.

Lemma BZ_division_zero1: forall a, inc a BZ -> a <> BZ_zero ->
  (BZ_divides a BZ_zero & BZquo BZ_zero a = BZ_zero).
Proof. ir. nin (BZ_division_zero H). split. red. eee; qprops. am.
Qed.

Lemma BZ_divides_trans: forall a b a', 
  BZ_divides a a'-> BZ_divides b a -> BZ_divides b a'.
Proof. ir. cp (BZdivides_pr H).  cp (BZdivides_pr H0).
  rwi H2 H1.  red in H; red in H0. ee. assert (inc (BZquo a b) BZ). qprops.
  wri BZ_mult_assoc H1. rw H1. app BZdivides_pr1. app BZ_stable_mult.
  wr H2. qprops. am. am. wr H2. qprops.
Qed.

Lemma BZ_divides_trans1: forall a b a', 
  BZ_divides a a'-> BZ_divides b a ->
  BZquo a' b =  BZmult (BZquo a' a)  (BZquo a b).
Proof. ir.  cp (BZdivides_pr H).  cp (BZdivides_pr H0).
  rwi H2 H1. red in H; red in H0. ee.
  assert (inc (BZquo a' a) BZ). qprops. assert (inc (BZquo a b) BZ). qprops.
  wri BZ_mult_assoc H1.  wri H2 H1.
  assert (inc (BZmult (BZquo a b) (BZquo a' a)) BZ).
  app BZ_stable_mult.  cp (BZdivides_pr2 H11 H3 H4 H1).  sy.
  rww BZ_mult_comm.  am. am. wr H2. am.
Qed.

Lemma BZ_divides_trans2: forall a b c, inc c BZ ->
  BZ_divides b a -> BZ_divides b (BZmult a c).
Proof. ir. nin (equal_or_not a BZ_zero). rw H1. qw. ue.
  assert (inc a BZ).  red in H0. ee. am.
  cp (BZdivides_pr1 H H2 H1). 
  app (BZ_divides_trans H3 H0).  
Qed.

Lemma BZquo_simplify:  forall a b c, inc a BZ -> inc b BZ -> inc c BZps ->
   b <> BZ_zero ->
   (BZquo (BZmult a c) (BZmult b c) = BZquo a b &
     BZrem (BZmult a c) (BZmult b c) = BZmult (BZrem a b) c).
Proof. ir. cp (BZdivision H H0 H2). nin H3. nin H4. nin H5.
  assert (BZmult a c = 
    BZplus (BZmult (BZmult b c) (BZquo a b)) (BZmult (BZrem a b) c)).
  rw (BZ_mult_comm b c). wrr BZ_mult_assoc. 
  rw (BZ_mult_comm c (BZmult b (BZquo a b))).
  wr BZmult_plus_distr_l. wrr H5.  qprops. app BZ_stable_mult. qprops. 
  qprops. assert (Ha:= sub_BZps_BZ H1).
  assert (BZ_lt (BZmult (BZrem a b) (BZabs c)) (BZabs (BZmult b c))).
  rw BZ_mult_abs. rw (BZ_abs_pos (sub_BZps_BZp H1)). 
  app BZ_prod_increasing2. nin H6. am. am. am.
  assert (inc (BZmult a c) BZ). app BZ_stable_mult.
  assert (inc (BZmult b c) BZ). app BZ_stable_mult.
  assert (inc (BZmult (BZrem a b) c) BZ). app BZ_stable_mult. qprops.
  assert (BZmult b c <> BZ_zero). app BZmult_0_rl. rwi inc_BZps H1.
  nin H1. am. assert(BZdivision_prop  (BZmult a c) (BZmult b c) (BZquo a b)
  (BZmult (BZrem a b) c)). split. am. split. 
  rwi (BZ_abs_pos (sub_BZps_BZp H1)) H8. am. app BZp_stable_mult.
  qprops. 
  nin (BZdivision_unique1 H9 H10 H3 H11 H12 H13). split. sy; am.
  sy; am.  
Qed.

Lemma BZ_divides_and_sum:  forall a a' b, 
  BZ_divides b a -> BZ_divides b a'
  -> (BZ_divides b  (BZplus a a') & 
    BZquo  (BZplus a a')  b =  BZplus(BZquo a b)  (BZquo a' b)).
Proof. ir. cp (BZdivides_pr H). cp (BZdivides_pr H0).
  red in H; red in H0.  nin H. nin H3. nin H0. nin H5. nin H4.
  assert (BZplus a  a' = BZmult b  (BZplus(BZquo a b)(BZquo a' b))). 
  rw BZmult_plus_distr_r. wr H1; wrr H2. qprops. qprops. qprops.
  split. rw H8.  app BZdivides_pr1. app BZ_stable_plus. qprops. qprops.
  sy. app BZdivides_pr2. app BZ_stable_plus. qprops. qprops.
Qed.
 
Lemma BZ_divides_and_difference:  forall a a' b,
  BZ_divides b a -> BZ_divides b a'
  -> (BZ_divides b  (BZminus a a') & 
    BZquo  (BZminus a a')  b =  BZminus (BZquo a b)  (BZquo a' b)).
Proof. ir. uf BZminus. cp H0. rwii BZ_divides_opp2 H1. 
  cp (BZ_divides_and_sum H H1). ee. am. rw H3. rww BZ_quotient_opp2.
  red in H0; eee; qprops.
Qed.

Definition BZ_ideal x:= 
  (forall a, inc a x -> inc a BZ)
  & (forall a b, inc a x -> inc b x -> inc (BZplus a b) x)
  & (forall a b, inc a x -> inc b BZ -> inc (BZmult a b) x).

Definition BZ_ideal2 a b := Zo BZ (fun z =>
  exists u, exists v, inc u BZ & inc v BZ &
  z = BZplus (BZmult a u) (BZmult b v)).

Lemma BZ_in_ideal1: forall a b, inc a BZ -> inc b BZ ->
  (inc a (BZ_ideal2 a b) & inc b (BZ_ideal2 a b)).
Proof. ir. split. uf BZ_ideal2. Ztac. exists BZ_one. exists BZ_zero. 
  eee; qprops. qw.  uf BZ_ideal2. Ztac. exists BZ_zero. exists BZ_one.
  eee; qprops. qw.
Qed.

Lemma BZ_plus_permute_2_in_4: forall a b c d, 
  inc a BZ -> inc b BZ -> inc c BZ -> inc d BZ ->
  BZplus (BZplus a b)  (BZplus c d) = BZplus (BZplus a c)  (BZplus b d).
Proof. ir. rww BZ_plus_assoc.  rww BZ_plus_assoc.
  cut (BZplus (BZplus a b) c =  (BZplus (BZplus a c) b)). ir. rww H3.
  wrr BZ_plus_assoc.  wrr BZ_plus_assoc.  rw (BZ_plus_comm b c). tv.
  app BZ_stable_plus. app BZ_stable_plus.
Qed.



Lemma BZ_mult_permute_2_in_4: forall a b c d, 
  inc a BZ -> inc b BZ -> inc c BZ -> inc d BZ ->
  BZmult (BZmult a b)  (BZmult c d) = BZmult (BZmult a c)  (BZmult b d).
Proof. ir. rww BZ_mult_assoc.  rww BZ_mult_assoc.
  cut (BZmult (BZmult a b) c =  (BZmult (BZmult a c) b)). ir. rww H3.
  wrr BZ_mult_assoc.  wrr BZ_mult_assoc.  rw (BZ_mult_comm b c). tv.
  app BZ_stable_mult. app BZ_stable_mult.
Qed.



Lemma BZ_is_ideal2: forall a b, inc a BZ -> inc b BZ -> 
  BZ_ideal (BZ_ideal2 a b).
Proof. ir. split. ir. ufi BZ_ideal2 H1.  Ztac. am. split.
  uf BZ_ideal2. ir. cp (Z_all H2). clear H2. cp (Z_all H1). clear H1. ee.
  nin H4. nin H4. nin H2. nin H2. ee. Ztac. 
  app BZ_stable_plus. exists (BZplus x x1). exists (BZplus x0 x2). ee.
  app BZ_stable_plus. app BZ_stable_plus. rw H8; rw H6. 
  rww BZmult_plus_distr_r. rww BZmult_plus_distr_r. rww BZ_plus_permute_2_in_4.
  rw  (BZ_plus_comm (BZmult a x1) (BZmult a x)).
  rw  (BZ_plus_comm (BZmult b x0) (BZmult b x2)). tv.
  app BZ_stable_mult. app BZ_stable_mult.
  app BZ_stable_mult. app BZ_stable_mult.

  uf BZ_ideal2. ir. cp (Z_all H1). clear H1. nin H3. nin H3. nin H3. 
  ee. Ztac. app BZ_stable_mult.
  exists (BZmult x b0).  exists (BZmult x0 b0). ee. app BZ_stable_mult.
  app BZ_stable_mult. rww BZ_mult_assoc. rww BZ_mult_assoc.  
  wr BZmult_plus_distr_l. ue. am.  app BZ_stable_mult. app BZ_stable_mult.
Qed.

(** ** Ideals *)

Definition BZ_ideal1 a := Zo BZ (fun z => exists u, inc u BZ & z = BZmult a u).

Lemma BZ_in_ideal3: forall a, inc a BZ ->  BZ_ideal1 a = BZ_ideal2 a a.
Proof. ir. uf BZ_ideal1; uf BZ_ideal2. set_extens.  Ztac. nin H2. clear H0. ee.
  Ztac. exists x0. exists BZ_zero. eee; qprops. qw.  app BZ_stable_mult.
  Ztac. clear H0. nin H2. nin H0. ee. wri BZmult_plus_distr_r H3.
  Ztac. exists  (BZplus x0 x1). ee.  app BZ_stable_plus.  am. am. am. am.
Qed.

Lemma BZ_in_ideal4: forall a, inc a BZ -> 
  (BZ_ideal (BZ_ideal1 a) & inc a (BZ_ideal1 a)).
Proof. ir.  rw BZ_in_ideal3. split. app  BZ_is_ideal2.
  nin (BZ_in_ideal1 H H). am. am.
Qed.

Lemma BZ_ideal_stable_opp: forall a x, BZ_ideal x -> inc a x ->
  inc (BZopp a) x.
Proof. ir.  wr BZ_mult_m1_r. red in H. ee. app H2. qprops.  nin H. app H.
Qed.

Lemma BZ_ideal_stable_abs: forall a x, BZ_ideal x -> inc a x ->
  inc (BZabs a) x.
Proof. ir. assert (inc a BZ). nin H. ap H. am.
  rwi inc_BZ H1. nin H1. rw BZ_abs_neg. app  BZ_ideal_stable_opp. am.
  rww BZ_abs_pos. 
Qed.

Lemma BZ_ideal_stable_minus: forall a b x, BZ_ideal x -> inc a x ->  inc b x ->
  inc (BZminus a b) x.
Proof. ir. uf BZminus. cp H. red in H. ee. app H3. app  BZ_ideal_stable_opp.
Qed.
 
Lemma BZ_ideal_stable_rem: forall a b x, BZ_ideal x -> inc a x ->  inc b x ->
  b <> BZ_zero  -> inc (BZrem a b) x.
Proof. ir. cp H. red in H. ee. cp (BZdivision (H _ H0) (H _ H1) H2).
  nin H6. nin H7. nin H8.
  set (c:= BZmult b (BZquo a b)). assert (inc c x). uf c. app H5.
  rw (BZplus_minus_ea (H _ H0)  (H _ H10) (sub_BZp_BZ H7) H8).
  app  BZ_ideal_stable_minus.
Qed.

Lemma BZ_ideal_0: forall a, inc a (BZ_ideal1 BZ_zero) = (a = BZ_zero).
Proof. ir. uf BZ_ideal1. app iff_eq; ir; Ztac.  nin H1. nin H1.
  rwii BZmult_0_l H2. rw H. qprops. exists BZ_one. split. qprops.
  rw H. qw. qprops.
Qed.

Lemma BZ_N_worder: forall X, sub X BZp -> nonempty X ->
  exists a, inc a X & forall b, inc b X -> BZ_le a b.
Proof. ir. set (Y:= fun_image X P). assert (sub Y Bnat). red. ir.
  ufi Y H1. awi H1. nin H1. ee. wr H2. assert (inc x0 BZ).
  app sub_BZp_BZ. app H. app inc_BZ_P. assert (nonempty Y). nin H0.
  exists (P y).  uf Y. aw. ex_tac. cp Bnat_order_worder. nin H3.
  rwi Bnat_order_substrate H4. nin (H4 _  H1 H2). nin H5.
  assert (sub Y (substrate  Bnat_order)). rw Bnat_order_substrate. fprops.
  awi H5. awi H6. assert (forall a, inc a Y -> cardinal_le x a). ir.
  cp (H6 _ H8). awii H9. rwi Bnat_order_le H9. red in H9. ee. am.
  exists (BZ_of_nat x). ee. ufi Y H5. awi H5. nin H5.  nin H5.
  assert (BZ_of_nat x = x0). wr H9. app inc_BZp_pr. app H. ue.
  ir. rw BZ_le_pr1. rw BZ_of_nat_P. app H8. uf Y. aw. ex_tac.
  app inc_BZ_of_natp. app H1. app H. am. am. am. am.
Qed.


Theorem BZ_principal: forall x, BZ_ideal x -> nonempty x ->
  exists a, (inc a BZp & BZ_ideal1 a = x).
Proof. ir. nin (p_or_not_p (exists a, inc a x & a <> BZ_zero)). 
  nin H1. ee. set (Z:= intersection2 BZps x).
  assert (nonempty Z). cp (BZ_ideal_stable_abs H H1).
  assert (inc x0 BZ). nin H. app H. 
  exists (BZabs x0). uf Z. app intersection2_inc. rw inc_BZps. split.
  app inc_BZabs_N.  dneg. app BZ_zero_if_abs0. cp (f_equal P H5).
  rwi BZabs_P H6. qwi H6. am. am. assert (sub Z BZp).
  apply sub_trans with BZps.  uf Z. app intersection2sub_first. qprops.
  nin (BZ_N_worder H4 H3). nin H5. cp (H4 _ H5). 
  assert (inc x1 x). ufi Z H5. cp (intersection2_second H5). am.
  assert (x1 <> BZ_zero). ufi Z H5. cp (intersection2_first H5). 
  rwi inc_BZps H9. ee; am.
  exists x1. ee. am. sy. set_extens. assert (inc x2 BZ). nin H. qprops.
  cp (BZdivision H11 (sub_BZp_BZ H7) H9). ee. nin H14. nin H15.
  cp (BZ_ideal_stable_rem H H10 H8 H9).  rwi BZ_abs_pos H15.
  nin (equal_or_not (BZrem x2 x1) BZ_zero). rwi H18 H14.
  rwi BZplus_0_r H14. rw H14. uf BZ_ideal1. Ztac.  app BZ_stable_mult.
  qprops. exists (BZquo x2 x1).
  eee.  app BZ_stable_mult. qprops. assert (inc (BZrem x2 x1) Z). uf Z. 
  app intersection2_inc.
  rw inc_BZps. split. am. am. cp (H6 _ H19). BZo_tac. am. 
  ufi BZ_ideal1 H10. Ztac. nin H12. ee. rw H13.  red in H. ee. app H15.
  exists BZ_zero. nin H0. split. app inc0_BZp. sy. set_extens.
  uf BZ_ideal1. Ztac.  nin H. app H. exists BZ_one. split. qprops.
  qw.  nin (equal_or_not x0 BZ_zero). am. elim H1. exists x0. au.
  qprops. ufi BZ_ideal1 H2. Ztac. nin H4. nin H4.
  assert (x0 =  BZmult y BZ_zero). rw H5. qw.  
  nin H.  app H. red in H. ee.  rw H6. app H8.  qprops.
Qed.


Lemma BZmult_1_inversion_more: forall a b c,
  inc a BZ -> inc b BZ -> inc c BZ -> a = BZmult a (BZmult b c) ->
  a = BZ_zero \/ b = BZ_one \/ b = BZ_mone.
Proof. ir. assert (a = BZmult a BZ_one). qw.
  assert (BZminus (BZmult a BZ_one) (BZmult a (BZmult b c)) = BZ_zero).
  wr H2. wr H3. app BZminus_diag. 
  assert (inc (BZmult b c) BZ).  app BZ_stable_mult.
  wrii BZmult_minus_distr_r H4.
  nin (equal_or_not a BZ_zero). au. right.
  nin (equal_or_not  (BZminus BZ_one (BZmult b c)) BZ_zero).
  cp (BZ_minus_inj inc1_BZ H5 H7).
  symmetry in H8. nin (BZmult_1_inversion_l H0 H1 H8). am.
  assert (inc (BZminus BZ_one (BZmult b c)) BZ). 
  app BZ_stable_minus.  qprops. 
  cp (BZmult_0_rl H H8 H6 H7). contradiction. qprops. 
Qed.


Lemma BZ_ideal_unique_generator: forall a b, inc a BZ -> inc b BZ ->
  BZ_ideal1 a = BZ_ideal1 b -> BZabs a = BZabs b.
Proof. ir.  cp (BZ_in_ideal4 H).  cp (BZ_in_ideal4 H0). ee.
  rwi H1 H5. wri H1 H4. ufi BZ_ideal1 H5. ufi BZ_ideal1 H4. Ztac. clear H4.
  Ztac. nin H7. nin H8. ee. rwi H10  H9. wri BZ_mult_assoc H9. rw H10.
  nin (BZmult_1_inversion_more H H7 H8 H9). rw H11. qw.  nin H11. rw H11.
  qw. rw H11. rww BZ_mult_m1_r. rww BZabs_opp. am. am. am.
Qed.

(** ** Ideals *)

Definition BZ_gcd a b := choose (fun z => 
  inc z BZp & BZ_ideal1 z = BZ_ideal2 a b).
Definition BZ_lcm a b := BZquo (BZmult a b) (BZ_gcd a b).
Definition BZ_coprime a b :=  BZ_gcd a b = BZ_one.

Lemma BZ_gcd_prop1: forall a b, inc a BZ -> inc b BZ ->
  (inc (BZ_gcd a b) BZp & BZ_ideal1  (BZ_gcd a b)  = BZ_ideal2 a b).
Proof. ir. uf BZ_gcd. app choose_pr. app BZ_principal. app BZ_is_ideal2.
  exists a. nin (BZ_in_ideal1 H H0). am.
Qed.

Lemma BZ_stable_gcd: forall a b, inc a BZ -> inc b BZ ->
  inc (BZ_gcd a b) BZp.
Proof. ir. nin (BZ_gcd_prop1 H H0). am. Qed.

Hint Resolve  BZ_stable_gcd : qprops.

Lemma BZ_gcd_diva: forall a b, inc a BZ -> inc b BZ ->
  a =  BZmult (BZ_gcd a b) (BZquo a  (BZ_gcd a b)).
Proof. ir. cp (BZ_gcd_prop1 H H0). nin H1. nin (BZ_in_ideal1 H H0). wri H2 H3.
  ufi BZ_ideal1 H3. Ztac. nin H6. ee.
  nin (equal_or_not  (BZ_gcd a b) BZ_zero). rw H8. qw. rwi H8 H7. qwi H7. am.
  am. qprops. app BZdivides_pr. set (u := BZ_gcd a b) in *. rw H7.
  app  BZdivides_pr1. uf u. qprops.
Qed.

Lemma BZ_gcd_divb: forall a b, inc a BZ -> inc b BZ ->
  b =  BZmult (BZ_gcd a b) (BZquo b (BZ_gcd a b)).
Proof. ir. cp (BZ_gcd_prop1 H H0). nin H1. nin (BZ_in_ideal1 H H0). wri H2 H4.
  ufi BZ_ideal1 H4. Ztac. nin H6. ee.
  nin (equal_or_not  (BZ_gcd a b) BZ_zero). rw H8. qw. rwi H8 H7. qwi H7. am.
  am. qprops. app BZdivides_pr. set (u := BZ_gcd a b) in *. rw H7.
  app  BZdivides_pr1. uf u. qprops.
Qed.

Lemma BZ_gcd_nonzero: forall a b, inc a BZ ->  inc b BZ -> 
  BZ_gcd a b = BZ_zero ->  (a = BZ_zero & b = BZ_zero).
Proof. ir. cp (BZ_gcd_divb H H0). cp (BZ_gcd_diva H H0). rwi H1 H2.
  rwi H1 H3. qwi H2. qwi H3. au. qprops. qprops.
Qed.

Lemma BZ_gcd_nonzero1: forall a b, inc a BZ ->  inc b BZ -> 
  (a <> BZ_zero \/ b <> BZ_zero) -> BZ_gcd a b <> BZ_zero.
Proof. ir. red. ir. nin (BZ_gcd_nonzero H H0 H2). nin H1; contradiction.
Qed.

Lemma BZ_gcd_prop2: forall a b, inc a BZ -> inc b BZ ->
  (exists x, exists y, inc x BZ & inc y BZ  & 
    (BZ_gcd a b = BZplus (BZmult a x) (BZmult b y))).
Proof. ir. cp (BZ_gcd_prop1 H H0). ee. 
  nin (BZ_in_ideal4 (sub_BZp_BZ H1)). rwi H2 H4. ufi BZ_ideal2 H4. Ztac. am.
Qed.

Lemma BZ_gcd_comm: forall a b, inc a BZ -> inc b BZ ->
  BZ_gcd a b = BZ_gcd b a.
Proof. ir. assert (BZ_ideal2 a b = BZ_ideal2 b a). 
  uf BZ_ideal2. set_extens; Ztac; clear H1; Ztac; nin H3; nin H1;
  exists  x1; exists x0; eee; app BZ_plus_comm.
  uf BZ_gcd.  rww H1. 
Qed.

Lemma BZ_gcd_opp: forall a b, inc a BZ -> inc b BZ ->
  BZ_gcd a b = BZ_gcd (BZopp a) b.
Proof. ir. assert (BZ_ideal2 a b = BZ_ideal2  (BZopp a) b).
  uf BZ_ideal2. set_extens; Ztac; clear H1; Ztac; nin H3; nin H1;
  exists  (BZopp x0); exists x1; eee; qprops. 
  rww BZmult_opp_opp. rww BZmult_opp_comm.
  uf BZ_gcd. rww H1. 
Qed.

Lemma BZ_gcd_abs: forall a b, inc a BZ -> inc b BZ ->
  BZ_gcd a b = BZ_gcd (BZabs a) (BZabs b).
Proof. ir.  assert (forall a b, inc a BZ -> inc b BZ ->
  BZ_gcd a b = BZ_gcd (BZabs a) b).
  ir.  rwi inc_BZ H1. nin H1. rw BZ_abs_neg. app BZ_gcd_opp. qprops. am.
  rww BZ_abs_pos. transitivity (BZ_gcd (BZabs a) b). app H1.
  assert (inc (BZabs a) BZ). qprops.
  assert (inc (BZabs b) BZ). qprops.
  rw (BZ_gcd_comm H2 H0).  rw (BZ_gcd_comm H2 H3). app H1.
Qed.

Lemma BZ_gcd_zero: forall a, inc a BZ -> BZ_gcd a BZ_zero = BZabs a.
Proof. ir. cp (BZ_gcd_prop1 H inc0_BZ). nin H0.
  assert (BZ_ideal1 a = BZ_ideal2 a BZ_zero). 
  uf BZ_ideal1; uf BZ_ideal2. set_extens; Ztac; clear H2; Ztac. nin H4.
  exists  x0. exists BZ_zero. nin H2. qw. ee; qprops. ue. qprops. 
  nin H4. nin H2. ee. qwii H5. exists x0. eee.  app BZ_stable_mult.
  wri H2 H1. cp (BZ_ideal_unique_generator (sub_BZp_BZ H0) H H1). 
  rwi (BZ_abs_pos H0) H3. am.
Qed.

Lemma BZ_non_zero_quo_gcd: forall a b, inc a BZ -> inc b BZ -> 
  a <> BZ_zero -> BZquo a (BZ_gcd a b) <> BZ_zero.
Proof. ir. cp (BZ_gcd_diva H H0). dneg. rwi H3 H2. qwi H2. am. qprops.
Qed.

Lemma BZ_positive_quo_gcd: forall a b, inc a BZp -> inc b BZ -> 
  inc (BZquo a (BZ_gcd a b)) BZp.
Proof. ir.  app BZp_stable_quo. qprops.
Qed.


Definition BZ_Bezout a b :=
  exists u, exists v, inc u BZ & inc v BZ &
    BZ_one = BZplus (BZmult a u) (BZmult b v).

Lemma BZ_exists_Bezout1: forall a b, inc a BZ -> inc b BZ ->
  BZ_coprime a b -> BZ_Bezout a b.
Proof. ir. red in H1. cp (BZ_gcd_prop2 H H0). rwii H1 H2.
Qed.

Lemma BZ_exists_Bezout2: forall a b, inc a BZ -> inc b BZ ->
  BZ_Bezout a b -> BZ_coprime a b.
Proof. ir. red. nin H1. nin H1. ee.
  rwi (BZ_gcd_diva H H0) H3. set (g := BZ_gcd a b) in H3. 
  rwi (BZ_gcd_divb H H0) H3. fold g in H3. assert (inc g BZ). uf g; qprops.
  assert (inc (BZquo b g) BZ). qprops.  assert (inc (BZquo a g) BZ). qprops.
  wrii BZ_mult_assoc H3. wrii BZ_mult_assoc H3.
  wri BZmult_plus_distr_r H3. 
  assert (inc (BZplus (BZmult (BZquo a g) x) (BZmult (BZquo b g) x0)) BZ).
  app BZ_stable_plus.  app BZ_stable_mult.  app BZ_stable_mult.
  symmetry in H3. nin (BZmult_1_inversion_l H4 H7 H3). nin H9. am.
  assert (inc g BZp). uf g. qprops.  assert (inc g BZms). rw H9. qprops.
  elim (BZ_disjoint_neg_pos H11 H10). am. app BZ_stable_mult. 
  app BZ_stable_mult.
Qed.


Lemma BZ_exists_Bezout3: forall a b, inc a BZ -> inc b BZ ->
 (a<> BZ_zero \/ b <> BZ_zero) ->
  BZ_Bezout (BZquo a (BZ_gcd a b)) (BZquo b (BZ_gcd a b)).
Proof. ir. nin (BZ_gcd_prop2 H H0). nin H2. ee. set (g := BZ_gcd a b) in *. 
  rwi (BZ_gcd_diva H H0) H4. fold g in H4.  
  rwi (BZ_gcd_divb H H0) H4. fold g in H4. assert (inc g BZ). uf g. qprops.
  exists x. exists x0. eee; qprops.
  assert (inc (BZquo b g) BZ). qprops.  assert (inc (BZquo a g) BZ). qprops.
  wrii BZ_mult_assoc H4. wrii BZ_mult_assoc H4.
  wri BZmult_plus_distr_r H4.  apply BZmult_reg_l with g. am. qprops.
  app BZ_stable_plus.  app BZ_stable_mult.  app BZ_stable_mult.
  uf g. app BZ_gcd_nonzero1. qw. am.  app BZ_stable_mult.  app BZ_stable_mult.
Qed.


Lemma BZ_divides_order_r: 
  order_r (fun a b => inc a BZps & inc b BZps & BZ_divides a b).
Proof. red. ee. red.  ir. eee. apply BZ_divides_trans with y; fprops.
  red. ir. ee. cp (BZdivides_pr H4). cp (BZdivides_pr H2).
  set (u:= BZquo x y) in *. set (v := BZquo y x) in *.
  assert (inc u BZ). uf u. qprops. assert (inc v BZ). uf v. qprops.
  rwi H5 H6. wri BZ_mult_assoc H6. 
  nin (BZmult_1_inversion_more  (sub_BZps_BZ H) H8 H7 H6).
  rwi inc_BZps H.  nin H. contradiction. nin H9.  rwi H9 H5. qwi H5. sy; am.
  qprops. rwi H9 H5. rwi BZ_mult_m1_r H5. cp (BZopp_positive1 H). wri H5 H10.
  elim (BZ_disjoint_neg_spos H10 H3). qprops. qprops. am. am.
  red. ir. nin H. nin H0. eee.  assert (inc x BZ). qprops.
  rwi inc_BZps H. ee. nin (BZdivision_itself H2 H3). am.
  assert (inc y BZ). qprops.
  rwi inc_BZps H0. ee. nin (BZdivision_itself H2 H3). am.
Qed.

Lemma BZdivides_pr6: forall a b, inc a BZ -> inc b BZ -> a<> BZ_zero ->
  b <> BZ_zero ->
  (BZ_divides a b = sub (BZ_ideal1 b)(BZ_ideal1 a)).
Proof. ir. ap iff_eq. ir. cp (BZdivides_pr H3). uf BZ_ideal1.
  red. ir. Ztac. nin H7. clear H5. nin H7. Ztac. 
  exists (BZmult  (BZquo b a) x0). assert (inc  (BZquo b a) BZ). qprops.
  split.  app BZ_stable_mult.  rww BZ_mult_assoc. wrr H4. ir.
  assert (inc b (BZ_ideal1 a)). app H3. nin (BZ_in_ideal4 H0). am.
  ufi BZ_ideal1 H4. Ztac. nin H6. ee. rw H7. app BZdivides_pr1.
Qed.

Lemma BZ_gcd_prop3: forall a b, inc a BZ -> inc b BZ ->
  (a<> BZ_zero \/ b <> BZ_zero) ->
  (BZ_divides (BZ_gcd a b) a & BZ_divides (BZ_gcd a b) b &
    forall u, inc u BZ -> BZ_divides u a -> BZ_divides u b ->
      BZ_divides u  (BZ_gcd a b)).
Proof. ir. cp (BZ_gcd_nonzero1 H H0 H1). 
  assert (inc (BZ_gcd a b) BZ). qprops.
  cp (BZ_gcd_diva H H0). cp (BZ_gcd_divb H H0). set (g := BZ_gcd a b) in *. 
  split. rw H4. app  BZdivides_pr1.  qprops. split.  rw H5. 
  app  BZdivides_pr1.  qprops.
  ir.  cp (BZdivides_pr H7). cp (BZdivides_pr H8).
  nin (BZ_gcd_prop2 H H0).  nin H11. ee.  fold g in H13.
  assert (inc (BZquo b u) BZ). qprops. assert (inc (BZquo a u) BZ). qprops.
  rwi H9 H13; rwi H10 H13.  wrii BZ_mult_assoc H13.  wrii BZ_mult_assoc H13.
  assert(inc (BZmult (BZquo a u) x) BZ). app BZ_stable_mult. 
  assert (inc (BZmult (BZquo b u) x0) BZ).  app BZ_stable_mult.
  wrii BZmult_plus_distr_r H13. 
  set (w:= (BZplus (BZmult (BZquo a u) x) (BZmult (BZquo b u) x0))) in *.
  assert (inc w BZ). uf w. app BZ_stable_plus. 
  rw H13. app  BZdivides_pr1. qprops. dneg. rwi H19 H13. qwi H13. am. am.
Qed.

Lemma BZ_gcd_prop4: forall a b x, inc a BZ -> inc b BZ -> inc x BZ ->
  (a<> BZ_zero \/ b <> BZ_zero) ->
  BZ_divides x a -> BZ_divides x b ->
  (forall u, inc u BZ -> BZ_divides u a -> BZ_divides u b -> BZ_divides u x)
  ->  BZ_gcd a b = BZabs x.
Proof. ir.  cp (BZ_gcd_nonzero1 H H0 H2). 
  assert (inc  (BZ_gcd a b) BZ). qprops. 
  cp (BZ_gcd_prop3 H H0 H2). ee.
  cp (H5 _ H7 H8 H9). cp (H10 _ H1 H3 H4).
  cp (BZdivides_pr H12). cp (BZdivides_pr H11).
  set (g := BZ_gcd a b) in *. rwi H13 H14.  wri BZ_mult_assoc H14.
  wri H13 H14. assert (inc (BZquo g x) BZ). qprops.
  assert (inc (BZquo x g) BZ). qprops. 
  assert (inc g BZp). nin (BZ_gcd_prop1 H H0). am.
  nin (BZmult_1_inversion_more H1 H15 H16 H14). rwi H18 H13.
  qwi H13. contradiction. qprops.
  nin H18. rwi H18 H13. qwi H13.  rw H13. wr H13. rww BZ_abs_pos. am.
  rwi H18 H13. rwii BZ_mult_m1_r H13. cp (f_equal BZopp H13).
  rwi BZopp_involutive H19. wr H19. rww BZ_abs_neg. rww BZopp_involutive.
  app BZopp_positive1. rw inc_BZps. split. am. am. am. am. qprops.
  app BZ_stable_quo.  app BZ_stable_mult. qprops.
Qed.

Lemma BZ_lcm_zero : forall a, inc a BZ -> BZ_lcm a BZ_zero = BZ_zero.
Proof. ir. uf BZ_lcm. qw.  assert (inc (BZ_gcd a BZ_zero) BZ). qprops.
  nin (BZ_division_zero H0). am.
Qed.

Lemma BZ_stable_lcm: forall a b, inc a BZ -> inc b BZ -> 
  inc (BZ_lcm a b) BZ.
Proof. ir.  uf BZ_lcm. app BZ_stable_quo.  app BZ_stable_mult. qprops.
Qed.

Hint Resolve  BZ_stable_lcm : qprops.

Lemma BZ_lcm_prop1 : forall a b, inc a BZ -> inc b BZ ->
  a <> BZ_zero -> b <> BZ_zero ->
  let g := BZ_gcd a b in
    (BZ_lcm a b = BZmult (BZquo a g) b & BZ_lcm a b = BZmult a (BZquo b g)
      &  BZ_lcm a b = BZmult (BZmult (BZquo a g) (BZquo b g)) g).
Proof. ir. uf BZ_lcm. assert (a<> BZ_zero \/ b <> BZ_zero). au.
  cp (BZ_gcd_prop3 H H0 H3).  nin H4. nin H5. assert (inc g BZ). uf g. qprops.
  cp (BZdivides_pr H4). cp (BZdivides_pr H5).
  set (t :=  BZquo (BZmult a b) (BZ_gcd a b)). fold g in t.
  set (u := BZquo a g) in *. set (v := BZquo b g) in *.
  set (t1 := BZmult (BZmult u v) g).
  assert (inc u BZ). uf u. qprops.  assert (inc v BZ). uf v. qprops.
  assert (t1 = BZmult u b).  uf t1.  wrr BZ_mult_assoc. rw (BZ_mult_comm v g).
  ue.  assert (t1 = BZmult a v).  uf t1. rw BZ_mult_comm. rww BZ_mult_assoc. ue.
  assert (BZmult a b = BZmult g t1). rw (BZ_mult_comm g t1). rw H13.
  wrr BZ_mult_assoc.  rw (BZ_mult_comm v g). ue.
  assert (inc t1 BZ). uf t1. app BZ_stable_mult.  app BZ_stable_mult. 
  cp (BZ_gcd_nonzero1 H H0 H3). 
  cp (BZdivides_pr2 H15 H7 H16 H14). fold t in H17. wr H17. au.
Qed.


Lemma BZ_lcm_nonzero : forall a b, inc a BZ -> inc b BZ ->
  a <> BZ_zero -> b <> BZ_zero -> BZ_lcm a b <> BZ_zero.
Proof. ir.  cp (BZ_lcm_prop1 H H0 H1 H2). ee. red. ir. simpl in H3. ee.
  rwi H4 H3. set (u := BZquo a (BZ_gcd a b)).  nin (equal_or_not u BZ_zero).
  assert (a<> BZ_zero \/ b <> BZ_zero). au.
  cp (BZ_gcd_prop3 H H0 H8). ee. clear H11.
  assert (inc (BZ_gcd a b) BZ). qprops. cp (BZdivides_pr H9). 
  fold u in H12. rwi H7 H12. qwi H12. contradiction. am.
  assert (inc u BZ). uf u. qprops. fold u in H3. cp (BZmult_0_rl H0 H8 H2 H7).
  rwi H3 H9. nin H9.  app BZ_mult_comm.
Qed.

Lemma BZ_divide_monotone: forall a b, inc a BZp -> inc b BZps ->
  BZ_divides a b -> BZ_le a b.
Proof. ir. cp (BZdivides_pr H1).
  assert (inc (BZquo b a) BZ). qprops. rwi inc_BZ1 H3. nin H3.
  rwi H3 H2. qwi H2. rwi inc_BZps H0. ee. contradiction. qprops.
  nin H3.   assert (BZ_le a a). app BZ_le_reflexive. qprops.
  cp (BZ_prod_increasing8 H H3 H4). ue.
  assert (inc a BZps). rw inc_BZps. split. am. nin H1. ee. am.
  assert (inc b BZms). rw H2. app BZpms_stable_mult.
  elim (BZ_disjoint_neg_spos H5 H0).
Qed.

Lemma BZ_lcm_prop2 : forall a b, inc a BZ -> inc b BZ ->
  a <> BZ_zero -> b <> BZ_zero ->
  (BZ_divides a (BZ_lcm a b) & BZ_divides b (BZ_lcm a b) &
    forall u, inc u BZ -> BZ_divides a u -> BZ_divides b u->
      BZ_divides  (BZ_lcm a b) u).
Proof. ir. assert (a<> BZ_zero \/ b <> BZ_zero). au.
  cp (BZ_gcd_prop3 H H0 H3).  nin H4. nin H5. clear H6.
  cp  (BZ_lcm_prop1 H H0 H1 H2). simpl in H6. ee. rw H7.
  app BZdivides_pr1. qprops. rw H6. rw BZ_mult_comm. app BZdivides_pr1. 
  qprops.  ir. set (g := BZ_gcd a b) in *.
  cp (BZ_exists_Bezout3 H H0 H3). fold g in H12.
  assert (inc g BZ). uf g. qprops. 
  cp (BZdivides_pr H4). cp (BZdivides_pr  H5). clear H4; clear H5.
  set (aq:= BZquo a g) in *. set (bq:= BZquo b g) in *.
  nin H12. nin H4. ee.
  assert (u = BZmult u BZ_one). qw. rwi H12 H16. 
  assert (inc aq BZ). uf aq. qprops. assert (inc bq BZ). uf bq. qprops.
  assert (inc (BZmult aq x) BZ).  app BZ_stable_mult. 
  assert(inc (BZmult bq x0) BZ).  app BZ_stable_mult. 
  rwii BZmult_plus_distr_r H16. assert (inc  (BZ_lcm a b) BZ). qprops.
  set (f1 := BZmult u (BZmult aq x)). assert (f1 = BZmult u (BZmult aq x)). tv.
  cp (BZdivides_pr H11). rwi H23 H22. rwii BZ_mult_permute_2_in_4 H22.
  rwi (BZ_mult_comm b aq) H22. wri H6 H22.
  cp (BZdivides_pr  H10). fold f1 in H16. 
  set (f2 := BZmult u (BZmult bq x0)). assert (f2 = BZmult u (BZmult bq x0)).
  tv. rwi H24 H25. rwii BZ_mult_permute_2_in_4 H25. wri H7 H25.
  fold f2 in H16. rwi H25 H16. rwi H22 H16. wrii BZmult_plus_distr_r H16. 
  rw H16. app  BZdivides_pr1.
  app BZ_stable_plus.  app BZ_stable_mult.  qprops.  app BZ_stable_mult. 
  qprops. app BZ_lcm_nonzero.  app BZ_stable_mult.  qprops.  app BZ_stable_mult. qprops.
  qprops. qprops.
Qed.

Lemma BZ_lcm_prop3 : forall a b u, inc a BZ -> inc b BZ -> inc u BZ -> 
  BZ_coprime a b -> 
  BZ_divides a u -> BZ_divides b u->  BZ_divides  (BZmult a b) u.
Proof. ir. assert (a <> BZ_zero). red in H3. ee. am. 
  assert (b <> BZ_zero). red in H4. ee. am.
  cp (BZ_lcm_prop2 H H0 H5 H6). ee. assert (inc (BZ_lcm a b) BZ).
  qprops. cp (H9 _ H1 H3 H4). red in H2. ufi BZ_lcm H11. rwi H2 H11.
  rwi BZquotient_by_one H11. am.  app BZ_stable_mult. 
Qed.

Lemma BZ_exists_Bezout5: forall a, inc a BZ -> BZ_coprime a BZ_one.
Proof. ir. app  BZ_exists_Bezout2.  qprops. exists BZ_zero. exists BZ_one.
  eee. qprops. qprops. qw. qprops. qprops.
Qed.


End RationalIntegers.
Export  RationalIntegers.
